# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import dataclasses
import datetime
import re
import time
from dataclasses import dataclass
from datetime import timezone
from pprint import pformat
from typing import List, Union, Optional

FIELD_NOT_PROVIDED = None
VALUE_NOT_PROVIDED = ""
DATETIME_EPOCH_UTC = datetime.datetime(1970, 1, 1).replace(tzinfo=timezone.utc)


class EventType(object):
    JOB = "JOB"
    TASK = "TASK"
    BOOT = "BOOT"
    BLOCK = "BLOCK"
    VNODE = "VNODE"
    RESERVATION = "RESERVATION"
    NODE_STATE = "NODE_STATE"
    LICENSE = "LICENSE"
    STATE = "STATE"


class RecType(object):
    #
    # PBS Record Types
    # see acct.h in OpenPBS
    X = "X"
    S = "S"  # PBS Job Start
    E = "E"  # PBS Job End
    D = "D"  # PBS Job Delete
    Q = "Q"  # PBS Job Queue
    A = "A"  # PBS Job Abort
    R = "R"  # PBS Job rerun
    C = "C"  # PBS Job checkpoint
    T = "T"  # PBS Job checkpoint_restart
    U = "U"  # PBS Reservation requested by user (awaiting confirmation)
    Y = "Y"  # PBS Reservation alteration confirmed by system
    B = "B"  # PBS Reservation begin
    F = "F"  # PBS Reservation finish
    K = "K"  # PBS Reservation removed by system
    k = "k"  # PBS Reservation removed by user
    L = "L"  # License
    a = "a"  # PBS Job alter
    r = "r"  # PBS Job resume
    z = "z"  # PBS Job suspend
    e = "e"  # PBS vnode released, shows job info, final phase
    c = "c"  # PBS vnode upcoming phase
    s = "s"  # PBS vnode request was trimmed
    u = "u"  # PBS vnode release, shows job info
    P = "P"  # PBS vnode provision start for job or reservation
    p = "p"  # PBS vnode provision end for job or reservation

    dct = {
        S: {"action": "start", "record": S, "event_type": EventType.JOB},
        E: {"action": "end", "record": E, "event_type": EventType.JOB},
        D: {"action": "delete", "record": D, "event_type": EventType.JOB},
        Q: {"action": "queue", "record": Q, "event_type": EventType.JOB},
        A: {"action": "abort", "record": A, "event_type": EventType.JOB},
        R: {"action": "rerun", "record": R, "event_type": EventType.JOB},
        C: {"action": "checkpoint", "record": C, "event_type": EventType.JOB},
        T: {"action": "checkpoint_restart", "record": T, "event_type": EventType.JOB},
        U: {"action": "unconfirmed", "record": U, "event_type": EventType.RESERVATION},
        Y: {"action": "confirmed", "record": Y, "event_type": EventType.RESERVATION},
        B: {"action": "begin", "record": B, "event_type": EventType.RESERVATION},
        F: {"action": "finish", "record": F, "event_type": EventType.RESERVATION},
        K: {"action": "system_remove", "record": K, "event_type": EventType.RESERVATION},
        k: {"action": "remove", "record": k, "event_type": EventType.RESERVATION},
        L: {"action": "license", "record": L, "event_type": EventType.LICENSE},
        a: {"action": "alter", "record": a, "event_type": EventType.JOB},
        r: {"action": "resume", "record": r, "event_type": EventType.JOB},
        z: {"action": "suspend", "record": z, "event_type": EventType.JOB},
        e: {"action": "released", "record": e, "event_type": EventType.VNODE},
        c: {"action": "upcoming", "record": c, "event_type": EventType.VNODE},
        s: {"action": "trip", "record": s, "event_type": EventType.VNODE},
        u: {"action": "release", "record": u, "event_type": EventType.VNODE},
        P: {"action": "provision_start", "record": P, "event_type": EventType.VNODE},
        p: {"action": "provision_end", "record": p, "event_type": EventType.VNODE},
        X: {"action": "none", "record": X, "event_type": EventType.STATE},
        None: {"action": None, "record": None, "event_type": None},
    }


class LOG_REGEX(object):
    """This class contains certain useful attributes for working with PBS accounting logs"""

    LOG_FORMAT_PBS = "%m/%d/%Y %H:%M:%S"
    LOG_FORMAT_PBS_US = "%m/%d/%Y %H:%M:%S.%f"
    LOG_FORMAT_PBS_US_MSB_NO_SPACE = "%Y-%m-%dT%H:%M:%S.%f"
    LOG_FORMAT_ISO8601 = "%Y-%m-%dT%H:%M:%S.%f"
    LOG_FORMAT_PBS_GROUPED = re.compile(
        r"^(?P<MONTH>[0-9]{2})/(?P<DAY>[0-9]{2})/(?P<YEAR>[0-9]{4}) "
        + r"(?P<HOUR>[0-9]{2}):(?P<MIN>[0-9]{2}):(?P<SEC>[0-9]{2})(?P<DATA>.*)$"
    )
    LOG_FORMAT_DATA_SEPERATOR_PBS = ";"
    LOG_FORMAT_DATA_SEPERATOR_PBS_RE = re.compile(r";([a-zA-Z]{1,3});")
    LOG_FORMAT_KV_SEPERATOR_PBS = " "
    PATTERN_KEY_VALUE_PAIRS = re.compile(r"^(?:\s*([\w.\-]+)=((?:([\"])(?:\\\3|(?!\3).)*\3|[^\"]*?)+))(\s*)(?=\s+[\w.\-]+=|$)")
    RECTYPE = re.compile(r"^([A-Za-z]+);.*")


def join_record(record_value, record_id, record_keyvals):
    return ";".join((record_value, record_id, record_keyvals))


def split_record(data):
    record_value, identifier, record_keyvals = data.split(";", 2)
    return record_value.strip(), identifier.strip(), record_keyvals


def format_key_value_pairs(kv_pairs: dict) -> str:
    """
    Given a dictionary, flatten it into a key value string used by PBS
    """
    kv_pair_lst = []
    for key, value in kv_pairs.items():
        # if "'" in value:
        #     #value = value.replace("'", "\'")
        #     value = f'"{str(value)}"'
        # if '"' in value:
        #     #value = value.replace('"', '\"')
        #     value = f"'{str(value)}'"
        kv_pair_lst.append(f"{key}={value}")
    return LOG_REGEX.LOG_FORMAT_KV_SEPERATOR_PBS.join(kv_pair_lst)


def func_dquote(s):
    return s.strip('"').replace('\\"', '"') if (s.startswith('"') and s.endswith('"')) else s


def func_squote(s):
    return s.strip("'").replace("\\'", "'") if (s.startswith("'") and s.endswith("'")) else s


def parse_key_value_pairs(row: str, record_value=None):
    """This parser will pivot around the = sign"""
    row = row.strip()
    dct = {}
    pos = 0
    # parse row and create dict, making sure there are no duplicate keys
    exit_status_lst = []
    while pos < len(row):
        tail = row[pos:]
        match = LOG_REGEX.PATTERN_KEY_VALUE_PAIRS.match(tail)
        if tail.find("=") != -1 and match:
            pos += match.end(4) + 1
            key, value, _, _ = match.groups()
            if key == "Exit_status":
                # there can be more than one Exit_status which can related to errors, gather them up.
                exit_status_lst.append(value)
                dct[key] = value
                continue
            elif key == "Variable_List":  # this can contain bash functions.
                # include everything because parsing this would take much more code.
                dct[key] = tail
                break
            if key in dct:
                # make sure the dups have the same value, if so then ok
                if dct[key] != value:
                    raise BadLineError(f'PBS KEY({key}) ERROR: {dct[key]=} != {value=}\n"{row}"')
                else:
                    continue
            dct[key] = value
        elif tail.find("=") == -1:
            # this fixes a message only line:
            # 10/26/2021 16:24:55;A;53.pdw-s1;Job deleted as result of dependency on job 52.pdw-s1
            dct["message"] = tail
            pos = len(row)
        else:
            # if this is an abort record then just accept this and move on.
            #  This is to handles lines like this:
            #  PBS KEY ERROR: parse error at char 1 in "Reject reply code=15001, aux=0, type=52, ..."
            if record_value == RecType.A:
                dct["message"] = tail
                pos = len(row)
            else:
                raise BadLineError(f'PBS KEY ERROR: parse error at char {pos + 1} in "{row}"')
    if "Exit_status" in dct:
        # we need to now correct the Exit_status
        if len(exit_status_lst) > 1:
            int_exit_status_lst = []
            oth_exit_status_lst = []
            for es in exit_status_lst:
                try:
                    es = int(es)
                except ValueError:
                    oth_exit_status_lst.append(es)
                else:
                    int_exit_status_lst.append(es)
            max_exit_status = max([_ for _ in int_exit_status_lst if isinstance(_, int)])
            if max_exit_status is not None:
                dct["Exit_status"] = str(max_exit_status)
            else:
                dct["Exit_status"] = sorted(oth_exit_status_lst).pop() if oth_exit_status_lst else "None"
            dct["Exit_list"] = ",".join([str(_) for _ in exit_status_lst])

    # fix the quotes on the values.  Note, this loses and adds data.
    dct = dict(map(lambda e: (e[0], func_squote(func_dquote(e[1].strip()))), dct.items()))
    return dct


class BadLineError(Exception):
    pass


class BadTimestampError(Exception):
    pass


def fix_newlines_extract_time(lines, include_lineno=False, version=1, iso8601=False):
    """
    times_and_lines is the list of tuples that will contain the good
    scheduler_timestamps and lines

    good_lines is recreating the file but with the fixed newlines
    it is used to walk backwards to find the good line and update
    it in place without updating lines
    """
    assert type(lines) == list, "type(lines) = %s should be a List[str]" % type(lines)

    def _sanitize(_):
        return _.replace("\n", " ")

    times_and_lines = []
    good_lines = []
    for lineno, raw_line in enumerate(lines):
        if version > 1:
            # This correctly sets the lineno to 1 for the first line of the file.  It could mess up past data.
            lineno = lineno + 1
        line = _sanitize(raw_line)
        try:
            scheduler_timestamp, data = pbs_line_extract_time(line, lineno=lineno, iso8601=iso8601)
        except BadLineError:
            # remove the last line of parsed data
            # remove the last good line and update it
            if not times_and_lines:
                continue
            times_and_lines.pop()  # it was bad/missing data
            # go get the previous line
            previous_lineno, previous_line = good_lines.pop()
            bigger_line = previous_line + line
            scheduler_timestamp, data = pbs_line_extract_time(bigger_line, lineno=lineno, iso8601=iso8601)
            good_lines.append((previous_lineno, bigger_line))
            if version > 1:
                # This sets the line to the root line.  It could mess up past data.
                lineno = previous_lineno
        else:
            good_lines.append((lineno, line))
        if include_lineno is True:
            result = (scheduler_timestamp, lineno, data)
        else:
            result = (scheduler_timestamp, data)
        times_and_lines.append(result)
    return times_and_lines


def pbs_line_extract_time(line, lineno=-1, iso8601=False):
    """pull the time and data out of a PBS line."""
    re_separator = LOG_REGEX.LOG_FORMAT_DATA_SEPERATOR_PBS_RE
    parsed = re.split(re_separator, line, 1)
    if len(parsed) == 3:
        # This was done to fix a problem with bash functions inside Variable_List.  - ERP
        scheduler_timestamp, record_value, data = parsed
        data = f"{record_value};{data}"
    else:
        raise BadLineError(f"problem splitting line#{lineno} using sep:{str(re_separator)}")
    try:
        if iso8601:
            scheduler_timestamp = datetime.datetime.strptime(scheduler_timestamp, LOG_REGEX.LOG_FORMAT_ISO8601)
        else:
            scheduler_timestamp = datetime.datetime.strptime(scheduler_timestamp, LOG_REGEX.LOG_FORMAT_PBS)
    except ValueError:
        raise BadTimestampError(f"problem converting line#{lineno} ts:{scheduler_timestamp}")
    return scheduler_timestamp, data


def seconds_to_hms(seconds: Union[int, float], use_days=False):
    one_day = 86400
    if use_days:
        days = seconds // one_day
        seconds = seconds % one_day
    else:
        days = 0
    microseconds = seconds % 1
    seconds = seconds // 1
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    days = int(days)
    hours = int(hours)
    minutes = int(minutes)
    seconds = int(seconds)
    microseconds = round(microseconds)
    if days:
        if days > 1:
            day_str = "days" if days > 1 else "day"
        else:
            day_str = "days" if days < -1 else "day"
        result = f"{days} {day_str}, {hours:0>2}:{minutes:0>2}:{seconds:0>2}"
    else:
        if hours:
            result = f"{hours:0>2}:{minutes:0>2}:{seconds:0>2}"
        else:
            if minutes:
                result = f"{minutes:0>2}:{seconds:0>2}"
            else:
                result = f"{seconds:0>2}"
    if microseconds:
        result = f"{result}.{microseconds:0>6}"
    return result


def hms_to_seconds(hms_str):
    """
    a standard hms_str has a format of hh:mm:ss
    hms_str may have only mm:ss or ss portions
    """
    if not hms_str or not isinstance(hms_str, str):
        return None
    hms_ls = hms_str.split(":")
    len_hms_ls = len(hms_ls)
    if len_hms_ls == 3:
        hours, minutes, seconds = hms_ls
    elif len_hms_ls == 2:
        hours = 0
        minutes, seconds = hms_ls
    elif len_hms_ls == 1:
        # seconds only
        hours = 0
        minutes = 0
        (seconds,) = hms_ls
    else:
        raise Exception(f"Invalid hms string {hms_str}")

    try:
        hours = int(hours) * 3600
        minutes = int(minutes) * 60
        seconds = int(seconds)
    except ValueError as err:
        raise ValueError(f"Invalid hms string.  Error:{err}")
    if hours < 0 or minutes < 0 or seconds < 0:
        result = 0
    else:
        result = hours + minutes + seconds
    return result


def get_time(scheduler_timestamp, pattern):
    """given a PBS log time, return a time object."""
    datetimeobj = datetime.datetime.strptime(scheduler_timestamp, pattern)
    datetimeobj = datetimeobj.replace(tzinfo=timezone.utc)
    scheduler_timestamp = datetimeobj.timestamp()
    return scheduler_timestamp


class Parse_Error_PBS(Exception):
    pass


def cast_to_int(value):
    return int(value) if value else None


def cast_to_list(value, list_delimiter):
    try:

        def cl(d, x):
            return x if type(x) == list else str(x).split(d)

        return None if value is None else cl(list_delimiter, value)
    except AttributeError:
        return None


def cast_to_vnode_list(value):
    node_group = re.split(r"[+]", value)
    node_lst = []
    for head in node_group:
        # pull off non-node values
        if "*" in head:
            # partial:  mom2/0*4
            # full:     mom2/0*4+mom1/0+mom1/1+mom1/2+mom1/3+mom1/4+mom1/5+mom1/6
            head_tail = head.rsplit("*", 1)
            if len(head_tail) == 1:
                head = head_tail[0]
            else:  # max split is two
                head, ncpus = head_tail
        elif ":" in head:
            # works w/at least these two cases:
            # partial:  (mom2:ncpus=2)
            # full:     (mom2:ncpus=2)+(mom2:ncpus=2)
            # partial:  (pdw-c3:ncpus=3:mem=1kb)
            # full:     (pdw-c3:ncpus=3:mem=1kb)+(pdw-c3:ncpus=3:mem=1kb)+(pdw-c4:ncpus=3:mem=1kb)
            head = head.strip("(").strip(")")
            head_tail = head.rsplit(":", 1)
            while len(head_tail) > 1:
                head_tail = head.rsplit(":")
                head = head_tail[0]
        elif head.count("(") == 1 and head.count(")") == 1:
            # partial:  (pdw-c03)
            head = head.strip("(").strip(")")
        # FIXME: this only works for the * case in head;
        #        otherwise it will only split the last head, need more examples!

        head_tail = head.rsplit("/", 1)
        if len(head_tail) == 1:
            node_job_index = 0
            head = head_tail[0]
        else:
            head, node_job_index = head_tail
        node_name = head
        node_lst.append(node_name)
    return node_lst


def cast_to_float(str_val, def_val=None):
    try:
        return None if str_val is None else float(str_val)
    except ValueError:
        return def_val


def cast_to_list_of_x(str_val, list_delimiter, item_cast_func):
    lst = cast_to_list(str_val, list_delimiter)
    if lst is not None:
        lst = list(map(item_cast_func, lst))
    return lst


def cast_to_place_dict(value) -> dict:
    # arrangement, sharing, group
    if value:
        valid_arrangement = ["pack", "scatter", "free", "vscatter"]
        valid_sharing = ["shared", "excl", "exclhost"]
        dct = dict(arrangement=None, sharing=None, groups=[])
        items = value.split(":")
        for item in items:
            if item in valid_arrangement:
                dct["arrangement"] = item
            elif item in valid_sharing:
                dct["sharing"] = item
            elif item.startswith("group="):
                _, group = item.split("=", 1)
                dct["groups"].append(group)
            # else:
            #     raise Exception("Illegal attribute or resource value Resource_List.place")
    else:
        dct = {}  # this allows an if check
    return dct


def join_place_dict(dct):
    place = []
    if dct:
        arrangement = dct["arrangement"]
        sharing = dct["sharing"]
        groups = dct["groups"]
        if arrangement:
            place.append(arrangement)
        if sharing:
            place.append(sharing)
        if groups:
            for group in groups:
                place.append(f"group={group}")
    return ":".join(place)


def cast_to_select_list(value) -> List[dict]:
    if not value:
        chunks = []
    else:
        chunk_lst = re.split(r'\+(?=(?:[^"]*"[^"]*")*[^"]*$)', value)
        chunks = []
        for chunk in chunk_lst:
            chunk_dct = {}
            sub_chunk_lst = re.split(r':(?=(?:[^"]*"[^"]*")*[^"]*$)', chunk)
            for sub_chunk in sub_chunk_lst:
                if "=" not in sub_chunk:
                    N = int(sub_chunk)
                    chunk_dct["N"] = N
                else:
                    # even if multiple are pushed in via qsub, only the last is taken thus overwriting the value is acceptable.
                    resource, sub_value = sub_chunk.split("=", 1)
                    if sub_value.startswith('"') and sub_value.endswith('"'):
                        sub_value = sub_value.strip('"')
                    chunk_dct[resource] = sub_value
            chunks.append(chunk_dct)
    return chunks


def join_select_list(lst):
    select = []
    for chunk_dct in lst:
        sub_select = []
        if "N" in chunk_dct:
            sub_select.append(str(chunk_dct["N"]))
        for key, value in chunk_dct.items():
            if key != "N":
                if ":" in value or "+" in value and not (value.startswith('"') and value.endswith('"')):
                    sub_select.append(f'{key}="{value}"')
                else:
                    sub_select.append(f"{key}={value}")
        select.append(":".join(sub_select))
    result = "+".join(select)
    return result


def timedelta_to_microtime(timedeltaobj):
    """Convert a timedelta object to microseconds"""
    return timedeltaobj.microseconds + (timedeltaobj.seconds + timedeltaobj.days * 86400) * 1000000


def epoch_to_datetime(time_val):
    # 0 is possible as a default and can really mess things up, thus "not timeval"
    # NOTE: loses precision.
    return None if not time_val else datetime.datetime(*time.gmtime(time_val)[0:6])


def datetime_to_epoch(datetime_obj: datetime) -> float:
    """Given a datetime object with timezone,
    return the number of seconds in epoch"""
    if datetime_obj.tzinfo is None:
        datetime_obj = datetime_obj.replace(tzinfo=timezone.utc)
    delta = datetime_obj - DATETIME_EPOCH_UTC
    return timedelta_to_microtime(delta) / 1000000


def convert_none_str_to_empty_list(str_val):
    return [] if str_val == "None" else str_val


def _get_requester(key_value_dct):
    requester = key_value_dct.get("requester", FIELD_NOT_PROVIDED)
    if requester == FIELD_NOT_PROVIDED:
        requester = key_value_dct.get("requestor", FIELD_NOT_PROVIDED)
    return requester


class FieldHandlers:
    encode_passthrough_lst = [
        "action",
        "record",
        "event_type",
        "identifier",
        "sub_identifier",
        "resource",
        "action",
        "requester",
        "raw_user_name",
        "raw_project_name",
        "raw_account_name",
        "raw_exec_vnode",
    ]

    # used when creating PBS records.
    encode_key_rename = {
        "raw_project_name": "project",
        "raw_account_name": "account",
        "raw_user_name": "user",
        "raw_exec_vnode": "exec_vnode",
    }

    # These are the keys that are set by the parser
    builtin_keys = [
        "action",
        "record",
        "event_type",
        "identifier",
        "sub_identifier",
        "scheduler_timestamp",
    ]

    parsers_from_str_lookup = {}
    encoders_to_str_lookup = {}
    record_handler = {}

    @staticmethod
    def update_parsers(dct):
        FieldHandlers.parsers_from_str_lookup.update(dct)

    @staticmethod
    def update_encoders(dct):
        FieldHandlers.encoders_to_str_lookup.update(dct)

    @staticmethod
    def update_record_handler(dct):
        FieldHandlers.record_handler.update(dct)

    @staticmethod
    def process_resource(field_handler, rec_type, resource, key_value_dct, dct):
        """resource can be sent in to be applied to the record, but it also can be in the record.
        If it's in the record, use the one in the record.
        """
        if "resource" in dct:
            new_resource = field_handler.parse_from_str("resource", key_value_dct)
            # if resource is not None:
            #     # make sure they match.
            #     assert new_resource == resource
            if new_resource is not None:
                dct["resource"] = new_resource

    @staticmethod
    def parse_from_str(field: str, key_value_dct: dict, bypass=False):
        try:
            parser = FieldHandlers.parsers_from_str_lookup[field]
        except KeyError:
            if bypass:
                parser = FieldHandlers.parse_pass_or_to_notprovided
            else:
                raise Exception(f"Parser for {field=} not found.")
        result = parser(field, key_value_dct)
        return result

    @staticmethod
    def parse_full_record(field_handler, rec_type, resource, key_value_dct, dct):
        """modifies and updates incoming dct with the parser result.
        must maintain the same signature as the other parsers"""
        for key, value in key_value_dct.items():
            dct[key] = field_handler.parse_from_str(key, key_value_dct, bypass=True)
        return dct

    @staticmethod
    def parse_record_no_builtin(field_handler, rec_type, resource, key_value_dct, dct):
        """modifies and updates incoming dct with the parser result.
        must maintain the same signature as the other parsers"""
        for key in field_handler.parsers_from_str_lookup:
            if key not in field_handler.builtin_keys:
                if key in key_value_dct:
                    dct[key] = field_handler.parse_from_str(key, key_value_dct)
        return dct

    @staticmethod
    def parse_pass_or_to_notprovided(field: str, key_value_dct: dict) -> object:
        result = key_value_dct.get(field, FIELD_NOT_PROVIDED)
        return result

    @staticmethod
    def parse_epoch_str_to_datetime(field: str, key_value_dct: dict) -> Optional[datetime.datetime]:
        result = epoch_to_datetime(cast_to_float(key_value_dct.get(field, FIELD_NOT_PROVIDED)))
        return result

    @staticmethod
    def parse_str_to_int(field: str, key_value_dct: dict) -> Optional[int]:
        result = cast_to_int(key_value_dct.get(field, FIELD_NOT_PROVIDED))
        return result

    @staticmethod
    def parse_hms_to_seconds(field: str, key_value_dct: dict) -> Optional[int]:
        result = hms_to_seconds(key_value_dct.get(field, FIELD_NOT_PROVIDED))
        return result

    @staticmethod
    def parse_vnode_lst(field: str, key_value_dct: dict) -> Optional[list]:
        value = key_value_dct.get(field, FIELD_NOT_PROVIDED)
        result = cast_to_vnode_list(value) if value else value
        return result

    @staticmethod
    def parse_exit_status(field: str, key_value_dct: dict) -> Optional[list]:
        try:
            result = str(int(key_value_dct.get(field, FIELD_NOT_PROVIDED)))
        except (TypeError, ValueError):
            result = FIELD_NOT_PROVIDED
        return result

    @staticmethod
    def parse_x_used(field: str, key_value_dct: dict) -> Optional[list]:
        result = cast_to_list_of_x(key_value_dct.get(field, FIELD_NOT_PROVIDED), ":", cast_to_int)
        return result

    @staticmethod
    def parse_place(field: str, key_value_dct: dict) -> Optional[dict]:
        value = key_value_dct.get(field, FIELD_NOT_PROVIDED)
        result = cast_to_place_dict(value)
        return result

    @staticmethod
    def parse_select(field: str, key_value_dct: dict) -> Optional[list]:
        value = key_value_dct.get(field, FIELD_NOT_PROVIDED)
        result = cast_to_select_list(value)
        return result

    @staticmethod
    def parse_pbs_ts_to_datetime(field: str, key_value_dct: dict) -> Optional[list]:
        value = key_value_dct.get(field, FIELD_NOT_PROVIDED)
        if value:
            # value2 = dateutil.parser.parse(value)
            value = datetime.datetime.strptime(value, LOG_REGEX.LOG_FORMAT_ISO8601)
            if hasattr(value, "tzinfo") and value.tzinfo:
                value = value.replace(tzinfo=None)
        return value

    # ^^^ parsers ^^^
    # vvv encoders vvv

    @staticmethod
    def encode_to_str(field: str, key_value_dct: dict, attempt_str=False):
        try:
            encoder = FieldHandlers.encoders_to_str_lookup[field]
        except KeyError:
            if attempt_str:
                encoder = FieldHandlers.encode_any_to_str
            else:
                raise Exception(f"Encoder for {field=} not found.")
        result = encoder(field, key_value_dct)
        return result

    @staticmethod
    def encode_full_record(field_handler, rec_type, resource, key_value_dct, dct):
        """modifies and updates incoming dct with the parser result."""
        result = ""
        for key in dct:
            result = field_handler.encode_to_str(key, key_value_dct, attempt_str=True)
        return result

    @staticmethod
    def encode_any_to_str(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, None)
        result = str(value) if value is not None else VALUE_NOT_PROVIDED
        return result

    @staticmethod
    def encode_datetime_to_epoch_str(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, None)
        # valid characters
        # quote only if a space.
        result = str(datetime_to_epoch(value))
        return result

    @staticmethod
    def encode_int_to_str(field: str, key_value_dct: dict) -> str:
        return FieldHandlers.encode_any_to_str(field, key_value_dct)

    @staticmethod
    def encode_float_to_str(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, None)
        result = str(value) if value is not None else VALUE_NOT_PROVIDED
        return result

    @staticmethod
    def encode_str_to_str(field: str, key_value_dct: dict):
        value = key_value_dct.get(field, None)
        result = value if value is not None else VALUE_NOT_PROVIDED
        return result

    @staticmethod
    def encode_datetime_to_str(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, None)
        # valid characters
        # quote only if a space.
        # result = value.strftime(LOG_REGEX.LOG_FORMAT_PBS_US)
        result = value.strftime(LOG_REGEX.LOG_FORMAT_PBS)
        return result

    @staticmethod
    def encode_datetime_to_iso8601str(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, None)
        # valid characters
        # quote only if a space.
        # result = value.strftime(LOG_REGEX.LOG_FORMAT_PBS_US)
        if type(value) == str:
            result = value.replace("+00:00", "") if value.endswith("+00:00") else value
        else:
            result = value.strftime(LOG_REGEX.LOG_FORMAT_ISO8601)
        return result

    @staticmethod
    def encode_seconds_to_hms(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, None)
        result = seconds_to_hms(value) if value is not None else VALUE_NOT_PROVIDED
        return result

    @staticmethod
    def encode_vnode_lst(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, None)
        result = "+".join(value) if value is not None else VALUE_NOT_PROVIDED
        return result

    @staticmethod
    def encode_exit_status(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, None)
        result = str(value) if value is not None else VALUE_NOT_PROVIDED
        return result

    @staticmethod
    def encode_x_used(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, FIELD_NOT_PROVIDED)
        result = ":".join(value) if value else FIELD_NOT_PROVIDED
        return result

    @staticmethod
    def encode_place(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, FIELD_NOT_PROVIDED)
        result = join_place_dict(value)
        return result

    @staticmethod
    def encode_select(field: str, key_value_dct: dict) -> str:
        value = key_value_dct.get(field, FIELD_NOT_PROVIDED)
        result = join_select_list(value)
        return result


FieldHandlers.update_parsers(
    {
        "ctime": FieldHandlers.parse_epoch_str_to_datetime,
        "qtime": FieldHandlers.parse_epoch_str_to_datetime,
        "etime": FieldHandlers.parse_epoch_str_to_datetime,
        "stime": FieldHandlers.parse_epoch_str_to_datetime,
        "start": FieldHandlers.parse_epoch_str_to_datetime,
        "end": FieldHandlers.parse_epoch_str_to_datetime,
        "Resource_List.nodect": FieldHandlers.parse_str_to_int,
        "Resource_List.ncpus": FieldHandlers.parse_str_to_int,
        "Resource_List.ngpus": FieldHandlers.parse_str_to_int,
        "Resource_List.nproc": FieldHandlers.parse_str_to_int,
        "Resource_List.place": FieldHandlers.parse_place,
        "Resource_List.select": FieldHandlers.parse_select,
        "account": FieldHandlers.parse_pass_or_to_notprovided,
        "project": FieldHandlers.parse_pass_or_to_notprovided,
        "event_type": FieldHandlers.parse_pass_or_to_notprovided,
        "resource": FieldHandlers.parse_pass_or_to_notprovided,
        "sub_identifier": FieldHandlers.parse_pass_or_to_notprovided,
        "requester": FieldHandlers.parse_pass_or_to_notprovided,
        "run_count": FieldHandlers.parse_str_to_int,
        "Resource_List.walltime": FieldHandlers.parse_hms_to_seconds,
        "exec_host": FieldHandlers.parse_vnode_lst,
        "exec_vnode": FieldHandlers.parse_vnode_lst,
        "nodes": FieldHandlers.parse_vnode_lst,
        "Exit_status": FieldHandlers.parse_exit_status,
        "resources_used.walltime": FieldHandlers.parse_hms_to_seconds,
        "resources_used.nodect": FieldHandlers.parse_str_to_int,
        "resources_used.nodes_used": FieldHandlers.parse_x_used,
        "resources_used.cores_used": FieldHandlers.parse_x_used,
        "session": FieldHandlers.parse_pass_or_to_notprovided,
        "args": FieldHandlers.parse_pass_or_to_notprovided,
        "queue": FieldHandlers.parse_pass_or_to_notprovided,
        "mode": FieldHandlers.parse_pass_or_to_notprovided,
        "jobname": FieldHandlers.parse_pass_or_to_notprovided,
        "resvID": FieldHandlers.parse_pass_or_to_notprovided,
        "resvname": FieldHandlers.parse_pass_or_to_notprovided,
        "duration": FieldHandlers.parse_pass_or_to_notprovided,
        "group": FieldHandlers.parse_pass_or_to_notprovided,
        "exe": FieldHandlers.parse_pass_or_to_notprovided,
        "cwd": FieldHandlers.parse_pass_or_to_notprovided,
        "user": FieldHandlers.parse_pass_or_to_notprovided,
    }
)


FieldHandlers.update_encoders(
    {
        "ctime": FieldHandlers.encode_datetime_to_epoch_str,
        "qtime": FieldHandlers.encode_datetime_to_epoch_str,
        "etime": FieldHandlers.encode_datetime_to_epoch_str,
        "stime": FieldHandlers.encode_datetime_to_epoch_str,
        "start": FieldHandlers.encode_datetime_to_epoch_str,
        "end": FieldHandlers.encode_datetime_to_epoch_str,
        "Resource_List.nodect": FieldHandlers.encode_int_to_str,
        "Resource_List.ncpus": FieldHandlers.encode_int_to_str,
        "Resource_List.ngpus": FieldHandlers.encode_int_to_str,
        "Resource_List.nproc": FieldHandlers.encode_int_to_str,
        "Resource_List.place": FieldHandlers.encode_place,
        "Resource_List.select": FieldHandlers.encode_select,
        "scheduler_timestamp": FieldHandlers.encode_datetime_to_str,
        "run_count": FieldHandlers.encode_int_to_str,
        "Resource_List.walltime": FieldHandlers.encode_seconds_to_hms,
        "exec_host": FieldHandlers.encode_vnode_lst,
        "nodes": FieldHandlers.encode_vnode_lst,
        "Exit_status": FieldHandlers.encode_exit_status,
        "resources_used.walltime": FieldHandlers.encode_seconds_to_hms,
        "resources_used.nodect": FieldHandlers.encode_int_to_str,
        "resources_used.nodes_used": FieldHandlers.encode_x_used,
        "resources_used.cores_used": FieldHandlers.encode_x_used,
        "account": FieldHandlers.encode_any_to_str,
        "project": FieldHandlers.encode_any_to_str,
        "action": FieldHandlers.encode_any_to_str,
        "record": FieldHandlers.encode_any_to_str,
        "event_type": FieldHandlers.encode_any_to_str,
        "session": FieldHandlers.encode_any_to_str,
        "user": FieldHandlers.encode_any_to_str,
        "queue": FieldHandlers.encode_any_to_str,
        "args": FieldHandlers.encode_any_to_str,
        "mode": FieldHandlers.encode_any_to_str,
        "jobname": FieldHandlers.encode_any_to_str,
        "resvname": FieldHandlers.encode_any_to_str,
        "resvID": FieldHandlers.encode_any_to_str,
        "duration": FieldHandlers.encode_any_to_str,
        "group": FieldHandlers.encode_any_to_str,
        "exe": FieldHandlers.encode_any_to_str,
        "cwd": FieldHandlers.encode_any_to_str,
    }
)


def process_record_Reservation(field_handler, rec_type, resource, key_value_dct, dct):
    dct["sub_identifier"] = key_value_dct.get("active_id", FIELD_NOT_PROVIDED)
    dct["requester"] = _get_requester(key_value_dct)
    dct["duration"] = cast_to_int(key_value_dct.get("duration", FIELD_NOT_PROVIDED))
    for key, value in key_value_dct.items():
        if key.startswith("Resource_List."):
            dct[key] = field_handler.parse_from_str(key, key_value_dct, bypass=True)
        elif key.startswith("resources_used."):
            dct[key] = field_handler.parse_from_str(key, key_value_dct, bypass=True)
    dct["Authorized_Users"] = cast_to_list(
        convert_none_str_to_empty_list(key_value_dct.get("Authorized_Users", FIELD_NOT_PROVIDED)),
        ",",
    )
    dct["owner"] = key_value_dct.get("owner", FIELD_NOT_PROVIDED)
    dct["users"] = cast_to_list(convert_none_str_to_empty_list(key_value_dct.get("users", FIELD_NOT_PROVIDED)), ",")
    dct["name"] = key_value_dct.get("name", FIELD_NOT_PROVIDED)
    dct["queue"] = key_value_dct.get("queue", FIELD_NOT_PROVIDED)
    return dct


def process_record_Job_SER(field_handler, rec_type, resource, key_value_dct, dct):
    dct["raw_user_name"] = dct["user"]
    dct["raw_exec_vnode"] = key_value_dct.get("exec_vnode", FIELD_NOT_PROVIDED)
    # pbs job arrays don't run on anything thus don't provide an exec_host
    dct["Resource_List.nodes_estimated"] = cast_to_list_of_x(
        key_value_dct.get("Resource_List.nodes_estimated", FIELD_NOT_PROVIDED), ":", cast_to_int
    )
    dct["Resource_List.cores_estimated"] = cast_to_list_of_x(
        key_value_dct.get("Resource_List.cores_estimated", FIELD_NOT_PROVIDED), ":", cast_to_int
    )
    if dct["record"] == rec_type.E:
        dct["Exit_status"] = field_handler.parse_from_str("Exit_status", key_value_dct)
        dct["resources_used.nodect"] = cast_to_int(key_value_dct.get("resources_used.nodect", FIELD_NOT_PROVIDED))
        dct["resources_used.nodes_used"] = cast_to_list_of_x(
            key_value_dct.get("resources_used.nodes_used", FIELD_NOT_PROVIDED), ":", cast_to_int
        )
        dct["resources_used.cores_used"] = cast_to_list_of_x(
            key_value_dct.get("resources_used.cores_used", FIELD_NOT_PROVIDED), ":", cast_to_int
        )
    return dct


def process_record_Job_Q(field_handler, rec_type, resource, key_value_dct, dct):
    return dct


def process_record_Job_D(field_handler, rec_type, resource, key_value_dct, dct):
    dct["requester"] = _get_requester(key_value_dct)
    return dct


FieldHandlers.update_record_handler(
    {
        RecType.A: lambda field_handler, rec_type, resource, key_value_dct, dct: dct,
        RecType.Y: process_record_Reservation,
        RecType.B: process_record_Reservation,
        RecType.F: process_record_Reservation,
        RecType.K: process_record_Reservation,
        RecType.k: process_record_Reservation,
        RecType.S: process_record_Job_SER,
        RecType.E: process_record_Job_SER,
        RecType.R: process_record_Job_SER,
        RecType.Q: process_record_Job_Q,
        RecType.D: process_record_Job_D,
    }
)


field_handler = FieldHandlers


def parse_pbs_organize_keyvals(dct, resource, key_value_dct, rec_type=None):
    """This merges dct and key_value_dct.
    if rec_type is specified, it will use the version given"""
    dct = dct.copy()
    # required fields for all records
    dct["resource"] = resource if resource else key_value_dct.get("resource", FIELD_NOT_PROVIDED)
    dct["sub_identifier"] = key_value_dct.get("sub_identifier", "0")
    rec_type = rec_type if rec_type else RecType
    for key in ["account", "project", "user"]:
        dct[key] = key_value_dct.get(key, FIELD_NOT_PROVIDED)
        if key in key_value_dct:
            if dct[key] != FIELD_NOT_PROVIDED:
                dct[f"raw_{key}_name"] = dct[key]

    record = dct["record"]
    field_handler.parse_record_no_builtin(field_handler, rec_type, resource, key_value_dct, dct)
    if record in field_handler.record_handler:
        func = field_handler.record_handler[record]
        dct = func(field_handler, rec_type, resource, key_value_dct, dct)
    else:
        dct = field_handler.parse_full_record(field_handler, rec_type, resource, key_value_dct, dct)
    field_handler.process_resource(field_handler, rec_type, resource, key_value_dct, dct)

    # create final dictionary that removes any fields that are tagged not provided
    final_dct = {}
    for k in dct:
        if dct[k] != FIELD_NOT_PROVIDED or k in ["event_type", "record"]:
            final_dct[k] = dct[k]
    return final_dct


def enhance_pbs_record(timestamp: datetime.datetime, record_value, identifier, dct=None, rec_type: RecType = None):
    """given pbs record data, enhance the record"""
    if not dct:
        dct = {}
    else:
        dct = dct.copy()
    rec_type = rec_type if rec_type else RecType
    try:
        dct_record_value = rec_type.dct[record_value]
    except (KeyError, IndexError):
        # no record type found
        dct_record_value = rec_type.dct[None]
    dct.update(**dct_record_value)
    # required fields for all records
    dct["identifier"] = identifier
    dct["scheduler_timestamp"] = timestamp
    if "sub_identifier" not in dct:
        dct["sub_identifier"] = "0"
    return dct


def parse_pbs_log_line(timestamp: datetime.datetime, data, resource=None, rec_type: RecType = None):
    """Parse the pbs log lines"""
    rec_type = rec_type if rec_type else RecType
    record_value, identifier, record_keyvals = split_record(data)
    dct = enhance_pbs_record(timestamp, record_value, identifier, rec_type=rec_type)
    if record_value == rec_type.L:
        dct["resource"] = resource if resource else FIELD_NOT_PROVIDED
        dct["sub_identifier"] = "0"
        dct["data"] = record_keyvals
        result = dct
    else:
        key_value_dct = parse_key_value_pairs(record_keyvals, record_value=record_value)
        result = parse_pbs_organize_keyvals(dct, resource, key_value_dct, rec_type)
    return result


def create_pbs_log_line(
    timestamp: datetime.datetime,
    record: str,
    identifier: str,
    kv_pairs: dict,
    attempt_str=False,
    add_extra=False,
    rec_type=None,
    iso8601=False,
):
    """
    timestamp: scheduler timestamp
    record RecType.*
    identifier: jobid or reservationid or the word license, could be an uuid
    kv_pairs:
    attempt_str will tell the encoder to execute str(x) to convert it to a line if there is no encoder found.

    """
    rec_type = rec_type if rec_type else RecType
    assert record in rec_type.dct, f"record {record} is not a valid RecType"
    sep = LOG_REGEX.LOG_FORMAT_DATA_SEPERATOR_PBS
    kv_sep = LOG_REGEX.LOG_FORMAT_KV_SEPERATOR_PBS
    ts_us = timestamp.strftime(LOG_REGEX.LOG_FORMAT_ISO8601)
    if iso8601:
        ts = ts_us
    else:
        ts = timestamp.strftime(LOG_REGEX.LOG_FORMAT_PBS)
    # make sure identifier is a specific size and characters.
    head = f"{ts}{sep}{record}{sep}{identifier}"
    kv_pair_lst = []

    # rename the fields to overwrite any keys that already may have that name.
    for key, value in field_handler.encode_key_rename.items():
        if key in kv_pairs:
            new_key = field_handler.encode_key_rename[key]
            kv_pairs[new_key] = kv_pairs[key]

    for key in kv_pairs.keys():
        if key in field_handler.encode_key_rename:
            key = field_handler.encode_key_rename[key]
        if key in field_handler.builtin_keys:
            continue  # we already have these.
        if key in field_handler.encode_passthrough_lst:
            value = kv_pairs.get(key, None)
            value = value if value is not None else VALUE_NOT_PROVIDED
        else:
            value = field_handler.encode_to_str(key, kv_pairs, attempt_str=attempt_str)
        kv_pair_lst.append(f"{key}={value}")
    if add_extra:
        kv_pair_lst.append(f"ts={ts_us}")

    tail = kv_sep.join(kv_pair_lst)
    line = f"{head}{sep}{tail}"
    return line


@dataclass
class PBSRecordKVPairs:
    ts: str
    te: str
    kwargs: dataclasses.field(default_factory=dict)

    def __post_init__(self):
        [setattr(self, k, v) for k, v in self.kwargs.items() if v is not None]


@dataclass
class PBSRecordBase:
    timestamp: datetime
    record: str
    identifier: str
    kv_pairs: PBSRecordKVPairs

    def get_dct(self):
        dct = self.kv_pairs.kwargs.copy()
        ts = self.kv_pairs.ts
        te = self.kv_pairs.te
        if ts:
            dct["ts"] = ts
        if te:
            dct["te"] = te
        dct["timestamp"] = self.timestamp
        dct["record"] = self.record
        dct["identifier"] = self.identifier
        return dct


def create_pbs_record(timestamp: datetime.datetime, record, identifier, dct) -> PBSRecordBase:
    """
    create a PBSRecordBase.
    The keys in dct of ts and te are special and will be flattened into strings.
    """
    ts_obj = dct["ts"].strftime(LOG_REGEX.LOG_FORMAT_ISO8601) if "ts" in dct else None
    te_obj = dct["te"].strftime(LOG_REGEX.LOG_FORMAT_ISO8601) if "te" in dct else None
    rkvp = PBSRecordKVPairs(ts_obj, te_obj, dct)
    rb = PBSRecordBase(timestamp, record, identifier, rkvp)
    return rb


def create_pbs_log_line_from_record(record_obj, attempt_str=False, iso8601=False):
    dct = record_obj.get_dct()
    dct_clean = dict((k, v) for k, v in dct.items() if v is not None)  # remove nulls
    del dct_clean["timestamp"]  # this is sent in the next line
    if "scheduler_timestamp" in dct_clean:
        del dct_clean["scheduler_timestamp"]
    line = create_pbs_log_line(
        record_obj.timestamp, record_obj.record, record_obj.identifier, dct_clean, attempt_str=attempt_str, iso8601=iso8601
    )
    return line
