# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import errno
import re
from functools import partial
from PBS_Utils.pbs_enum import PBS_BATCH, PBS_SCHED, PBS_REPLY, DIS_ERR
from PBS_Utils.pbs_enum import LogEventTypes

try:
    import pbs
except ImportError:
    pbs = None


def resolver(enum, template, match):
    result = None
    match_group = match.group(1)
    if match_group is not None:
        try:
            result = template.format(name=enum(int(match_group)).name, orig=match_group)
        except:
            result = template.format(name="???", orig=match_group)
    result = result.replace("PBS_BATCH_", "")
    return result


def resolver_f(func, template, match):
    result = None
    value = match.group(1)
    if value is not None:
        try:
            result = template.format(name=func(int(value)), orig=value)
        except:
            result = template.format(name="???", orig=value)
    return result


def convert_state(template, match):
    result = None
    if match:
        vnode_state_lst = []
        value = match.group(1)
        value_int = int(value, 16)
        # bit_indexes = [2**i for i in range(16) if value_int & 2**i > 0]
        if value_int == pbs.ND_STATE_FREE:
            vnode_state_lst.append("ND_STATE_FREE")
        else:
            vnode_state_lst = [val for (mask, val) in sorted(pbs.REVERSE_NODE_STATE.items()) if value_int & mask]
        try:
            vnode_state_str = ",".join(vnode_state_lst)
            result = template.format(name=vnode_state_str, orig=value)
        except:
            result = template.format(name="???", orig=value)
    return result


def ERRNO(value):
    if value == 0:
        result = "0"
    else:
        result = errno.errorcode[value]
    return result


DECODERS = [
    # all decoders must have {name}({orig})
    ("Req;;Type {name}({orig})", re.compile(r"Req;;Type (\d+)"), partial(resolver, PBS_BATCH)),
    ("Scheduler sent command {name}({orig})", re.compile(r"Scheduler sent command (\d+)"), partial(resolver, PBS_SCHED)),
    ("Reject reply code={name}({orig})", re.compile(r"Reject reply code=(\d+)"), partial(resolver, PBS_REPLY)),
    ("dis error {name}({orig})", re.compile(r"dis error (\d+)"), partial(resolver, DIS_ERR)),
    ("errno {name}({orig})", re.compile(r"errno[ =](\d+)"), partial(resolver_f, ERRNO)),
]
if pbs:
    DECODERS.append((".state={name}({orig})", re.compile(r"\.state=0x([0-9a-fA-F]+)"), convert_state))
    DECODERS.append(("state_bits={name}({orig})", re.compile(r"state_bits=0x([0-9a-fA-F]+)"), convert_state))
# Updated job state to 72 and substate to 20

DECODERS = list(
    [(template.replace("{name}({orig})", ""), template, pattern, partial(func, template)) for template, pattern, func in DECODERS]
)


def process_log_event_type(line):
    try:
        ts, event_type, tail = line.split(";", 2)
        event_type_int = int(event_type, 16)
        # LogEventTypes
        event_type_lst = []
        for key, value in dict([(e.name, e.value) for e in LogEventTypes]).items():
            if event_type_int & value:
                event_type_lst.append(key.replace("PBSEVENT_", ""))
        line = f"{ts};{','.join(event_type_lst)}({event_type});{tail}"
    except:
        pass
    return line


def process_log_line(line):
    for precheck, template, pattern, func in DECODERS:
        if precheck in line:
            line = re.sub(pattern, func, line)
    line = process_log_event_type(line)
    return line
