# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import copy
import getpass
import json
import os
import socket
import traceback
import uuid
from datetime import datetime
from PBS_Utils.pbs_util import datetime_to_str, get_now


class EmitterRabbitMQ:
    def __init__(self, pika, rabbitmq_host, rabbitmq_port, event_type, debug_func=None, error_func=None):
        self.pika = pika
        self.rabbitmq_host = rabbitmq_host
        self.rabbitmq_port = rabbitmq_port
        self.event_type = event_type
        self.error_func = error_func
        self.debug_func = debug_func

    def send(self, queue, json_dump):
        pika = self.pika
        rabbitmq_host = self.rabbitmq_host
        rabbitmq_port = self.rabbitmq_port
        connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbitmq_host))
        try:
            channel = connection.channel()
            channel.queue_declare(
                queue=queue,
                durable=True,  # True: D
                auto_delete=False,  # True: AD
            )
            channel.basic_publish(exchange="", routing_key=queue, body=json_dump)
        except Exception:
            lines = traceback.format_exc().strip().splitlines()
            for lineno, line in enumerate(lines):
                self.message_error(f"Error sending rabbitmq message: #{lineno}:Exception: {line}")
        finally:
            connection.close()

    def get_metadata(self, message_type):
        """could move this into its own function."""
        metadata_dct = {
            "hostname": socket.gethostname(),
            "timestamp": datetime_to_str(get_now()),
            "pid": os.getpid(),
            "ppid": os.getppid(),
            "uuid": uuid.uuid4().hex,
            "name": __name__,
            "user": getpass.getuser(),
            "message_type": message_type,
        }
        return metadata_dct

    def emit_pbs_log_line(self, record_queue, record_obj):
        metadata_dct = self.get_metadata(record_queue)
        dct = record_obj.get_dct()
        dct["timestamp"] = datetime_to_str(dct["timestamp"])
        dct["message_type"] = "pbs"
        json_dump = json.dumps({"metadata": metadata_dct, "data": dct}, default=str)
        self.send(record_queue, json_dump)

    def emit_system(self, record_queue, ts: datetime, te: datetime, dct):
        metadata_dct = self.get_metadata(record_queue)
        metadata_dct["ts"] = datetime_to_str(ts)
        metadata_dct["te"] = datetime_to_str(te)
        if self.event_type != "HOOK_EVENT_PERIODIC":
            dct = copy.deepcopy(dct)
        # this structure has metadata, ps, event, server, hook keys.
        dct["metadata"] = metadata_dct
        json_dump = json.dumps(dct, default=str)
        self.send(record_queue, json_dump)

    def message_error(self, msg):
        if self.error_func:
            self.error_func(msg)

    def message_debug(self, msg):
        if self.debug_func:
            self.debug_func(msg)
