# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import os
import uuid
import socket
import getpass
from datetime import datetime, timedelta, timezone

HOSTNAME = socket.gethostname()
utc_tz = timezone(timedelta(seconds=0), name="UTC")


def get_now(notz=False) -> datetime:
    """Return a datetime object right now as the local server time then convert it to utc"""
    if notz:
        result = datetime.utcnow()
    else:
        result = datetime.now(utc_tz)
    return result


def get_now_day_str() -> str:
    datetime_now = get_now()
    datetime_str = datetime_now.strftime("%Y%m%d")
    return datetime_str


def datetime_to_str(obj):
    return obj.strftime("%Y-%m-%dT%H:%M:%S.%f")


def get_metadata(machine_name, message_type):
    metadata_dct = {
        "hostname": HOSTNAME,
        "timestamp": get_now().strftime("%Y-%m-%d %H:%M:%S.%f"),
        "pid": os.getpid(),
        "ppid": os.getppid(),
        "uuid": uuid.uuid4().hex,
        "machine_name": machine_name,
        "message_type": message_type,
        "user": getpass.getuser(),
    }
    return metadata_dct


def timedelta_str_to_td(td_str: str):
    """Convert a timedelta string, such as a walltime into a timedelta."""
    if td_str.startswith("-"):
        negative = True
        td_str = td_str.lstrip("-")
    else:
        negative = False
    if "." in td_str:
        dhms, us = td_str.split(".", 1)
    else:
        us = 0.0
        dhms = td_str

    if "days" in dhms:
        d, hms = dhms.split(" days, ", 1)
    elif "day" in dhms:
        d, hms = dhms.split(" day, ", 1)
    else:
        d = "0"
        hms = dhms
    h, m, s = hms.split(":", 2)
    d = int(d)
    h = int(h)
    m = int(m)
    s = int(s)
    us = int(us)
    if negative:
        d = -d
    td = timedelta(days=int(d), hours=int(h), minutes=int(m), seconds=int(s), microseconds=us)
    return td


def timedelta_to_str(td: timedelta):
    """Convert a timedelta into a string."""
    days = td.days
    seconds = td.seconds
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    microseconds = td.microseconds
    days = int(days)
    hours = int(hours)
    minutes = int(minutes)
    seconds = int(seconds)
    microseconds = round(microseconds)
    if days:
        if days > 1:
            day_str = "days" if days > 1 else "day"
        else:
            day_str = "days" if days < -1 else "day"
        result = f"{days} {day_str}, {hours:0>2}:{minutes:0>2}:{seconds:0>2}"
    else:
        result = f"{hours:0>2}:{minutes:0>2}:{seconds:0>2}"
    if microseconds:
        result = f"{result}.{microseconds:0>6}"
    return result


def df_to_lstofdct(df):
    """convert a dataframe to a list of dct"""
    df = df.reset_index()
    lst_of_dct = list(df.T.to_dict().values())
    return lst_of_dct
