#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import gzip
from pprint import pformat
from glob import glob
import click
import logging
from dataclasses import dataclass
from pathlib import Path
from multiprocessing import Queue, Process, Manager

logging.basicConfig()
log = logging.getLogger(__name__)

from PBS_Utils.pbs_accounting import parse_pbs_log_line, fix_newlines_extract_time


@dataclass
class Settings:
    def __init__(self) -> None:
        self.verbose: int = 0


pass_settings = click.make_pass_decorator(Settings, ensure=True)


@click.group()
@pass_settings
@click.option("--verbose", "-v", help="Increase verbosity level", count=True)
def cli(settings: Settings, verbose: int) -> None:
    settings.verbose = verbose
    if settings.verbose > 2:
        log.setLevel(logging.DEBUG)
    elif settings.verbose > 1:
        log.setLevel(logging.INFO)
    else:
        log.setLevel(logging.WARN)


class WorkComplete:
    def __init__(self, count):
        self.count = count


class FileProcessor:
    def __init__(self, work_queue: Queue, log_queue: Queue, err_queue: Queue):
        self.work_queue = work_queue
        self.log_queue = log_queue
        self.err_queue = err_queue

    def process_file(self, file):
        path = Path(file)
        if str(path).endswith(".gz"):
            opener = gzip.open
        else:
            opener = open

        with opener(path, "rb") as f:
            contents = f.read().decode("utf-8")

        lines = contents.split("\n")
        # fixes linebreaks
        times_and_data = fix_newlines_extract_time(lines, include_lineno=True)
        for logtime, lineno, data in times_and_data:
            try:
                parse_dct = parse_pbs_log_line(logtime, data)
            except Exception as err:
                line = lines[lineno]
                self.err_queue.put(f"X:{file}:{logtime}#{lineno}\n    {err=}    \n{line=}    \n{data=}")
            else:
                self.log_queue.put(f"O:{file}:{logtime}#{lineno}\n    {pformat(parse_dct, indent=4)}")
        return len(times_and_data)


def worker(fp, work_queue: Queue, log_queue: Queue, err_queue: Queue):
    total_result = 0
    while not work_queue.empty():
        # ... race ...
        work = work_queue.get(timeout=0.1)
        print(".", end="")
        result = fp.process_file(work)
        total_result += result
    log_queue.put(WorkComplete(total_result))
    err_queue.put(WorkComplete(total_result))


def worker_logger(log_queue: Queue, complete_count):
    while complete_count > 0:
        message = log_queue.get()
        if type(message) == WorkComplete:
            complete_count -= 1
            log.info(f"worker completed {message.count} lines.")
        else:
            log.debug(message)


def worker_logger_error(err_queue: Queue, complete_count):
    while complete_count > 0:
        message = err_queue.get()
        if type(message) == WorkComplete:
            complete_count -= 1
        else:
            log.error(message)


@cli.command()
@pass_settings
@click.argument("files", nargs=-1)
def test_parse(settings: Settings, files):
    with Manager() as manager:
        file_lst = []
        for file in files:
            file_lst.extend(glob(file))
        work_queue = manager.Queue()
        log_queue = manager.Queue()
        err_queue = manager.Queue()

        fp = FileProcessor(work_queue, log_queue, err_queue)
        for file in files:
            work_queue.put(file)
        processes = 8  # mp.cpu_count()
        process_lst = []
        for i in range(processes):
            process_lst.append(Process(target=worker, args=(fp, work_queue, log_queue, err_queue)))

        for process in process_lst:
            process.start()

        process = Process(target=worker_logger, args=(log_queue, processes))
        process.start()
        process_lst.append(process)

        process = Process(target=worker_logger_error, args=(err_queue, processes))
        process.start()
        process_lst.append(process)

        for process in process_lst:
            process.join()


if __name__ == "__main__":
    cli()
