#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""Experimental!"""

import re
import gzip
import copy
import json
import logging
import os
import socket
import subprocess
import sys
import datetime
from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from pprint import pformat
from typing import Dict, Tuple, List, Optional
import pandas as pd
import numpy as np
from pandas.core.groupby.generic import DataFrameGroupBy

import click

from PBS_Utils.pbs_library import PBSIFLInterface, pbs_load_enums
from PBS_Utils.pbs_util import get_now, df_to_lstofdct
from PBS_Utils.pbs_accounting import cast_to_vnode_list, LOG_REGEX
from PBS_Utils.pbs_accounting import epoch_to_datetime, hms_to_seconds

logging.basicConfig()
log = logging.getLogger(__name__)

np.seterr(over="raise", under="raise")
np.set_printoptions(threshold=sys.maxsize, linewidth=132)
pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)
pd.set_option("display.width", None)
pd.set_option("display.max_colwidth", None)
pd.set_option("display.float_format", "{:.3f}".format)

"""
There needs to be a way to filter the data before the processing and after.  --fin --fout?
Option to get all columns on output, this requires processing the data.
"""


@dataclass
class Settings:
    def __init__(self) -> None:
        self.verbose: int = 0
        self.server = None
        self.pbs_interface = None
        self.format_json = False
        self.data_path = None
        self.data_prefix = None
        self.sortby = None
        self.filt: dict = {}
        self.filt_or: dict = {}
        self.groupby: list = []
        self.agg: Dict[str, List[str]] = {}
        self.header: list = []
        self.bare = False
        self.bare_delim = ","
        self.data_latest = False


pass_settings = click.make_pass_decorator(Settings, ensure=True)


def cb_filt(ctx, param, value) -> dict:
    bad_args = []
    result = {}
    for filt in value:
        try:
            col, value = filt.split("=", 1)
        except ValueError:
            bad_args.append(filt)
        else:
            result[col] = value
    if bad_args:
        raise click.BadParameter(f"filter must be key=value.  Bad filts:{','.join(bad_args)}")
    return result


def cb_filt_multiple(ctx, param, value) -> dict:
    bad_args = []
    result = defaultdict(list)
    for filt in value:
        try:
            col, value = filt.split("=", 1)
        except ValueError:
            bad_args.append(filt)
        else:
            result[col].append(value)
    if bad_args:
        raise click.BadParameter(f"filter must be key=value.  Bad filts:{','.join(bad_args)}")
    return result


def cb_agg(ctx, param, value) -> Dict[str, List[str]]:
    bad_args = []
    result = {}
    for filt in value:
        try:
            col, value = filt.split("=", 1)
        except ValueError:
            bad_args.append(filt)
        else:
            try:
                value_lst = value.split(",")
            except:
                bad_args.append(filt)
            else:
                result[col] = value_lst
    if bad_args:
        raise click.BadParameter(f"agg must be key=value or key=value,value.  Bad filts:{','.join(bad_args)}")
    return result


def cb_header(ctx, param, value) -> List[str]:
    if value:
        value_lst = value.replace(" ", "").split(",")
    else:
        value_lst = []
    return value_lst


def cb_groupby(ctx, param, value) -> List[str]:
    if value:
        value_lst = list(value)
    else:
        value_lst = []
    return value_lst


def process_prefix(ctx, settings) -> List[str]:
    data_path = settings.data_path

    if ctx.invoked_subcommand == "dump":
        if not data_path:
            raise click.BadParameter("--data-path or PLPATH is required.")
        value = get_now().strftime("%Y%m%dT%H%M%S")
    else:
        if settings.data_prefix:
            if not data_path:
                raise click.BadParameter("--data-path or PLPATH is required.")
            value = settings.data_prefix
        elif settings.data_latest:
            if not data_path:
                raise click.BadParameter("--data-path or PLPATH is required.")
            files = os.listdir(data_path)
            pattern = re.compile(r"\d{8}T\d{6}_metadata.json")
            latest_lookup = {}
            for file_name in files:
                match = pattern.match(file_name)
                if match:
                    # this is just for order
                    ts, _ = file_name.split("_", 1)
                    ts = ts.replace("T", ".")
                    ts = float(ts)
                    latest_lookup[ts] = file_name
            file_name = latest_lookup[max(latest_lookup.keys())]
            value, _ = file_name.split("_", 1)
            log.info(f"Using latest data {value} from {data_path}")
        else:
            value = None
    if value:
        value = f"{value}_"
    return value


def cb_latest(ctx, param, value) -> bool:
    if value:
        value = True if value in ["True", True, 1, "1"] else False
    else:
        value = os.environ.get("PLLATEST", False)
        value = True if value in ["True", True, 1, "1"] else False
    return value


def cb_sortby(ctx, param, value) -> str:
    return value


@click.group()
@click.pass_context
@pass_settings
@click.option("--server", default=None, help="server host to connect via ifl")
@click.option("--verbose", "-v", count=True, help="Increase verbosity level")
@click.option("--format-json", "-j", is_flag=True, show_default=True, default=False, help="JSON output")
@click.option("--data-path", default=lambda: os.environ.get("PLPATH", ""), help="The data path.  This can be set via PLPATH")
@click.option("--sortby", "-s", help="--sortby job_state", callback=cb_sortby)
@click.option("--filt", "--filt-and", "-f", help="--filt queue=workq", multiple=True, callback=cb_filt)
@click.option("--filt-or", help="--filt-or vnode=abc --filt-or vnode=def", multiple=True, callback=cb_filt_multiple)
@click.option("--groupby", help="groupby", multiple=True, callback=cb_groupby)
@click.option("--agg", help="aggregate", multiple=True, callback=cb_agg)
@click.option("--header", show_default=True, callback=cb_header, help="The header to display, comma separated list.")
@click.option("--bare", is_flag=True, show_default=True, default=False, help="bare output which can be used for other commands.")
@click.option("--bare-delim", show_default=True, default=",", help="no header")
@click.option(
    "--data-latest",
    "-l",
    is_flag=True,
    show_default=True,
    default=False,
    callback=cb_latest,
    help="Use the latest data in directory.  The default can be set via PLLATEST",
)
@click.option("--data-prefix", help="Select a specific time.")
def cli(
    settings: Settings,
    ctx: click.Context,
    server,
    verbose: int,
    format_json,
    data_path,
    sortby,
    filt,
    filt_or,
    groupby,
    agg,
    header: list,
    bare: bool,
    bare_delim: str,
    data_latest: bool,
    data_prefix,
) -> None:
    settings.server = server
    settings.verbose = verbose
    settings.format_json = format_json
    settings.data_path = data_path
    settings.data_latest = data_latest
    settings.data_prefix = data_prefix
    settings.data_prefix = process_prefix(ctx, settings)
    settings.sortby = sortby
    settings.filt = filt
    settings.filt_or = filt_or
    settings.groupby = groupby
    settings.agg = agg
    settings.header = header
    settings.bare = bare
    settings.bare_delim = bare_delim
    if settings.verbose >= 2:
        log.setLevel(logging.DEBUG)
    elif settings.verbose == 1:
        log.setLevel(logging.INFO)
    else:
        log.setLevel(logging.WARN)


def flatten_key_into_value(_dct: dict, key_type, value: str) -> dict:
    dct = copy.deepcopy(_dct)
    dct[key_type] = value
    return dct


def flatten_dct(_dct: Dict, key_name: str, delim=".") -> Dict:
    dct = {}
    for key, value in _dct.items():
        dct[f"{key_name}{delim}{key}"] = value
    return dct


def flatten_dct_of_dct(_dct: Dict) -> Dict:
    dct = {}
    for key, value in _dct.items():
        value_type = type(value)
        if value_type in [dict]:
            dct.update(flatten_dct(value, key))
        else:
            dct[key] = value
    return dct


def flatten_pbs_structure(data: dict, key_type="identifier"):
    """given a pbs_ifl data structure in the format of an key that is the identifier, move
    that identifier into key_type:identifier within the value dictionary with the rest of the fields.
    Then take any values that are a dictionary and flatten them using the . format for depth."""
    lst = []
    for key, value in data.items():
        lst.append(flatten_key_into_value(value, key_type, key))
    lst = [flatten_dct_of_dct(dct) for dct in lst]
    return lst


def convert_to_bool(value):
    # return True if value in ['True', True, 1, '1'] else False
    return 1 if value in ["True", True, 1, "1"] else 0


def state_to_strings(enum_lookup, key, state):
    mask = True if key.startswith("MASK") else False
    lookup = enum_lookup[key]
    lst = []
    for key, value in lookup.items():
        if mask:
            if value & state:
                lst.append(key)
        else:
            if value == state:
                lst.append(key)
    return lst


def short_id(value):
    return value.split(".", 1)[0]


@dataclass
class PBSData:
    df_jobs: pd.DataFrame
    df_queues: pd.DataFrame
    df_nodes: pd.DataFrame
    df_vnodes: pd.DataFrame
    df_reservations: pd.DataFrame
    df_server: pd.DataFrame
    enum_lookup: dict
    metadata: dict


def query_all(pbs_interface) -> PBSData:
    ts = get_now().strftime("%Y-%m-%dT%H:%M:%S.%f")
    jobs = pbs_interface.get_pbs_jobs(score=True)["data"]
    queues = pbs_interface.get_pbs_queues()["data"]
    nodes = pbs_interface.get_pbs_nodes()["data"]
    vnodes = pbs_interface.get_pbs_vnodes()["data"]
    reservations = pbs_interface.get_pbs_reservations()["data"]
    server = pbs_interface.get_pbs_server()["data"]

    jobs = flatten_pbs_structure(jobs, key_type="jobid")
    queues = flatten_pbs_structure(queues, key_type="queue")
    nodes = flatten_pbs_structure(nodes, key_type="node")
    vnodes = flatten_pbs_structure(vnodes, key_type="vnode")
    reservations = flatten_pbs_structure(reservations, key_type="resvid")
    server = flatten_pbs_structure(server, key_type="server")

    df_jobs = pd.DataFrame(jobs)
    if not df_jobs.empty:
        df_jobs["jobid"] = df_jobs["jobid"].apply(short_id)
        df_jobs = df_jobs.apply(pd.to_numeric, errors="ignore")
        df_jobs["Rerunable"] = df_jobs["Rerunable"].apply(convert_to_bool)
        df_jobs["jobid"] = df_jobs["jobid"].astype(object)

    df_queues = pd.DataFrame(queues)
    df_queues = df_queues.apply(pd.to_numeric, errors="ignore")

    df_queues["started"] = df_queues["started"].apply(convert_to_bool)
    df_queues["enabled"] = df_queues["enabled"].apply(convert_to_bool)
    df_queues = df_queues.rename(columns={"started": "queue_started"})

    df_nodes = pd.DataFrame(nodes)
    df_nodes = df_nodes.apply(pd.to_numeric, errors="ignore")
    df_nodes["resv_enable"] = df_nodes["resv_enable"].apply(convert_to_bool)

    df_vnodes = pd.DataFrame(vnodes)
    df_vnodes = df_vnodes.apply(pd.to_numeric, errors="ignore")
    df_vnodes["resv_enable"] = df_vnodes["resv_enable"].apply(convert_to_bool)
    df_vnodes = df_vnodes.rename(columns={"state": "node_state"})

    df_reservations = pd.DataFrame(reservations)
    if not df_reservations.empty:
        df_reservations["resvid"] = df_reservations["resvid"].apply(short_id)
        df_reservations = df_reservations.apply(pd.to_numeric, errors="ignore")
        df_reservations["resvid"] = df_reservations["resvid"].astype(object)

    df_server = pd.DataFrame(server)
    df_server = df_server.apply(pd.to_numeric, errors="ignore")
    df_server["scheduling"] = df_server["scheduling"].apply(convert_to_bool)
    df_server["query_other_jobs"] = df_server["query_other_jobs"].apply(convert_to_bool)
    df_server["job_history_enable"] = df_server["job_history_enable"].apply(convert_to_bool)
    df_server["resv_enable"] = df_server["resv_enable"].apply(convert_to_bool)
    df_server["flatuid"] = df_server["flatuid"].apply(convert_to_bool)
    df_server["eligible_time_enable"] = df_server["eligible_time_enable"].apply(convert_to_bool)

    metadata = dict(
        hostname=socket.gethostname(),
        timestamp=ts,
    )
    enum_lookup = pbs_load_enums()
    pbsdata = PBSData(df_jobs, df_queues, df_nodes, df_vnodes, df_reservations, df_server, enum_lookup, metadata)
    log.info(f"started queries at {ts}")
    return pbsdata


def wrap_cast_to_vnode_list(value):
    if pd.isna(value):
        result = []
    else:
        result = cast_to_vnode_list(value)
    return result


def wrap_cast_to_vnode_count(value):
    if pd.isna(value):
        result = []
    else:
        result = cast_to_vnode_list(value)
    return len(result)


def split_column_lst_into_series(df, column, new_column):
    """given a dataframe with a column that is a list, split that list into multiple series
    and put the new rows back into the dataframe and remove the column.
    TODO: look into the possible cartesian product here..  Need to remove dups after."""
    if not df.empty:
        # ser = df[column].apply(pd.Series, 1).stack() # FutureWarning
        ser = df[column].apply(lambda s: pd.Series(s, dtype=object), 1).stack()
        ser.index = ser.index.droplevel(-1)
        ser.name = new_column
        df = df.join(ser)
        df.drop([column], axis=1, inplace=True)
    return df


# FIXME: add load and dump with hdf5, requires pytables.
def load_all(data_path, data_prefix):
    log.info(f"loading prefix {data_prefix} from {data_path}")
    df_jobs = pd.read_csv(os.path.join(data_path, f"{data_prefix}df_jobs.csv.gz"), dtype={"jobid": str})
    df_queues = pd.read_csv(os.path.join(data_path, f"{data_prefix}df_queues.csv.gz"), dtype={"queue": str})
    df_nodes = pd.read_csv(os.path.join(data_path, f"{data_prefix}df_nodes.csv.gz"), dtype={"node": str})
    df_vnodes = pd.read_csv(os.path.join(data_path, f"{data_prefix}df_vnodes.csv.gz"), dtype={"vnode": str})
    df_reservations = pd.read_csv(os.path.join(data_path, f"{data_prefix}df_reservations.csv.gz"), dtype={"resvid": str})
    df_server = pd.read_csv(os.path.join(data_path, f"{data_prefix}df_server.csv.gz"))
    with gzip.open(os.path.join(data_path, f"{data_prefix}pbs_enums.json.gz"), "rb") as fd:
        enum_lookup = json.loads(fd.read())
    with gzip.open(os.path.join(data_path, f"{data_prefix}metadata.json.gz"), "rb") as fd:
        metadata = json.loads(fd.read())

    if not df_jobs.empty:
        df_jobs["jobid"] = df_jobs["jobid"].astype(object)
    if not df_reservations.empty:
        df_reservations["resvid"] = df_reservations["resvid"].astype(object)
    pbsdata = PBSData(df_jobs, df_queues, df_nodes, df_vnodes, df_reservations, df_server, enum_lookup, metadata)
    return pbsdata


def get_data(settings):
    if settings.data_path and settings.data_prefix is not None:
        pbsdata = load_all(settings.data_path, settings.data_prefix)
    else:
        pbs_interface = PBSIFLInterface(host=settings.server)
        pbsdata = query_all(pbs_interface)
    log.info(pformat(pbsdata.metadata))
    return pbsdata


@cli.command()
@pass_settings
@click.option("--dump-pbsnodes", "--dpn", is_flag=True, show_default=True, default=False, help="dump pbsnodes")
@click.option("--dump-qstat", "--dqs", is_flag=True, show_default=True, default=False, help="dump qstat")
def dump(settings: Settings, dump_pbsnodes, dump_qstat):
    data_path = settings.data_path
    data_prefix = settings.data_prefix
    pbs_interface = PBSIFLInterface(host=settings.server)
    pbsdata = query_all(pbs_interface)
    df_jobs = pbsdata.df_jobs
    df_queues = pbsdata.df_queues
    df_nodes = pbsdata.df_nodes
    df_vnodes = pbsdata.df_vnodes
    df_reservations = pbsdata.df_reservations
    df_server = pbsdata.df_server
    enum_lookup = pbsdata.enum_lookup
    metadata = pbsdata.metadata
    lst = zip(
        (df_jobs, df_queues, df_nodes, df_vnodes, df_reservations, df_server),
        ("df_jobs", "df_queues", "df_nodes", "df_vnodes", "df_reservations", "df_server"),
    )
    for df, df_name in lst:
        fullpath = os.path.join(data_path, f"{data_prefix}{df_name}.csv.gz")
        df.to_csv(fullpath, compression="gzip")
        print(f"wrote {fullpath}")

    fullpath = os.path.join(data_path, f"{data_prefix}pbs_enums.json.gz")
    with gzip.open(fullpath, "wb") as fd:
        fd.write(json.dumps(enum_lookup).encode("utf-8"))
    print(f"wrote {fullpath}")

    fullpath = os.path.join(data_path, f"{data_prefix}metadata.json.gz")
    with gzip.open(fullpath, "wb") as fd:
        fd.write(json.dumps(metadata).encode("utf-8"))
    print(f"wrote {fullpath}")

    if dump_pbsnodes:
        fullpath = os.path.join(data_path, f"{data_prefix}pbsnodes.json.gz")
        cmd = f"set -o pipefail; pbsnodes -avF json | gzip -9c > {fullpath}"
        ret = subprocess.run(cmd, shell=True)
        if ret.returncode == 0:
            print(f"wrote {fullpath}")
        else:
            print(f"failed to write pbsnodes output to {fullpath}")
    if dump_qstat:
        fullpath = os.path.join(data_path, f"{data_prefix}qstat.json.gz")
        cmd = f"set -o pipefail; qstat -fF json | gzip -9c > {fullpath}"
        ret = subprocess.run(cmd, shell=True)
        if ret.returncode == 0:
            print(f"wrote {fullpath}")
        else:
            print(f"failed to write qstat output to {fullpath}")


def job_state_to_word(value):
    # taken from _base_types.py
    if value == "T":
        value = "transit"
    elif value == "Q":
        value = "queued"
    elif value == "H":
        value = "held"
    elif value == "W":
        value = "waiting"
    elif value == "R":
        value = "running"
    elif value == "E":
        value = "exiting"
    elif value == "X":
        value = "expired"
    elif value == "B":
        value = "begun"
    elif value == "S":
        value = "suspend"
    elif value == "U":
        value = "suspend_user"
    elif value == "M":
        value = "moved"
    elif value == "F":
        value = "finished"
    return value


def logic_avail_nodes(ser) -> Tuple[int, str]:
    jobid = ser["jobid"] if "jobid" in ser else None
    queue_started = ser["queue_started"] if "queue_started" in ser else None
    avail = 1
    comment_lst = []
    if ser["node_state"] != "free":
        avail = 0
        comment_lst.append("Node state is not free.")
    if (
        pd.notna(ser["node_state"])
        and "resv-exclusive" in ser["node_state"]
        and (pd.isna(queue_started) or pd.isnull(queue_started))
    ):
        avail = 0
        comment_lst.append("Node state is bad.")
    if queue_started and pd.notna(queue_started) and pd.notnull(queue_started):
        avail = 0
        comment_lst.append("Resv running.")
    if jobid and pd.notna(jobid) and pd.notnull(jobid):
        avail = 0
        comment_lst.append("Job running.")
    comment = " ".join(comment_lst)
    return avail, comment


def logic_avail_resv_nodes(ser) -> Tuple[int, str]:
    jobid = ser["jobid"] if "jobid" in ser else None
    queue_started = ser["queue_started"] if "queue_started" in ser else None
    resvid = ser["resvid"] if "resvid" in ser else None
    avail = 1
    comment_lst = []
    if ser["node_state"] != "resv-exclusive":
        avail = 0
        comment_lst.append("Node state must be only resv-exclusive")
    if resvid and "free" in ser["node_state"]:  # opposite of problem above, reservation started, but node marked free.
        avail = 0
        comment_lst.append("Node state is bad.")
    if jobid and pd.notna(jobid) and pd.notnull(jobid):
        avail = 0
        comment_lst.append("Job running.")
    comment = " ".join(comment_lst)
    return avail, comment


def df_apply_options(df, settings: Settings):
    sortby = settings.sortby
    filt: dict = settings.filt
    filt_or: dict = settings.filt_or
    groupby = settings.groupby
    agg: Optional[Dict[str, List[str]]] = settings.agg
    header: Optional[list] = settings.header
    if type(df) == pd.DataFrame:
        if df.empty:
            pass
        else:
            pd_and_masks = []
            for col, value in filt.items():
                try:
                    # bool is a special case
                    if df[col].dtype.type == np.bool_:
                        value = convert_to_bool(value)
                    # we need to make the filter compareable via the column type.
                    tvalue = df[col].dtype.type(value)
                    pd_and_masks.append(df[col] == tvalue)
                except KeyError:
                    log.warning(f"filt {col}={value} failed.")
            if pd_and_masks:
                df = df[np.logical_and.reduce(pd_and_masks)]

            pd_or_masks = []
            for col, value_lst in filt_or.items():
                for value in value_lst:
                    try:
                        # bool is a special case
                        if df[col].dtype.type == np.bool_:
                            value = convert_to_bool(value)
                        # we need to make the filter compareable via the column type.
                        try:
                            tvalue = df[col].dtype.type(value)
                        except ValueError:
                            tvalue = value
                        pd_or_masks.append(df[col] == tvalue)
                    except KeyError:
                        log.warning(f"filt_or {col}={value} failed.")
            if pd_or_masks:
                df = df[np.logical_or.reduce(pd_or_masks)]
            if header:
                df = pd.DataFrame(df, columns=header)
            if groupby:
                df = df.groupby(groupby)
            if agg:
                # if type(df) == DataFrameGroupBy and not df.ngroups:
                #    raise click.BadParameter(f"cannot operate an agg on an empty dataframe.")
                if type(df) == DataFrameGroupBy and not df.ngroups:
                    df = pd.DataFrame()
                else:
                    df = df.agg(agg)
                df = df.reset_index()  # this allows df.to_string(index=False) to work.
    elif type(df) == DataFrameGroupBy:
        pass
    if sortby:
        df = df.sort_values(by=[sortby])
    return df


def df_display(df: pd.DataFrame, settings):
    if settings.format_json:
        if df.empty:
            print(pformat({}))
        else:
            print(pformat(df.to_dict()))
    else:
        if settings.bare and len(settings.header) == 1:
            if df.empty:
                print("", end="")
            else:
                col = settings.header[0]
                print(settings.bare_delim.join(df[col].to_list()), end="")
        elif settings.bare and len(settings.header) != 1:
            raise click.BadParameter(f"bare must be selected with exactly one header and it probably should be vnode")
        else:
            if not df.empty:
                df = df.fillna("-")
            print(df.to_string(index=False))


def get_avail(df_queues, df_jobs, df_vnodes, df_reservations, enum_lookup):
    # remove all other queue types
    df_queues = df_queues[df_queues["queue_type"] == "Execution"]  # we don't need routing.
    df_queues = pd.DataFrame(df_queues, columns=["queue", "queue_started"])
    df_jobs = pd.DataFrame(df_jobs, columns=["jobid", "exec_vnode"])
    df_jobs["exec_vnode_lst"] = df_jobs["exec_vnode"].apply(wrap_cast_to_vnode_list)
    df_jobs.drop(["exec_vnode"], axis=1, inplace=True)
    df_jobs = split_column_lst_into_series(df_jobs, "exec_vnode_lst", "vnode")
    if not df_jobs.empty:
        df_jobs = df_jobs[df_jobs["vnode"].notnull()]
    df_vnodes = pd.DataFrame(df_vnodes, columns=["vnode", "node_state"])
    df_reservations = df_reservations.merge(df_queues, on=["queue"], how="left")
    df_reservations = df_reservations[df_reservations["queue_started"] == 1]
    df: pd.DataFrame = df_vnodes
    if not df_jobs.empty:
        df = df.merge(df_jobs, on=["vnode"], how="left")
    if not df_reservations.empty:
        df = df.merge(df_reservations, on=["vnode"], how="left")
    df[["isavailable", "comment"]] = df.apply(logic_avail_nodes, axis=1, result_type="expand")
    df[["isavailable_resv", "comment_resv"]] = df.apply(logic_avail_resv_nodes, axis=1, result_type="expand")
    df = df.sort_values(by=["vnode"])
    df.drop(["resv_nodes", "reserve_state", "reserve_state_lst", "queue_started"], axis=1, inplace=True, errors="ignore")
    df = df.reset_index()
    return df


@cli.command()
@pass_settings
def queue_jobs_agg(settings: Settings):
    agg = settings.agg
    groupby = settings.groupby
    filt = settings.filt
    header = settings.header
    valid_filters = {"job_state", "queue"}
    if filt and not set(filt.keys()).issubset(valid_filters):
        raise click.BadParameter(f"invalid filters provided: {set(filt.keys()) - valid_filters}. Valid:{valid_filters}")

    data = get_data(settings)
    df_jobs = data.df_jobs
    if not df_jobs.empty:
        if "exec_vnode" in df_jobs.columns:
            df_jobs["vnodes"] = df_jobs["exec_vnode"].apply(wrap_cast_to_vnode_count)
        else:
            df_jobs["vnodes"] = 0
        df_jobs["job_state"] = df_jobs["job_state"].apply(job_state_to_word)
        for col, value in filt.items():
            df_jobs = df_jobs[df_jobs[col] == value]
        dfts = df_jobs.groupby(["queue", "job_state"]).agg(count_jobs=("jobid", "count"), count_vnodes=("vnodes", "sum"))
        df = dfts.reset_index()
    else:
        df = df_jobs
    df = df_apply_options(df, settings)
    df_display(df, settings)


@cli.command()
@pass_settings
def vnode(settings: Settings):
    filt = settings.filt

    data = get_data(settings)
    df_vnodes = data.df_vnodes

    valid_filters = set(df_vnodes.columns.to_list())
    if filt and not set(filt.keys()).issubset(valid_filters):
        raise click.BadParameter(f"invalid filters provided: {set(filt.keys()) - valid_filters}. Valid:{valid_filters}")
    df = df_vnodes

    df = df_apply_options(df, settings)
    df_display(df, settings)


def find_subset(df_a, df_b, column):
    """given two dataframes and a column that they share, find if a is a subset of b."""
    # FIXME: this cand be done much better in pandas.
    alstofdct = df_to_lstofdct(df_a[[column]])
    blstofdct = df_to_lstofdct(df_b[[column]])
    lstofdct = []
    for adct in alstofdct:
        for bdct in blstofdct:
            abit = adct[column]
            bbit = bdct[column]
            subset = not (bbit & np.logical_not(abit)).any()
            if subset:
                adct["subset"] = True
            else:
                adct["subset"] = False
            lstofdct.append(adct)
    df = pd.DataFrame(lstofdct, columns=["subset"])
    return df["subset"]


@cli.command()
@click.option("--filt-in-or", help="--filt-in-or node_state_lst=down,offline,state-unknown,Stale", callback=cb_filt, multiple=True)
@pass_settings
def avail(settings: Settings, filt_in_or):
    agg = settings.agg
    groupby = settings.groupby
    filt = settings.filt
    valid_aggs = {"isavailable", "vnode", "isdown", "resvid"}
    if agg and not set(agg.keys()).issubset(valid_aggs):
        raise click.BadParameter(f"invalid agg provided: {set(agg.keys()) - valid_aggs}. Valid:{valid_aggs}")
    valid_groupby = {"isavailable", "resvid"}
    if groupby and not set(groupby).issubset(valid_groupby):
        raise click.BadParameter(f"invalid groupby provided: {set(groupby) - valid_groupby}. Valid:{valid_groupby}")
    valid_filters = {"isavailable", "resources_available.broken", "node_state", "queue", "isdown"}
    if filt and not set(filt.keys()).issubset(valid_filters):
        raise click.BadParameter(f"invalid filters provided: {set(filt.keys()) - valid_filters}. Valid:{valid_filters}")

    data = get_data(settings)
    df_queues = data.df_queues
    df_jobs = data.df_jobs
    df_vnodes = data.df_vnodes
    df_reservations = data.df_reservations
    enum_lookup = data.enum_lookup

    df_queues = df_queues[df_queues["queue_type"] == "Execution"]  # we don't need routing.
    df_reservations = split_reservations_into_nodes(df_reservations, enum_lookup)
    df = get_avail(df_queues, df_jobs, df_vnodes, df_reservations, enum_lookup)
    # the node is available, now we need to get the queue information merged with the vnodes.

    if "queue" in filt:
        # all nodes are not available to all queues.  If we specify a queue, we have to look
        # what is available to it.  If no default chunks, it should be all.
        if type(filt["queue"]) == list and len(filt["queue"]) > 1:
            raise click.BadParameter("filt for avail must have only one queue specified.")
        df_queues, df_vnodes = extract_queue_vnode(df_queues, df_vnodes)
        df_queues = df_queues[df_queues["queue"] == filt["queue"]]
        df_vnodes.drop(["jobs"], axis=1, inplace=True, errors="ignore")
        df = df.merge(df_vnodes, on=["vnode"], how="left")
        subset_mask = find_subset(df, df_queues, "resc_sel_bitmask")
        df = df[subset_mask]
        df = df.rename(columns={"node_state_y": "node_state"})
        df = pd.DataFrame(
            df,
            columns=[
                "vnode",
                "node_state",
                "jobid",
                "resvid",
                "queue",
                "isavailable",
                "comment",
                "isavailable_resv",
                "comment_resv",
            ],
        )
        df = df[df["queue"] == filt["queue"]]
        del settings.filt["queue"]

    df = df.drop_duplicates()
    df["node_state_lst"] = df["node_state"].apply(lambda s: s.split(",") if s and pd.notna(s) else [])

    def func_filt_in_or(filt_in_or, ser):
        include = 0
        for key, filt_lst in filt_in_or.items():
            intersection = set(ser["node_state_lst"]).intersection(set(filt_lst))
            if intersection:
                include = 1
                break
        return include

    # fixme: special case:
    if filt_in_or:
        filt_in_or = dict([(k, v.split(",")) for k, v in filt_in_or.items()])
        df["isdown"] = df.apply(partial(func_filt_in_or, filt_in_or), axis=1)
    df.drop(["node_state_lst"], axis=1, inplace=True, errors="ignore")

    df = df_apply_options(df, settings)
    df_display(df, settings)


def func_bitmask(bitmask_imap, value_lst):
    bitmask = np.zeros(64, dtype=np.int8)
    for value in value_lst:
        try:
            bitmask[bitmask_imap[value]] = 1
        except:
            pass
    return bitmask


def func_resc_sel(queue_resc_cols, columns, sel_rename, ser):
    lst = []
    for col in columns:
        new_column = col.replace(sel_rename, f"{ser[col]}__")
        if new_column in queue_resc_cols:
            value = ser[col]
            if pd.notna(value):
                lst.append(new_column)
    return lst


def extract_queue_vnode(df_queues, df_vnodes):
    """Process default_chunks on the queue and resources_available on the nodes.  Create a bitmask of each the resources
    so that later they can be used for an intersection using the inhib logic.
    for qdct in ..
        for vdct in ..
            qbit = qdct['resc_sel_bitmask']
            vbit = vdct['resc_sel_bitmask']
            subset = not (qbit & np.logical_not(vbit)).any()
    """
    df_queues = df_queues.copy()
    vnode_columns = set(filter(lambda c: c.startswith("resources_available."), df_vnodes.columns.to_list()))
    queue_columns = set(filter(lambda c: c.startswith("default_chunk."), df_queues.columns.to_list()))
    # queues_all = set(df_queues['queue'].unique())  # later we need this to find the queues that have no restrictions.
    queue_resc_cols = set([col.replace("default_chunk.", f"{val}__") for col in queue_columns for val in [True, False]])
    vnode_resc_cols = set([col.replace("resources_available.", f"{val}__") for col in vnode_columns for val in [True, False]])
    assert queue_resc_cols.issubset(
        vnode_resc_cols
    ), f"a resource is defined in default_chunk that doesn't exist on a node. {queue_resc_cols - vnode_resc_cols}"
    bitmask_imap = dict([(c, i) for i, c in enumerate(sorted(list(queue_resc_cols)), start=0)])
    # get the default_chunk.* columns into TRUE_value and FALSE_value or X_value
    df_queues["resc_sel"] = df_queues.apply(partial(func_resc_sel, queue_resc_cols, queue_columns, "default_chunk."), axis=1)
    df_queues["resc_sel_bitmask"] = df_queues["resc_sel"].apply(partial(func_bitmask, bitmask_imap))
    # build the bitmask on the fly right here.
    df_vnodes["resc_sel"] = df_vnodes.apply(partial(func_resc_sel, queue_resc_cols, vnode_columns, "resources_available."), axis=1)
    df_vnodes["resc_sel_bitmask"] = df_vnodes["resc_sel"].apply(partial(func_bitmask, bitmask_imap))
    return df_queues, df_vnodes


def split_reservations_into_nodes(df_reservations, enum_lookup):
    """given the reservations dataframe, get the resv_nodes and create a row, one for each node."""
    df_reservations = pd.DataFrame(df_reservations, columns=["resvid", "reserve_state", "resv_nodes", "queue"])
    df_reservations["reserve_state_lst"] = df_reservations["reserve_state"].apply(
        partial(state_to_strings, enum_lookup, "LONG__REVERSE_RESV_STATE")
    )
    df_reservations["resv_nodes_lst"] = df_reservations["resv_nodes"].apply(wrap_cast_to_vnode_list)
    df_reservations = split_column_lst_into_series(df_reservations, "resv_nodes_lst", "vnode")
    df_reservations["reserve_state_lst"] = df_reservations["reserve_state_lst"].apply(lambda x: ",".join(x))
    df_reservations.drop(
        [
            "resv_nodes",
            "reserve_state",
        ],
        axis=1,
        inplace=True,
        errors="ignore",
    )
    df_reservations = df_reservations.rename(columns={"reserve_state_lst": "reserve_state"})
    return df_reservations


@cli.command()
@pass_settings
def queue_avail(settings: Settings):
    filt = settings.filt
    valid_filters = {"queue", "resvid"}
    if filt and not set(filt.keys()).issubset(valid_filters):
        raise click.BadParameter(f"invalid filters provided: {set(filt.keys()) - valid_filters}. Valid:{valid_filters}")

    data = get_data(settings)
    df_queues = data.df_queues
    df_jobs = data.df_jobs
    df_vnodes = data.df_vnodes
    df_reservations = data.df_reservations
    enum_lookup = data.enum_lookup

    if "resvid" in filt:
        # take the data down a bit.
        subfilt = {"resvid": filt["resvid"]}
        df_reservations = df_reservations.loc[(df_reservations[list(subfilt)] == pd.Series(subfilt)).all(axis=1)]

    df_reservations = split_reservations_into_nodes(df_reservations, enum_lookup)
    df_avail = get_avail(df_queues, df_jobs, df_vnodes, df_reservations, enum_lookup)
    vnodes_all = set(df_vnodes["vnode"].unique())

    df_queues, df_vnodes = extract_queue_vnode(df_queues, df_vnodes)

    qlstofdct = df_to_lstofdct(df_queues[["queue", "resc_sel", "resc_sel_bitmask"]])
    vlstofdct = df_to_lstofdct(df_vnodes[["vnode", "resc_sel", "resc_sel_bitmask", "node_state"]])
    queue_vnodes_agg_lookup = {}
    available_vnodes = set(df_avail[df_avail["isavailable"] == 1]["vnode"].to_list())
    unavailable_vnodes = set(df_avail[df_avail["isavailable"] == 0]["vnode"].to_list())

    # if queue has nodes, available can only come from it.
    queues_with_nodes = defaultdict(set)
    df_reservations = df_reservations[df_reservations["reserve_state"] == "RESV_STATE_RUNNING"]

    if df_reservations.empty:
        rlstofdct = []
    else:
        rlstofdct = df_to_lstofdct(df_reservations[["resvid", "vnode"]])
    for dct in rlstofdct:
        queues_with_nodes[dct["resvid"]].add(dct["vnode"])

    for qdct in qlstofdct:
        queue = qdct["queue"]
        if queue in queues_with_nodes:
            available_vnodes_resv = set(
                df_avail[((df_avail["isavailable_resv"] == 1) & (df_avail["queue"] == queue))]["vnode"].to_list()
            )
            queue_avail_nodes = available_vnodes_resv
            queue_possible_nodes = queues_with_nodes[queue]
        else:
            queue_avail_nodes = available_vnodes
            queue_possible_nodes = vnodes_all
        queue_vnodes_lookup = {
            "possible": 0,
            "available": 0,
            "has_down": 0,
            "has_offline": 0,
            "has_free": 0,
            "has_state-unknown": 0,
            "has_Stale": 0,
            "vnodes": set(),
        }
        for vdct in vlstofdct:
            vnode = vdct["vnode"]
            if vnode not in queue_possible_nodes:
                continue
            qbit = qdct["resc_sel_bitmask"]
            vbit = vdct["resc_sel_bitmask"]
            subset = not (qbit & np.logical_not(vbit)).any()
            if subset:
                queue_vnodes_lookup["vnodes"].add(vnode)
                queue_vnodes_lookup["possible"] += 1
                if "down" in vdct["node_state"]:
                    queue_vnodes_lookup["has_down"] += 1
                if "offline" in vdct["node_state"]:
                    queue_vnodes_lookup["has_offline"] += 1
                if "free" in vdct["node_state"]:
                    queue_vnodes_lookup["has_free"] += 1
                if "state-unknown" in vdct["node_state"]:
                    queue_vnodes_lookup["has_state-unknown"] += 1
                if "Stale" in vdct["node_state"]:
                    queue_vnodes_lookup["has_Stale"] += 1
        queue_vnodes_available = queue_vnodes_lookup["vnodes"].intersection(queue_avail_nodes)
        queue_vnodes_lookup["available"] = len(queue_vnodes_available)
        queue_vnodes_agg_lookup[queue] = queue_vnodes_lookup

    lstofdct = []
    for queue, agg_dct in queue_vnodes_agg_lookup.items():
        new_agg_dct = dict((k, v) for k, v in agg_dct.items() if type(v) == int)
        new_agg_dct["queue"] = queue
        lstofdct.append(new_agg_dct)
    df = pd.DataFrame(lstofdct)

    df = df_apply_options(df, settings)
    df_display(df, settings)


@cli.command()
@pass_settings
def job(settings: Settings):
    filt = settings.filt
    data = get_data(settings)
    df_jobs = data.df_jobs
    if not settings.header:
        settings.header = [
            "jobid",
            "queue",
            "job_state",
            "substate",
            "stime",
            # 'Resource_List.nodect',
            "nodect",
            # 'Resource_List.walltime', 'resources_used.walltime',
            "walltime",
            "runtime",
            "run_count",
            "real_runtime",
            "job_stuck",
            "vnode",
        ]
    df_jobs["substate"] = df_jobs.get("substate", default=pd.Series(dtype=object)).apply(
        lambda ser: "".join(state_to_strings(data.enum_lookup, "LONG__REVERSE_JOB_SUBSTATE", ser)).replace("JOB_SUBSTATE_", "")
    )
    valid_filters = set(df_jobs.columns.to_list())
    if filt and not set(filt.keys()).issubset(valid_filters):
        raise click.BadParameter(f"invalid filters provided: {set(filt.keys()) - valid_filters}. Valid:{valid_filters}")
    if not df_jobs.empty:
        df_jobs["job_state"] = df_jobs["job_state"].apply(job_state_to_word)
        # this makes lots of rows.  Maybe have a split option for this data.
        # df_jobs['exec_vnode'] = df_jobs['exec_vnode'].apply(wrap_cast_to_vnode_list)
        df_jobs["exec_vnode"] = df_jobs.get("exec_vnode", default=pd.Series(dtype=object)).apply(wrap_cast_to_vnode_list)
        df_jobs = split_column_lst_into_series(df_jobs, "exec_vnode", "vnode")
    df = df_jobs
    df = df_apply_options(df, settings)
    df_display(df, settings)


@cli.command()
@click.option("--option", help="special job query option --option stuck")
@pass_settings
def job_special(settings: Settings, option):
    """return the jobs that are in R or E that are not progressing/stime is not changing and is 10 minute different than now."""
    valid_options = ["stuck"]
    if option not in valid_options:
        raise click.BadOptionUsage(f"Invalid option {option} selected.  Valid:{','.join(valid_options)}")

    data = get_data(settings)

    if option == "stuck":
        df_jobs = data.df_jobs
        metadata = data.metadata
        if not settings.header:
            settings.header = [
                "jobid",
                "queue",
                "job_state",
                "substate",
                "stime",
                # 'Resource_List.nodect',
                "nodect",
                # 'Resource_List.walltime', 'resources_used.walltime',
                "walltime",
                "runtime",
                "run_count",
                "real_runtime",
                "job_stuck",
            ]

        df = df_jobs
        if not df.empty:
            df["substate"] = df.get("substate", default=pd.Series(dtype=object)).apply(
                lambda ser: "".join(state_to_strings(data.enum_lookup, "LONG__REVERSE_JOB_SUBSTATE", ser)).replace(
                    "JOB_SUBSTATE_", ""
                )
            )
            timestamp = datetime.datetime.strptime(metadata["timestamp"], LOG_REGEX.LOG_FORMAT_ISO8601)
            df["stime"] = df.get("stime", default=pd.Series(dtype=object)).apply(
                lambda ts: epoch_to_datetime(ts) if pd.notna(ts) else np.nan
            )
            df["resources_used.walltime_seconds"] = df.get("resources_used.walltime", pd.Series(dtype=object)).apply(
                lambda ts: hms_to_seconds(ts) if ts else np.nan
            )
            df["time_query"] = timestamp

        if not df.empty:

            def func_real_runtime(ser):
                if "resources_used.walltime" in ser and pd.notna(ser["resources_used.walltime"]):
                    try:
                        result = (ser["time_query"] - ser["stime"]).total_seconds()
                    except:
                        result = np.nan
                else:
                    result = np.nan
                return result

            def func_stuck(ser):
                epsilon = 10 * 60
                real_delta = abs(ser["resources_used.walltime_seconds"] - ser["real_runtime"])
                return 1 if real_delta > epsilon else 0

            df["real_runtime"] = df.apply(func_real_runtime, axis=1)
            df["job_stuck"] = df.apply(func_stuck, axis=1)
            df = df.rename(
                columns={
                    "resources_used.walltime": "runtime",
                    "Resource_List.walltime": "walltime",
                    "Resource_List.nodect": "nodect",
                }
            )
            df = df[((df["job_state"].isin(["E", "R"]) & (df["job_stuck"] == 1)))]

            df = df_apply_options(df, settings)
        df_display(df, settings)


if __name__ == "__main__":
    cli()

"""
# This is a bunch of commands to create some scenerios to test with. 
pbs_jobid=$(/opt/pbs/bin/qsub -h -- /bin/sleep 1000);echo ${pbs_jobid}; qalter -o job_${pbs_jobid}.stdout -e job_${pbs_jobid}.stderr ${pbs_jobid}; qrls ${pbs_jobid};
qmgr -c "set node pdw-c03 state=offline"
pbs_jobid2=$(/opt/pbs/bin/qsub -h -l select=4:ncpus=4,place=scatter -- /bin/sleep 1000);echo ${pbs_jobid2}; qalter -o job_${pbs_jobid2}.stdout -e job_${pbs_jobid2}.stderr ${pbs_jobid2}; qrls ${pbs_jobid2};
pbs_resv_now=$(pbs_rsub -R $(date '+%H%M' -d "1 min") -D 01:00:00 -l select=1:ncpus=16|cut -d "." -f 1); echo $pbs_resv_now
pbs_resv_ltr=$(pbs_rsub -R $(date '+%H%M' -d "1 day") -D 01:00:00 -l select=1:ncpus=16|cut -d "." -f 1); echo $pbs_resv_ltr
qmgr -c "set server reserve_retry_time=60"
# stop pbs on pdw-c03, which is the reservation node of $pbs_resv_now
# the reservation will hop to somewhere

pbs_rdel $pbs_resv_now
pbs_rdel $pbs_resv_ltr
qdel $pbs_jobid
qdel $pbs_jobid2
qmgr -c "set node pdw-c03 state=free"

pbs_resv_main=$(pbs_rsub -R $(date '+%H%M' -d "1 min") -D 01:00:00 --hosts pdw-c03|cut -d "." -f 1); echo $pbs_resv_main
"""
