# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

from typing import Tuple, List

import pytest
from datetime import datetime
from textwrap import dedent
from pprint import pprint, pformat

from dateutil.parser import parse as date_parse

from PBS_Utils.pbs_accounting import (
    BadLineError,
    get_time,
    format_key_value_pairs,
    create_pbs_log_line,
    seconds_to_hms,
    enhance_pbs_record,
    Parse_Error_PBS,
    FieldHandlers,
)
from PBS_Utils.pbs_accounting import LOG_REGEX, RecType
from PBS_Utils.pbs_accounting import fix_newlines_extract_time
from PBS_Utils.pbs_accounting import hms_to_seconds
from PBS_Utils.pbs_accounting import parse_key_value_pairs
from PBS_Utils.pbs_accounting import parse_pbs_log_line
from PBS_Utils.pbs_accounting import pbs_line_extract_time
from PBS_Utils.pbs_accounting import split_record, func_squote, func_dquote
from PBS_Utils.pbs_accounting import cast_to_place_dict, join_place_dict
from PBS_Utils.pbs_accounting import cast_to_select_list, join_select_list
from PBS_Utils.pbs_accounting import create_pbs_log_line_from_record, create_pbs_record


def test_LOG_REGEX():
    assert LOG_REGEX.LOG_FORMAT_PBS
    assert LOG_REGEX.LOG_FORMAT_PBS_GROUPED
    assert LOG_REGEX.LOG_FORMAT_DATA_SEPERATOR_PBS == ";"
    assert LOG_REGEX.PATTERN_KEY_VALUE_PAIRS


def test_RecType():
    """this tests to make sure the record types have associated entries in the RecType.dct dictionary."""
    for key, value in RecType.dct.items():
        if key is None:
            continue
        assert getattr(RecType, key, None) is not None
        assert value["record"] == key

    missing = []
    for key, value in RecType.__dict__.items():
        if (not key.startswith("_")) and (type(value) == str):
            try:
                RecType.dct[key]
            except KeyError:
                missing.append(key)
    assert len(missing) == 0, pformat(missing)


def input_pbs_line_extract_time(ids=False) -> List[Tuple[str, Tuple[datetime, str]]]:
    deck = [
        # vvv line          vvv                          (  correct log time                    correct line)
        (
            "03/24/2011 13:50:42;D;362971;requester=someuser99",
            (date_parse("2011-03-24 13:50:42"), "D;362971;requester=someuser99"),
        ),
        (
            "11/18/2014 17:29:02;Q;248;Resource_List.ncpus=512 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "account=someaccount32 queue=default resource=default user=userd",
            (
                date_parse("2014-11-18 17:29:02"),
                "Q;248;Resource_List.ncpus=512 Resource_List.nodect=512 Resource_List.walltime=0:10:00 account=someaccount32 queue=default "
                "resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:29:16;D;248;Resource_List.ncpus=512 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "requester=userd resource=default user=userd",
            (
                date_parse("2014-11-18 17:29:16"),
                "D;248;Resource_List.ncpus=512 Resource_List.nodect=512 Resource_List.walltime=0:10:00 requester=userd "
                "resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:29:49;Q;249;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "account=someaccount32 queue=default resource=default user=userd",
            (
                date_parse("2014-11-18 17:29:49"),
                "Q;249;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 account=someaccount32 "
                "queue=default resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:30:13;Q;249;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "account=someaccount32 queue=default resource=default user=userd",
            (
                date_parse("2014-11-18 17:30:13"),
                "Q;249;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 account=someaccount32 "
                "queue=default resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:39:26;S;243;Resource_List.ncpus=512 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "account=someaccount32 args= ctime=1416269680.72 cwd=/fs/somedirectoryobj43 etime=1416269680.72 exe=/bin/date "
            "exec_host=IRN-48400-7B731-512 group=unknown jobname=N/A mode=c1 qtime=1416269680.72 queue=default resource=default "
            "session=unknown start=1416353966.65 user=userd",
            (
                date_parse("2014-11-18 17:39:26"),
                "S;243;Resource_List.ncpus=512 Resource_List.nodect=512 Resource_List.walltime=0:10:00 account=someaccount32 args= "
                "ctime=1416269680.72 cwd=/fs/somedirectoryobj43 etime=1416269680.72 exe=/bin/date exec_host=IRN-48400-7B731-512 "
                "group=unknown jobname=N/A mode=c1 qtime=1416269680.72 queue=default resource=default session=unknown "
                "start=1416353966.65 user=userd",
            ),
        ),
        (
            "11/18/2014 17:39:44;S;249;Resource_List.ncpus=65536 Resource_List.nodect=1024 Resource_List.walltime=0:10:00 "
            "account=someaccount32 args= ctime=1416353389.79 cwd=/fs/somedirectoryobj43 etime=1416353425.79 exe=/bin/date "
            "exec_host=IRN-48480-7B7F1-1-1024 group=unknown jobname=N/A mode=c64 qtime=1416353413.72 queue=default "
            "resource=default session=unknown start=1416353984.31 user=userd",
            (
                date_parse("2014-11-18 17:39:44"),
                "S;249;Resource_List.ncpus=65536 Resource_List.nodect=1024 Resource_List.walltime=0:10:00 account=someaccount32 args= "
                "ctime=1416353389.79 cwd=/fs/somedirectoryobj43 etime=1416353425.79 exe=/bin/date exec_host=IRN-48480-7B7F1-1-1024 "
                "group=unknown jobname=N/A mode=c64 qtime=1416353413.72 queue=default resource=default session=unknown "
                "start=1416353984.31 user=userd",
            ),
        ),
        (
            "11/18/2014 17:40:02;Q;250;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:05:00 "
            "account=someaccount32 queue=default resource=default user=userd",
            (
                date_parse("2014-11-18 17:40:02"),
                "Q;250;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:05:00 account=someaccount32 "
                "queue=default resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:40:30;S;250;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:05:00 "
            "account=someaccount32 args=--timeout,5000000 ctime=1416354002.41 cwd=/fs/somedirectoryobj43 etime=1416354016.26 exe=/bin/date "
            "exec_host=IRN-48440-7B771-512 group=unknown jobname=N/A mode=c64 qtime=1416354002.41 queue=default resource=default "
            "session=unknown start=1416354030.13 user=userd",
            (
                date_parse("2014-11-18 17:40:30"),
                "S;250;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:05:00 account=someaccount32 "
                "args=--timeout,5000000 ctime=1416354002.41 cwd=/fs/somedirectoryobj43 etime=1416354016.26 exe=/bin/date "
                "exec_host=IRN-48440-7B771-512 group=unknown jobname=N/A mode=c64 qtime=1416354002.41 queue=default "
                "resource=default session=unknown start=1416354030.13 user=userd",
            ),
        ),
        (
            "11/18/2014 17:41:33;E;243;Exit_status=0 Resource_List.ncpus=512 Resource_List.nodect=512 "
            "Resource_List.walltime=0:10:00 account=someaccount32 args= ctime=1416269680.72 cwd=/fs/somedirectoryobj43 end=1416354093.12 "
            "etime=1416269680.72 exe=/bin/date exec_host=IRN-48400-7B731-512 group=unknown jobname=N/A mode=c1 "
            "qtime=1416269680.72 queue=default resource=default resources_used.location=IRN-48400-7B731-512 "
            "resources_used.nodect=512 resources_used.walltime=0:02:06 session=unknown start=1416353966.65 user=userd",
            (
                date_parse("2014-11-18 17:41:33"),
                "E;243;Exit_status=0 Resource_List.ncpus=512 Resource_List.nodect=512 Resource_List.walltime=0:10:00 account=someaccount32 "
                "args= ctime=1416269680.72 cwd=/fs/somedirectoryobj43 end=1416354093.12 etime=1416269680.72 exe=/bin/date "
                "exec_host=IRN-48400-7B731-512 group=unknown jobname=N/A mode=c1 qtime=1416269680.72 queue=default "
                "resource=default resources_used.location=IRN-48400-7B731-512 resources_used.nodect=512 "
                "resources_used.walltime=0:02:06 session=unknown start=1416353966.65 user=userd",
            ),
        ),
        (
            "11/18/2014 17:42:14;E;249;Exit_status=0 Resource_List.ncpus=65536 Resource_List.nodect=1024 "
            "Resource_List.walltime=0:10:00 account=someaccount32 args= ctime=1416353389.79 cwd=/fs/somedirectoryobj43 end=1416354134.03 "
            "etime=1416353425.79 exe=/bin/date exec_host=IRN-48480-7B7F1-1-1024 group=unknown jobname=N/A mode=c64 "
            "qtime=1416353413.72 queue=default resource=default resources_used.location=IRN-48480-7B7F1-1-1024 "
            "resources_used.nodect=1024 resources_used.walltime=0:02:29 session=unknown start=1416353984.31 user=userd",
            (
                date_parse("2014-11-18 17:42:14"),
                "E;249;Exit_status=0 Resource_List.ncpus=65536 Resource_List.nodect=1024 Resource_List.walltime=0:10:00 "
                "account=someaccount32 args= ctime=1416353389.79 cwd=/fs/somedirectoryobj43 end=1416354134.03 etime=1416353425.79 exe=/bin/date "
                "exec_host=IRN-48480-7B7F1-1-1024 group=unknown jobname=N/A mode=c64 qtime=1416353413.72 queue=default "
                "resource=default resources_used.location=IRN-48480-7B7F1-1-1024 resources_used.nodect=1024 "
                "resources_used.walltime=0:02:29 session=unknown start=1416353984.31 user=userd",
            ),
        ),
        (
            "11/18/2014 17:45:38;A;250;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:05:00 "
            "resource=default user=userd",
            (
                date_parse("2014-11-18 17:45:38"),
                "A;250;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:05:00 resource=default "
                "user=userd",
            ),
        ),
        (
            "11/18/2014 17:45:58;E;250;Exit_status=143 Resource_List.ncpus=32768 Resource_List.nodect=512 "
            "Resource_List.walltime=0:05:00 account=someaccount32 args=--timeout,5000000 ctime=1416354002.41 cwd=/fs/somedirectoryobj43 "
            "end=1416354358.67 etime=1416354016.26 exe=/bin/date exec_host=IRN-48440-7B771-512 group=unknown jobname=N/A mode=c64 "
            "qtime=1416354002.41 queue=default resource=default resources_used.location=IRN-48440-7B771-512 "
            "resources_used.nodect=512 resources_used.walltime=0:05:28 session=unknown start=1416354030.13 user=userd",
            (
                date_parse("2014-11-18 17:45:58"),
                "E;250;Exit_status=143 Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:05:00 "
                "account=someaccount32 args=--timeout,5000000 ctime=1416354002.41 cwd=/fs/somedirectoryobj43 end=1416354358.67 "
                "etime=1416354016.26 exe=/bin/date exec_host=IRN-48440-7B771-512 group=unknown jobname=N/A mode=c64 "
                "qtime=1416354002.41 queue=default resource=default resources_used.location=IRN-48440-7B771-512 "
                "resources_used.nodect=512 resources_used.walltime=0:05:28 session=unknown start=1416354030.13 user=userd",
            ),
        ),
        (
            "11/18/2014 17:46:12;Q;251;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "account=someaccount32 queue=default resource=default user=userd",
            (
                date_parse("2014-11-18 17:46:12"),
                "Q;251;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 account=someaccount32 "
                "queue=default resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:46:27;D;251;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "requester=userd resource=default user=userd",
            (
                date_parse("2014-11-18 17:46:27"),
                "D;251;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 requester=userd "
                "resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:46:35;Q;252;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "account=someaccount32 queue=default resource=default user=userd",
            (
                date_parse("2014-11-18 17:46:35"),
                "Q;252;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 account=someaccount32 "
                "queue=default resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:46:52;S;252;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "account=someaccount32 args= ctime=1416354395.79 cwd=/fs/somedirectoryobj43 etime=1416354395.79 exe=/bin/date "
            "exec_host=IRN-48840-7BB71-512 group=unknown jobname=N/A mode=c64 qtime=1416354395.79 queue=default "
            "resource=default session=unknown start=1416354412.88 user=userd",
            (
                date_parse("2014-11-18 17:46:52"),
                "S;252;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 account=someaccount32 args= "
                "ctime=1416354395.79 cwd=/fs/somedirectoryobj43 etime=1416354395.79 exe=/bin/date exec_host=IRN-48840-7BB71-512 "
                "group=unknown jobname=N/A mode=c64 qtime=1416354395.79 queue=default resource=default session=unknown "
                "start=1416354412.88 user=userd",
            ),
        ),
        (
            "11/18/2014 17:47:26;D;252;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
            "requester=userd resource=default user=userd",
            (
                date_parse("2014-11-18 17:47:26"),
                "D;252;Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 requester=userd "
                "resource=default user=userd",
            ),
        ),
        (
            "11/18/2014 17:47:43;E;252;Exit_status=143 Resource_List.ncpus=32768 Resource_List.nodect=512 "
            "Resource_List.walltime=0:10:00 account=someaccount32 args= ctime=1416354395.79 cwd=/fs/somedirectoryobj43 end=1416354463.84 "
            "etime=1416354395.79 exe=/bin/date exec_host=IRN-48840-7BB71-512 group=unknown jobname=N/A mode=c64 "
            "qtime=1416354395.79 queue=default resource=default resources_used.location=IRN-48840-7BB71-512 "
            "resources_used.nodect=512 resources_used.walltime=0:00:50 session=unknown start=1416354412.88 user=userd",
            (
                date_parse("2014-11-18 17:47:43"),
                "E;252;Exit_status=143 Resource_List.ncpus=32768 Resource_List.nodect=512 Resource_List.walltime=0:10:00 "
                "account=someaccount32 args= ctime=1416354395.79 cwd=/fs/somedirectoryobj43 end=1416354463.84 etime=1416354395.79 exe=/bin/date "
                "exec_host=IRN-48840-7BB71-512 group=unknown jobname=N/A mode=c64 qtime=1416354395.79 queue=default "
                "resource=default resources_used.location=IRN-48840-7BB71-512 resources_used.nodect=512 "
                "resources_used.walltime=0:00:50 session=unknown start=1416354412.88 user=userd",
            ),
        ),
    ]
    if ids:
        deck = [f"{i:0>2}" for i, _ in enumerate(deck)]
    return deck


def input_parse_kv_pairs(ids=False):
    deck = [
        (
            "job record",
            "01/01/1970 01:01:01;Q;77777777;Resource_List.ncpus=8192 Resource_List.nodect=8192 Resource_List.walltime=1:16:00 "
            "args= ctime=1295298682.99 cwd=/fs/somedirectoryobjB etime=1295298682.99 "
            "exe=/fs/somedirectoryobjC exec_host=ANL-R30-R37-8192 group=unknown jobname=N/A "
            "mode=co qtime=1295298682.99 queue=default session=unknown start=1295298688.17 user=someuser99",
            {
                "exec_host": "ANL-R30-R37-8192",
                "exe": "/fs/somedirectoryobjC",
                "group": "unknown",
                "ctime": "1295298682.99",
                "Resource_List.ncpus": "8192",
                "qtime": "1295298682.99",
                "Resource_List.walltime": "1:16:00",
                "etime": "1295298682.99",
                "queue": "default",
                "start": "1295298688.17",
                "session": "unknown",
                "mode": "co",
                "Resource_List.nodect": "8192",
                "jobname": "N/A",
                "cwd": "/fs/somedirectoryobjB",
                "user": "someuser99",
                "args": "",
            },
            True,
        ),
        (
            "job record",
            "01/01/1970 01:01:01;Q;77777777;Resource_List.ncpus=8192 Resource_List.nodect=8192 Resource_List.walltime=1:16:00 "
            "args= ctime=1295298682.99 cwd=/fs/somedirectoryobjB etime=1295298682.99 "
            "exe=/fs/somedirectoryobjC exec_host=ANL-R30-R37-8192 group=unknown jobname=N/A "
            "mode=co qtime=1295298682.99 queue=default session=unknown start=1295298688.17 user=someuser99",
            {
                "exec_host": "ANL-R30-R37-8192",
                "exe": "/fs/somedirectoryobjC",
                "group": "unknown",
                "ctime": "1295298682.99",
                "Resource_List.ncpus": "8192",
                "qtime": "1295298682.99",
                "Resource_List.walltime": "1:16:00",
                "etime": "1295298682.99",
                "queue": "default",
                "start": "1295298688.17",
                "session": "unknown",
                "mode": "co",
                "Resource_List.nodect": "8192",
                "jobname": "N/A",
                "cwd": "/fs/somedirectoryobjB",
                "user": "someuser99",
                "args": "",
            },
            True,
        ),
        (
            "",
            "01/01/1970 01:01:01;Q;77777777;Resource_List.ncpus=8192 Resource_List.nodect=8192 Resource_List.walltime=1:16:00 "
            "args= ctime=1295298682.99 "
            "cwd=/fs/somedirectoryobjB etime=1295298682.99 "
            "exe=/fs/somedirectoryobjC exec_host=ANL-R30-R37-8192 group=unknown jobname=N/A "
            'mode=co qtime=1295298682.99 queue="default QUE" session=unknown start=1295298688.17 user=someuser99',
            {
                "exec_host": "ANL-R30-R37-8192",
                "exe": "/fs/somedirectoryobjC",
                "group": "unknown",
                "ctime": "1295298682.99",
                "Resource_List.ncpus": "8192",
                "qtime": "1295298682.99",
                "Resource_List.walltime": "1:16:00",
                "etime": "1295298682.99",
                "queue": "default QUE",
                "start": "1295298688.17",
                "session": "unknown",
                "mode": "co",
                "Resource_List.nodect": "8192",
                "jobname": "N/A",
                "cwd": "/fs/somedirectoryobjB",
                "user": "someuser99",
                "args": "",
            },
            True,
        ),
        (
            "job record",
            "01/01/1970 01:01:01;Q;77777777;Resource_List.ncpus=256 Resource_List.nodect=256 Resource_List.walltime=12:00:00 "
            "account=esp "
            "args=/fs/somedirectoryobj8,9,ADBC ctime=1358307943.61 "
            "cwd=/fs/somedirectoryobj9 etime=1358349957.24 "
            "exe=/fs/somedirectoryobjA exec_host=MIR-00CC0-33FF1-512 "
            "group=unknown jobname=somejobname mode=script qtime=1358307962.54 queue=somequeue78 session=unknown "
            "start=1358349989.3 user=someuser1",
            {
                "Resource_List.ncpus": "256",
                "Resource_List.nodect": "256",
                "Resource_List.walltime": "12:00:00",
                "account": "esp",
                "args": "/fs/somedirectoryobj8,9,ADBC",
                "ctime": "1358307943.61",
                "cwd": "/fs/somedirectoryobj9",
                "etime": "1358349957.24",
                "exe": "/fs/somedirectoryobjA",
                "exec_host": "MIR-00CC0-33FF1-512",
                "group": "unknown",
                "jobname": "somejobname",
                "mode": "script",
                "qtime": "1358307962.54",
                "queue": "somequeue78",
                "session": "unknown",
                "start": "1358349989.3",
                "user": "someuser1",
            },
            True,
        ),
        (
            "simple kv pairs",
            "01/01/1970 01:01:01;Q;77777777;key_a=pqr args=abc def ghi key_b=jkl key_c=mno",
            {
                "key_a": "pqr",
                "key_b": "jkl",
                "key_c": "mno",
                "args": "abc def ghi",
            },
            True,
        ),
        (
            "embedded structures",
            '01/01/1970 01:01:01;Q;77777777;key_a=pqr args=abc=234def:asdfghi:{"zxcv":asdf} key_b=jkl key_c=mno',
            {
                "key_a": "pqr",
                "key_b": "jkl",
                "key_c": "mno",
                "args": 'abc=234def:asdfghi:{"zxcv":asdf}',
            },
            True,
        ),
        (
            "unicode characters embedded",
            r"01/01/1970 01:01:01;Q;77777777;"
            r"Resource_List.ncpus=961 Resource_List.nodect=961 Resource_List.walltime=1:00:00 account=Performance args=["
            r"u'\u2013i', 'i_lsms', u'\u2013mode', '1d', u'\u2013size_lsms', '1024', u'\u2013num_lsms', '30', "
            r"u'\u2013num_steps', '600'] ctime=1380743728.06 "
            r"cwd=/fs/somedirectoryobjD etime=1380743728.06 "
            r"exe=/fs/somedirectoryobj92 exec_host=MIR-48880-7BBF1-1-1024 group=unknown "
            r"jobname=output mode=c1 qtime=1380743728.06 queue=somequeue09 session=unknown start=1380749287.08 user=someuser2",
            {
                "Resource_List.ncpus": "961",
                "Resource_List.nodect": "961",
                "Resource_List.walltime": "1:00:00",
                "account": "Performance",
                "args": "[u'\\u2013i', 'i_lsms', u'\\u2013mode', '1d', u'\\u2013size_lsms', '1024', u'\\u2013num_lsms', "
                "'30', u'\\u2013num_steps', '600']",
                "ctime": "1380743728.06",
                "cwd": "/fs/somedirectoryobjD",
                "etime": "1380743728.06",
                "exe": "/fs/somedirectoryobj92",
                "exec_host": "MIR-48880-7BBF1-1-1024",
                "group": "unknown",
                "jobname": "output",
                "mode": "c1",
                "qtime": "1380743728.06",
                "queue": "somequeue09",
                "session": "unknown",
                "start": "1380749287.08",
                "user": "someuser2",
            },
            False,
        ),  # BADDATA note this is false due to the embedded quotes and the lack of the ability to handle them.
        (
            """simple k/v pairs with spaces""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=8 abc=123 789 oiu=645""",
            {
                "Exit_status": "2",
                "tyu": "8",
                "abc": "123 789",
                "oiu": "645",
            },
            True,
        ),
        (
            """k/v pair with double quotes""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=8 abc="123 789" oiu=645""",
            {
                "Exit_status": "2",
                "tyu": "8",
                "abc": "123 789",
                "oiu": "645",
            },
            True,
        ),
        (
            """k/v pair with single quotes""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=8 abc='123 789' oiu=645""",
            {
                "Exit_status": "2",
                "tyu": "8",
                "abc": "123 789",
                "oiu": "645",
            },
            True,
        ),
        (
            """k/v pair with single quote in double quotes""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=8 abc="123'789" oiu=645""",
            {
                "Exit_status": "2",
                "tyu": "8",
                "abc": "123'789",
                "oiu": "645",
            },
            False,
        ),  # BADDATA note this is false due to the embedded quotes and the lack of the ability to handle them.
        (
            """k/v pair with double quote in single quotes""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=8 abc='123"789' oiu=645""",
            BadLineError,
            # removed the single quoting around the right side, 20231025, Eric
            # {
            #     'Exit_status': '2',
            #     'tyu': '8',
            #     'abc': '123"789',
            #     'oiu': '645',
            # },
            False,
        ),  # BADDATA note this is false due to the embedded quotes and the lack of the ability to handle them.
        (
            """k/v pair with escaped double quote in double quotes""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=8 abc="123\\"789" oiu=645""",
            {
                "Exit_status": "2",
                "tyu": "8",
                "abc": '123"789',
                "oiu": "645",
            },
            False,
        ),  # BADDATA note this is false due to the embedded quotes and the lack of the ability to handle them.
        (
            """k/v pair with escaped single quote in single quotes""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=8 abc='123\\'789' oiu=645""",
            {
                "Exit_status": "2",
                "tyu": "8",
                "abc": "123'789",
                "oiu": "645",
            },
            False,
        ),  # BADDATA note this is false due to the embedded quotes and the lack of the ability to handle them.
        (
            """k/v pairs in k/v pair""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 abc="qwe=123 tyu=456 ljk='789'" oiu=645""",
            {
                "Exit_status": "2",
                "abc": "qwe=123 tyu=456 ljk='789'",
                "oiu": "645",
            },
            False,
        ),  # BADDATA note this is false due to the embedded quotes and the lack of the ability to handle them.
        (
            """k/v pairs in k/v pair with duplicate tyu and escaped quotes""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=987 abc="qwe=\"123\" tyu=456 ljk='789'" oiu=645""",
            {
                "Exit_status": "2",
                "tyu": "987",
                "abc": "qwe=\"123\" tyu=456 ljk='789'",
                "oiu": "645",
            },
            False,
        ),  # BADDATA note this is false due to the embedded quotes and the lack of the ability to handle them.
        (
            """k/v pairs in k/v pair with inner k/v pairs values quoted""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=987 abc=qwe="123 759" someuser3 "tyu=456" 'ljk=789' bar oiu=645""",
            {
                "Exit_status": "2",
                "tyu": "987",
                "abc": 'qwe="123 759" someuser3 "tyu=456" \'ljk=789\' bar',
                "oiu": "645",
            },
            True,
        ),
        (
            """really complex k/v pairs""",
            "01/01/1970 01:01:01;Q;77777777;"
            "Exit_status=2 Resource_List.ncpus=1 Resource_List.nodect=1 Resource_List.walltime=0:30:00 account=Foo "
            "args=-c,\"env 	WORLD='hello' PATTERN='^$' 	make -C /fs/somedirectoryobj97 "
            "EPREFIX=/fs/somedirectoryobj90 	NODES=1 MAP_BY=node 	MPIRUN_ARGS='' 	"
            'logdir/my-prog-7-map-by-node.log" ctime=1603595566.28 cwd=/fs/somedirectoryobj96 '
            "end=1603595710.77 etime=1603595566.28 exe=/fs/somedirectoryobj98 exec_host=3834 group=unknown "
            "jobname=my.job mode=script qtime=1603595566.28 queue=somequeue1 resource=somemachine3 "
            "resources_used.location=3834 resources_used.nodect=1 resources_used.walltime=0:01:46 session=unknown "
            "start=1603595604.7 user=someuser3 ",
            {
                "Exit_status": "2",
                "Resource_List.ncpus": "1",
                "Resource_List.nodect": "1",
                "Resource_List.walltime": "0:30:00",
                "account": "Foo",
                "args": "-c,\"env \tWORLD='hello' PATTERN='^$' \tmake -C /fs/somedirectoryobj97 "
                "EPREFIX=/fs/somedirectoryobj90 \tNODES=1 MAP_BY=node \tMPIRUN_ARGS='' "
                '\tlogdir/my-prog-7-map-by-node.log"',
                "ctime": "1603595566.28",
                "cwd": "/fs/somedirectoryobj96",
                "end": "1603595710.77",
                "etime": "1603595566.28",
                "exe": "/fs/somedirectoryobj98",
                "exec_host": "3834",
                "group": "unknown",
                "jobname": "my.job",
                "mode": "script",
                "qtime": "1603595566.28",
                "queue": "somequeue1",
                "resource": "somemachine3",
                "resources_used.location": "3834",
                "resources_used.nodect": "1",
                "resources_used.walltime": "0:01:46",
                "session": "unknown",
                "start": "1603595604.7",
                "user": "someuser3",
            },
            True,
        ),
        (
            """k/v pairs with missing quote""",
            """01/01/1970 01:01:01;Q;77777777;Exit_status=2 tyu=987 abc='qwe=123 759 tyu=456 ljk=789' oiu='645""",
            BadLineError,
            True,
        ),
        (
            "duplicate Exit_status",
            "05/19/2022 21:33:28;R;244329.somehost;user=someuser432 group=Group project=someuser72 "
            "jobname=rfm_job queue=somequeue32 ctime=1652994845 qtime=1652994845 etime=1652994845 start=1652995653 "
            "exec_host=x3104c0s19b1n0/0*64+x3104c0s1b0n0/0*64+x3104c0s1b1n0/0*64+x3104c0s25b0n0/0*64+x3104c0s25b1n0/0*64"
            "+x3104c0s31b0n0/0*64+x3104c0s31b1n0/0*64+x3104c0s37b0n0/0*64+x3104c0s37b1n0/0*64+x3104c0s7b0n0/0*64+x3105c0s13b0n0/0"
            "*64+x3105c0s13b1n0/0*64+x3105c0s19b0n0/0*64+x3105c0s19b1n0/0*64+x3105c0s1b0n0/0*64+x3105c0s1b1n0/0*64+x3105c0s25b0n0"
            "/0*64+x3105c0s25b1n0/0*64+x3105c0s31b0n0/0*64+x3105c0s31b1n0/0*64+x3105c0s37b0n0/0*64+x3105c0s37b1n0/0*64"
            "+x3105c0s7b0n0/0*64+x3105c0s7b1n0/0*64+x3106c0s13b0n0/0*64+x3106c0s13b1n0/0*64+x3106c0s19b0n0/0*64+x3106c0s19b1n0/0"
            "*64+x3106c0s1b0n0/0*64+x3106c0s1b1n0/0*64+x3106c0s25b0n0/0*64+x3106c0s25b1n0/0*64+x3106c0s31b0n0/0*64+x3106c0s31b1n0"
            "/0*64+x3106c0s37b0n0/0*64+x3106c0s37b1n0/0*64+x3106c0s7b0n0/0*64+x3106c0s7b1n0/0*64+x3107c0s13b0n0/0*64"
            "+x3107c0s13b1n0/0*64+x3107c0s19b0n0/0*64+x3107c0s19b1n0/0*64+x3107c0s25b0n0/0*64+x3107c0s25b1n0/0*64+x3107c0s31b0n0"
            "/0*64+x3107c0s31b1n0/0*64+x3107c0s37b0n0/0*64+x3107c0s37b1n0/0*64+x3107c0s7b0n0/0*64+x3107c0s7b1n0/0*64"
            "+x3108c0s13b0n0/0*64+x3108c0s13b1n0/0*64+x3108c0s19b0n0/0*64+x3108c0s19b1n0/0*64+x3108c0s1b0n0/0*64+x3108c0s1b1n0/0"
            "*64+x3108c0s25b0n0/0*64+x3108c0s25b1n0/0*64+x3108c0s31b0n0/0*64+x3108c0s31b1n0/0*64+x3108c0s37b0n0/0*64"
            "+x3108c0s37b1n0/0*64+x3208c0s37b1n0/0*64+x3208c0s7b0n0/0*64 exec_vnode=(x3104c0s19b1n0:ncpus=64)+("
            "x3104c0s1b0n0:ncpus=64)+(x3104c0s1b1n0:ncpus=64)+(x3104c0s25b0n0:ncpus=64)+(x3104c0s25b1n0:ncpus=64)+("
            "x3104c0s31b0n0:ncpus=64)+(x3104c0s31b1n0:ncpus=64)+(x3104c0s37b0n0:ncpus=64)+(x3104c0s37b1n0:ncpus=64)+("
            "x3104c0s7b0n0:ncpus=64)+(x3105c0s13b0n0:ncpus=64)+(x3105c0s13b1n0:ncpus=64)+(x3105c0s19b0n0:ncpus=64)+("
            "x3105c0s19b1n0:ncpus=64)+(x3105c0s1b0n0:ncpus=64)+(x3105c0s1b1n0:ncpus=64)+(x3105c0s25b0n0:ncpus=64)+("
            "x3105c0s25b1n0:ncpus=64)+(x3105c0s31b0n0:ncpus=64)+(x3105c0s31b1n0:ncpus=64)+(x3105c0s37b0n0:ncpus=64)+("
            "x3105c0s37b1n0:ncpus=64)+(x3105c0s7b0n0:ncpus=64)+(x3105c0s7b1n0:ncpus=64)+(x3106c0s13b0n0:ncpus=64)+("
            "x3106c0s13b1n0:ncpus=64)+(x3106c0s19b0n0:ncpus=64)+(x3106c0s19b1n0:ncpus=64)+(x3106c0s1b0n0:ncpus=64)+("
            "x3106c0s1b1n0:ncpus=64)+(x3106c0s25b0n0:ncpus=64)+(x3106c0s25b1n0:ncpus=64)+(x3106c0s31b0n0:ncpus=64)+("
            "x3106c0s31b1n0:ncpus=64)+(x3106c0s37b0n0:ncpus=64)+(x3106c0s37b1n0:ncpus=64)+(x3106c0s7b0n0:ncpus=64)+("
            "x3106c0s7b1n0:ncpus=64)+(x3107c0s13b0n0:ncpus=64)+(x3107c0s13b1n0:ncpus=64)+(x3107c0s19b0n0:ncpus=64)+("
            "x3107c0s19b1n0:ncpus=64)+(x3107c0s25b0n0:ncpus=64)+(x3107c0s25b1n0:ncpus=64)+(x3107c0s31b0n0:ncpus=64)+("
            "x3107c0s31b1n0:ncpus=64)+(x3107c0s37b0n0:ncpus=64)+(x3107c0s37b1n0:ncpus=64)+(x3107c0s7b0n0:ncpus=64)+("
            "x3107c0s7b1n0:ncpus=64)+(x3108c0s13b0n0:ncpus=64)+(x3108c0s13b1n0:ncpus=64)+(x3108c0s19b0n0:ncpus=64)+("
            "x3108c0s19b1n0:ncpus=64)+(x3108c0s1b0n0:ncpus=64)+(x3108c0s1b1n0:ncpus=64)+(x3108c0s25b0n0:ncpus=64)+("
            "x3108c0s25b1n0:ncpus=64)+(x3108c0s31b0n0:ncpus=64)+(x3108c0s31b1n0:ncpus=64)+(x3108c0s37b0n0:ncpus=64)+("
            "x3108c0s37b1n0:ncpus=64)+(x3208c0s37b1n0:ncpus=64)+(x3208c0s7b0n0:ncpus=64) Resource_List.ncpus=4096 "
            "Resource_List.nodect=64 Resource_List.place=scatter Resource_List.select=64:system=somemachine4 "
            "Resource_List.system=somemachine4 Resource_List.walltime=00:15:00 session=28643 end=1652996008 Exit_status=0 "
            "resources_used.cpupercent=24565 resources_used.cput=24:56:06 resources_used.mem=238265672kb "
            "resources_used.ncpus=4096 resources_used.vmem=253929436kb resources_used.walltime=00:07:22 Exit_status=-20 "
            "eligible_time=00:00:00 run_count=2",
            {
                "Exit_status": "0",
                "Exit_list": "0,-20",
                "Resource_List.ncpus": "4096",
                "Resource_List.nodect": "64",
                "Resource_List.place": "scatter",
                "Resource_List.select": "64:system=somemachine4",
                "Resource_List.system": "somemachine4",
                "Resource_List.walltime": "00:15:00",
                "ctime": "1652994845",
                "eligible_time": "00:00:00",
                "end": "1652996008",
                "etime": "1652994845",
                "exec_host": "x3104c0s19b1n0/0*64+x3104c0s1b0n0/0*64+x3104c0s1b1n0/0*64+x3104c0s25b0n0/0*64+x3104c0s25b1n0/0*64"
                "+x3104c0s31b0n0/0*64+x3104c0s31b1n0/0*64+x3104c0s37b0n0/0*64+x3104c0s37b1n0/0*64+x3104c0s7b0n0/0*64"
                "+x3105c0s13b0n0/0*64+x3105c0s13b1n0/0*64+x3105c0s19b0n0/0*64+x3105c0s19b1n0/0*64+x3105c0s1b0n0/0*64"
                "+x3105c0s1b1n0/0*64+x3105c0s25b0n0/0*64+x3105c0s25b1n0/0*64+x3105c0s31b0n0/0*64+x3105c0s31b1n0/0*64"
                "+x3105c0s37b0n0/0*64+x3105c0s37b1n0/0*64+x3105c0s7b0n0/0*64+x3105c0s7b1n0/0*64+x3106c0s13b0n0/0*64"
                "+x3106c0s13b1n0/0*64+x3106c0s19b0n0/0*64+x3106c0s19b1n0/0*64+x3106c0s1b0n0/0*64+x3106c0s1b1n0/0*64"
                "+x3106c0s25b0n0/0*64+x3106c0s25b1n0/0*64+x3106c0s31b0n0/0*64+x3106c0s31b1n0/0*64+x3106c0s37b0n0/0*64"
                "+x3106c0s37b1n0/0*64+x3106c0s7b0n0/0*64+x3106c0s7b1n0/0*64+x3107c0s13b0n0/0*64+x3107c0s13b1n0/0*64"
                "+x3107c0s19b0n0/0*64+x3107c0s19b1n0/0*64+x3107c0s25b0n0/0*64+x3107c0s25b1n0/0*64+x3107c0s31b0n0/0*64"
                "+x3107c0s31b1n0/0*64+x3107c0s37b0n0/0*64+x3107c0s37b1n0/0*64+x3107c0s7b0n0/0*64+x3107c0s7b1n0/0*64"
                "+x3108c0s13b0n0/0*64+x3108c0s13b1n0/0*64+x3108c0s19b0n0/0*64+x3108c0s19b1n0/0*64+x3108c0s1b0n0/0*64"
                "+x3108c0s1b1n0/0*64+x3108c0s25b0n0/0*64+x3108c0s25b1n0/0*64+x3108c0s31b0n0/0*64+x3108c0s31b1n0/0*64"
                "+x3108c0s37b0n0/0*64+x3108c0s37b1n0/0*64+x3208c0s37b1n0/0*64+x3208c0s7b0n0/0*64",
                "exec_vnode": "(x3104c0s19b1n0:ncpus=64)+(x3104c0s1b0n0:ncpus=64)+(x3104c0s1b1n0:ncpus=64)+("
                "x3104c0s25b0n0:ncpus=64)+(x3104c0s25b1n0:ncpus=64)+(x3104c0s31b0n0:ncpus=64)+("
                "x3104c0s31b1n0:ncpus=64)+(x3104c0s37b0n0:ncpus=64)+(x3104c0s37b1n0:ncpus=64)+("
                "x3104c0s7b0n0:ncpus=64)+(x3105c0s13b0n0:ncpus=64)+(x3105c0s13b1n0:ncpus=64)+("
                "x3105c0s19b0n0:ncpus=64)+(x3105c0s19b1n0:ncpus=64)+(x3105c0s1b0n0:ncpus=64)+("
                "x3105c0s1b1n0:ncpus=64)+(x3105c0s25b0n0:ncpus=64)+(x3105c0s25b1n0:ncpus=64)+("
                "x3105c0s31b0n0:ncpus=64)+(x3105c0s31b1n0:ncpus=64)+(x3105c0s37b0n0:ncpus=64)+("
                "x3105c0s37b1n0:ncpus=64)+(x3105c0s7b0n0:ncpus=64)+(x3105c0s7b1n0:ncpus=64)+("
                "x3106c0s13b0n0:ncpus=64)+(x3106c0s13b1n0:ncpus=64)+(x3106c0s19b0n0:ncpus=64)+("
                "x3106c0s19b1n0:ncpus=64)+(x3106c0s1b0n0:ncpus=64)+(x3106c0s1b1n0:ncpus=64)+("
                "x3106c0s25b0n0:ncpus=64)+(x3106c0s25b1n0:ncpus=64)+(x3106c0s31b0n0:ncpus=64)+("
                "x3106c0s31b1n0:ncpus=64)+(x3106c0s37b0n0:ncpus=64)+(x3106c0s37b1n0:ncpus=64)+("
                "x3106c0s7b0n0:ncpus=64)+(x3106c0s7b1n0:ncpus=64)+(x3107c0s13b0n0:ncpus=64)+("
                "x3107c0s13b1n0:ncpus=64)+(x3107c0s19b0n0:ncpus=64)+(x3107c0s19b1n0:ncpus=64)+("
                "x3107c0s25b0n0:ncpus=64)+(x3107c0s25b1n0:ncpus=64)+(x3107c0s31b0n0:ncpus=64)+("
                "x3107c0s31b1n0:ncpus=64)+(x3107c0s37b0n0:ncpus=64)+(x3107c0s37b1n0:ncpus=64)+("
                "x3107c0s7b0n0:ncpus=64)+(x3107c0s7b1n0:ncpus=64)+(x3108c0s13b0n0:ncpus=64)+("
                "x3108c0s13b1n0:ncpus=64)+(x3108c0s19b0n0:ncpus=64)+(x3108c0s19b1n0:ncpus=64)+("
                "x3108c0s1b0n0:ncpus=64)+(x3108c0s1b1n0:ncpus=64)+(x3108c0s25b0n0:ncpus=64)+("
                "x3108c0s25b1n0:ncpus=64)+(x3108c0s31b0n0:ncpus=64)+(x3108c0s31b1n0:ncpus=64)+("
                "x3108c0s37b0n0:ncpus=64)+(x3108c0s37b1n0:ncpus=64)+(x3208c0s37b1n0:ncpus=64)+(x3208c0s7b0n0:ncpus=64)",
                "group": "Group",
                "jobname": "rfm_job",
                "project": "someuser72",
                "qtime": "1652994845",
                "queue": "somequeue32",
                "resources_used.cpupercent": "24565",
                "resources_used.cput": "24:56:06",
                "resources_used.mem": "238265672kb",
                "resources_used.ncpus": "4096",
                "resources_used.vmem": "253929436kb",
                "resources_used.walltime": "00:07:22",
                "run_count": "2",
                "session": "28643",
                "start": "1652995653",
                "user": "someuser432",
            },
            True,
        ),
        (
            "malformed command line leakage: Resources_List.-I=",
            """02/23/2024 22:33:55;Q;621652.somehost;user=someuser92 group=somegroup21
            account="someproject" project=someproject jobname=hello_world.sh queue=somequeue ctime=1708727635 qtime=1708727635
            etime=0 Resource_List.allow_account_check_failure=True Resource_List.allow_negative_allocation=True
            Resource_List.award_category=someword34 Resource_List.award_type=someword234 Resource_List.burn_ratio=0
            Resource_List.current_allocation=-280471872 Resource_List.hbm_mode=flat
            Resource_List.hpcm_image=compute_test_20240120T003017_0051f5d Resource_List.ncpus=208 Resource_List.ni_resource=aurora
            Resource_List.nodect=1 Resource_List.overburn=False Resource_List.place=scatter Resource_List.project_priority=2
            Resource_List.route_backfill=True Resource_List.select=1 Resource_List.snc_mode=quad Resource_List.total_allocation=0
            Resource_List.walltime=00:30:00 Resource_List.-I=""",
            {
                "Resource_List.-I": "",
                "Resource_List.allow_account_check_failure": "True",
                "Resource_List.allow_negative_allocation": "True",
                "Resource_List.award_category": "someword34",
                "Resource_List.award_type": "someword234",
                "Resource_List.burn_ratio": "0",
                "Resource_List.current_allocation": "-280471872",
                "Resource_List.hbm_mode": "flat",
                "Resource_List.hpcm_image": "compute_test_20240120T003017_0051f5d",
                "Resource_List.ncpus": "208",
                "Resource_List.ni_resource": "aurora",
                "Resource_List.nodect": "1",
                "Resource_List.overburn": "False",
                "Resource_List.place": "scatter",
                "Resource_List.project_priority": "2",
                "Resource_List.route_backfill": "True",
                "Resource_List.select": "1",
                "Resource_List.snc_mode": "quad",
                "Resource_List.total_allocation": "0",
                "Resource_List.walltime": "00:30:00",
                "account": "someproject",
                "ctime": "1708727635",
                "etime": "0",
                "group": "somegroup21",
                "jobname": "hello_world.sh",
                "project": "someproject",
                "qtime": "1708727635",
                "queue": "somequeue",
                "user": "someuser92",
            },
            True,
        ),
        (
            "'-' in Resource name: Resources_List.foo-bar",
            """02/23/2024 22:33:55;Q;621652.somehost;user=someuser92 group=somegroup21
            account="someproject" project=someproject jobname=hello_world.sh queue=somequeue ctime=1708727635 qtime=1708727635
            etime=0 Resource_List.allow_account_check_failure=True Resource_List.allow_negative_allocation=True
            Resource_List.award_category=someword34 Resource_List.award_type=someword234 Resource_List.burn_ratio=0
            Resource_List.current_allocation=-280471872 Resource_List.hbm_mode=flat
            Resource_List.hpcm_image=compute_test_20240120T003017_0051f5d Resource_List.ncpus=208 Resource_List.ni_resource=aurora
            Resource_List.nodect=1 Resource_List.overburn=False Resource_List.place=scatter Resource_List.project_priority=2
            Resource_List.route_backfill=True Resource_List.select=1 Resource_List.snc_mode=quad Resource_List.total_allocation=0
            Resource_List.walltime=00:30:00 Resource_List.foo-bar=baz""",
            {
                "Resource_List.foo-bar": "baz",
                "Resource_List.allow_account_check_failure": "True",
                "Resource_List.allow_negative_allocation": "True",
                "Resource_List.award_category": "someword34",
                "Resource_List.award_type": "someword234",
                "Resource_List.burn_ratio": "0",
                "Resource_List.current_allocation": "-280471872",
                "Resource_List.hbm_mode": "flat",
                "Resource_List.hpcm_image": "compute_test_20240120T003017_0051f5d",
                "Resource_List.ncpus": "208",
                "Resource_List.ni_resource": "aurora",
                "Resource_List.nodect": "1",
                "Resource_List.overburn": "False",
                "Resource_List.place": "scatter",
                "Resource_List.project_priority": "2",
                "Resource_List.route_backfill": "True",
                "Resource_List.select": "1",
                "Resource_List.snc_mode": "quad",
                "Resource_List.total_allocation": "0",
                "Resource_List.walltime": "00:30:00",
                "account": "someproject",
                "ctime": "1708727635",
                "etime": "0",
                "group": "somegroup21",
                "jobname": "hello_world.sh",
                "project": "someproject",
                "qtime": "1708727635",
                "queue": "somequeue",
                "user": "someuser92",
            },
            True,
        ),
    ]
    if ids:
        deck = [f"{i:0>2}" for i, _ in enumerate(deck)]
    return deck


def input_parse_pbs_log_line(ids=False):
    deck = [
        (
            "",
            "01/17/2011 15:11:22;Q;20;queue=default",
            {
                "event_type": "JOB",
                "sub_identifier": "0",
                "action": "queue",
                "scheduler_timestamp": datetime(2011, 1, 17, 15, 11, 22),
                "identifier": "20",
                "queue": "default",
                "record": "Q",
            },
            True,
        ),
        (
            "",
            "01/17/2011 15:11:23;Q;21;queue=default",
            {
                "event_type": "JOB",
                "sub_identifier": "0",
                "action": "queue",
                "scheduler_timestamp": datetime(2011, 1, 17, 15, 11, 23),
                "identifier": "21",
                "queue": "default",
                "record": "Q",
            },
            True,
        ),
        (
            "",
            """01/17/2011 15:11:23;Q;22;queue=default""",
            {
                "event_type": "JOB",
                "sub_identifier": "0",
                "action": "queue",
                "scheduler_timestamp": datetime(2011, 1, 17, 15, 11, 23),
                "identifier": "22",
                "queue": "default",
                "record": "Q",
            },
            True,
        ),
        (
            "3",
            "01/17/2011 15:11:28;S;20;Resource_List.ncpus=8192 Resource_List.nodect=8192 Resource_List.walltime=1:16:00 args= "
            "ctime=1295298682.99 cwd=/fs/somedirectoryobjB etime=1295298682.99 "
            "exe=/fs/somedirectoryobjC exec_host=ANL-R30-R37-8192 group=unknown jobname=N/A "
            "mode=co qtime=1295298682.99 queue=default session=unknown start=1295298688.17 user=someuser99",
            {
                "event_type": "JOB",
                "sub_identifier": "0",
                "action": "start",
                "ctime": datetime(2011, 1, 17, 21, 11, 22),
                "cwd": "/fs/somedirectoryobjB",
                "scheduler_timestamp": datetime(2011, 1, 17, 15, 11, 28),
                "etime": datetime(2011, 1, 17, 21, 11, 22),
                "exe": "/fs/somedirectoryobjC",
                "exec_host": ["ANL-R30-R37-8192"],
                "group": "unknown",
                "identifier": "20",
                "jobname": "N/A",
                "mode": "co",
                "Resource_List.nodect": 8192,
                "qtime": datetime(2011, 1, 17, 21, 11, 22),
                "queue": "default",
                "record": "S",
                "start": datetime(2011, 1, 17, 21, 11, 28),
                "user": "someuser99",
                "raw_user_name": "someuser99",
                "Resource_List.walltime": 4560,
                "Resource_List.ncpus": 8192,
                "args": "",
                "session": "unknown",
            },
            True,
        ),
        (
            "4",
            "01/17/2011 15:11:39;S;21;Resource_List.ncpus=8192 Resource_List.nodect=8192 Resource_List.walltime=1:16:00 args= "
            "ctime=1295298683.38 cwd=/fs/somedirectoryobj77 "
            "etime=1295298683.38 exe=/fs/somedirectoryobj76 exec_host=ANL-R20-R27-8192 "
            "group=unknown jobname=N/A mode=co qtime=1295298683.38 queue=default session=unknown start=1295298699.15 "
            "user=someuser99",
            {
                "event_type": "JOB",
                "sub_identifier": "0",
                "action": "start",
                "ctime": datetime(2011, 1, 17, 21, 11, 23),
                "etime": datetime(2011, 1, 17, 21, 11, 23),
                "qtime": datetime(2011, 1, 17, 21, 11, 23),
                "cwd": "/fs/somedirectoryobj77",
                "scheduler_timestamp": datetime(2011, 1, 17, 15, 11, 39),
                "exe": "/fs/somedirectoryobj76",
                "exec_host": ["ANL-R20-R27-8192"],
                "group": "unknown",
                "identifier": "21",
                "jobname": "N/A",
                "mode": "co",
                "Resource_List.nodect": 8192,
                "queue": "default",
                "record": "S",
                "start": datetime(2011, 1, 17, 21, 11, 39),
                "user": "someuser99",
                "raw_user_name": "someuser99",
                "Resource_List.walltime": 4560,
                "Resource_List.ncpus": 8192,
                "args": "",
                "session": "unknown",
            },
            True,
        ),
        (
            "5",
            "01/17/2011 15:13:14;E;20;Exit_status=0 Resource_List.ncpus=8192 Resource_List.nodect=8192 "
            "Resource_List.walltime=1:16:00 args= ctime=1295298682.99 "
            "cwd=/fs/somedirectoryobjB end=1295298794.04 etime=1295298682.99 "
            "exe=/fs/somedirectoryobjC exec_host=ANL-R30-R37-8192 group=unknown jobname=N/A "
            "mode=co qtime=1295298682.99 queue=default resources_used.location=ANL-R30-R37-8192 "
            "resources_used.nodect=8192 resources_used.walltime=0:01:45 session=unknown start=1295298688.17 user=someuser99",
            {
                "event_type": "JOB",
                "sub_identifier": "0",
                "Exit_status": "0",
                "action": "end",
                "ctime": datetime(2011, 1, 17, 21, 11, 22),
                "cwd": "/fs/somedirectoryobjB",
                "scheduler_timestamp": datetime(2011, 1, 17, 15, 13, 14),
                "etime": datetime(2011, 1, 17, 21, 11, 22),
                "exe": "/fs/somedirectoryobjC",
                "exec_host": ["ANL-R30-R37-8192"],
                "group": "unknown",
                "identifier": "20",
                "jobname": "N/A",
                "mode": "co",
                "Resource_List.nodect": 8192,
                "resources_used.nodect": 8192,
                "qtime": datetime(2011, 1, 17, 21, 11, 22),
                "queue": "default",
                "record": "E",
                "resources_used.walltime": 105,
                "end": datetime(2011, 1, 17, 21, 13, 14),
                "start": datetime(2011, 1, 17, 21, 11, 28),
                "user": "someuser99",
                "raw_user_name": "someuser99",
                "Resource_List.walltime": 4560,
                "Resource_List.ncpus": 8192,
                "args": "",
                "session": "unknown",
            },
            True,
        ),
        (
            "6",
            "01/17/2011 15:15:05;E;21;Exit_status=0 Resource_List.ncpus=8192 Resource_List.nodect=8192 "
            "Resource_List.walltime=1:16:00 args= ctime=1295298683.38 "
            "cwd=/fs/somedirectoryobj77 end=1295298905.96 etime=1295298683.38 "
            "exe=/fs/somedirectoryobj76 exec_host=ANL-R20-R27-8192 group=unknown jobname=N/A "
            "mode=co qtime=1295298683.38 queue=default resources_used.location=ANL-R20-R27-8192 "
            "resources_used.nodect=8192 resources_used.walltime=0:03:26 session=unknown start=1295298699.15 user=someuser99",
            {
                "event_type": "JOB",
                "sub_identifier": "0",
                "Exit_status": "0",
                "action": "end",
                "ctime": datetime(2011, 1, 17, 21, 11, 23),
                "etime": datetime(2011, 1, 17, 21, 11, 23),
                "qtime": datetime(2011, 1, 17, 21, 11, 23),
                "cwd": "/fs/somedirectoryobj77",
                "scheduler_timestamp": datetime(2011, 1, 17, 15, 15, 5),
                "exe": "/fs/somedirectoryobj76",
                "exec_host": ["ANL-R20-R27-8192"],
                "group": "unknown",
                "identifier": "21",
                "jobname": "N/A",
                "mode": "co",
                "Resource_List.nodect": 8192,
                "resources_used.nodect": 8192,
                "queue": "default",
                "record": "E",
                "resources_used.walltime": 206,
                "end": datetime(2011, 1, 17, 21, 15, 5),
                "start": datetime(2011, 1, 17, 21, 11, 39),
                "user": "someuser99",
                "raw_user_name": "someuser99",
                "Resource_List.walltime": 4560,
                "Resource_List.ncpus": 8192,
                "args": "",
                "session": "unknown",
            },
            True,
        ),
        (
            "Job start: 1 exec_host, 2 exec_vnodes (qsub -l obj_type=gpu -l ncpus=5 -- /bin/sleep 10",
            "04/16/2021 14:17:55;S;107.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=STDIN queue=somequeue32 "
            "ctime=1618582675 qtime=1618582675 etime=1618582675 start=1618582675 exec_host=pdw-c1/0*0 exec_vnode=("
            "pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=1) Resource_List.ncpus=5 Resource_List.nodect=1 Resource_List.obj_type=gpu "
            "Resource_List.place=pack Resource_List.select=1:ncpus=5:obj_type=gpu resource_assigned.ncpus=6",
            {
                "action": "start",
                "record": "S",
                "event_type": "JOB",
                "identifier": "107.pdw-s1",
                "scheduler_timestamp": datetime(2021, 4, 16, 14, 17, 55),
                "sub_identifier": "0",
                "user": "someuser73",
                "raw_user_name": "someuser73",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "queue": "somequeue32",
                "jobname": "STDIN",
                "group": "someuser73",
                "Resource_List.nodect": 1,
                "Resource_List.ncpus": 5,
                # exec_host=pdw-c1/0*0
                "exec_host": ["pdw-c1"],
                "raw_exec_vnode": "(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=1)",
                "exec_vnode": ["pdw-c1-v00", "pdw-c1-v01"],
                "ctime": datetime(2021, 4, 16, 14, 17, 55),
                "qtime": datetime(2021, 4, 16, 14, 17, 55),
                "etime": datetime(2021, 4, 16, 14, 17, 55),
                "start": datetime(2021, 4, 16, 14, 17, 55),
                "Resource_List.place": {
                    "arrangement": "pack",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 1,
                        "ncpus": "5",
                        "obj_type": "gpu",
                    },
                ],
            },
            True,
        ),
        (
            "Job end: 1 exec_host, 2 exec_vnodes (qsub -l obj_type=gpu -l ncpus=5 -- /bin/sleep 10)",
            "04/16/2021 14:18:06;E;107.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=STDIN queue=somequeue32 "
            "ctime=1618582675 qtime=1618582675 etime=1618582675 start=1618582675 exec_host=pdw-c1/0*0 exec_vnode=("
            "pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=1) Resource_List.ncpus=5 Resource_List.nodect=1 Resource_List.obj_type=gpu "
            "Resource_List.place=pack Resource_List.select=1:ncpus=5:obj_type=gpu session=2015 end=1618582686 Exit_status=0 "
            "resources_used.cpupercent=0 resources_used.cput=00:00:00 resources_used.mem=0kb resources_used.ncpus=5 "
            "resources_used.vmem=0kb resources_used.walltime=00:00:10 run_count=1",
            {
                "action": "end",
                "record": "E",
                "event_type": "JOB",
                "identifier": "107.pdw-s1",
                "scheduler_timestamp": datetime(2021, 4, 16, 14, 18, 6),
                "sub_identifier": "0",
                "user": "someuser73",
                "raw_user_name": "someuser73",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "queue": "somequeue32",
                "jobname": "STDIN",
                "group": "someuser73",
                "Resource_List.nodect": 1,
                "Resource_List.ncpus": 5,
                # exec_host=pdw-c1/0*0
                "exec_host": ["pdw-c1"],
                "raw_exec_vnode": "(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=1)",
                "exec_vnode": ["pdw-c1-v00", "pdw-c1-v01"],
                "ctime": datetime(2021, 4, 16, 14, 17, 55),
                "qtime": datetime(2021, 4, 16, 14, 17, 55),
                "etime": datetime(2021, 4, 16, 14, 17, 55),
                "start": datetime(2021, 4, 16, 14, 17, 55),
                "end": datetime(2021, 4, 16, 14, 18, 6),
                "Exit_status": "0",
                "resources_used.walltime": 10,
                "run_count": 1,
                "Resource_List.place": {
                    "arrangement": "pack",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 1,
                        "ncpus": "5",
                        "obj_type": "gpu",
                    },
                ],
                "session": "2015",
            },
            True,
        ),
        (
            "Job start: 3 exec_hosts, 21 exec_vnodes (qsub -l select=3:ncpus=11:obj_type=gpu -- /bin/sleep 10)",
            "04/16/2021 18:59:22;S;117.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=STDIN queue=somequeue32 "
            "ctime=1618599562 qtime=1618599562 etime=1618599562 start=1618599562 exec_host=pdw-c1/0*0+pdw-c2/0*0+pdw-c3/0*0 "
            "exec_vnode=(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=2+pdw-c1-v02:ncpus=1+pdw-c1-v03:ncpus=1+pdw-c1-v04:ncpus=1+pdw-c1"
            "-v04-s0:ncpus=1+pdw-c1-v04-s1:ncpus=1)+("
            "pdw-c2-v00:ncpus=4+pdw-c2-v01:ncpus=2+pdw-c2-v02:ncpus=1+pdw-c2-v03:ncpus=1+pdw-c2-v04:ncpus=1+pdw-c2-v04-s0:ncpus=1"
            "+pdw-c2-v04-s1:ncpus=1)+(pdw-c3-v00:ncpus=4+pdw-c3-v01:ncpus=2+pdw-c3-v02:ncpus=1+pdw-c3-v03:ncpus=1+pdw-c3-v04"
            ":ncpus=1+pdw-c3-v04-s0:ncpus=1+pdw-c3-v04-s1:ncpus=1) Resource_List.ncpus=33 Resource_List.nodect=3 "
            "Resource_List.place=free Resource_List.select=3:ncpus=11:obj_type=gpu resource_assigned.ncpus=33",
            {
                "action": "start",
                "record": "S",
                "event_type": "JOB",
                "identifier": "117.pdw-s1",
                "scheduler_timestamp": datetime(2021, 4, 16, 18, 59, 22),
                "sub_identifier": "0",
                "user": "someuser73",
                "raw_user_name": "someuser73",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "queue": "somequeue32",
                "jobname": "STDIN",
                "group": "someuser73",
                "Resource_List.nodect": 3,
                "Resource_List.ncpus": 33,
                "exec_host": ["pdw-c1", "pdw-c2", "pdw-c3"],
                "exec_vnode": [
                    "pdw-c1-v00",
                    "pdw-c1-v01",
                    "pdw-c1-v02",
                    "pdw-c1-v03",
                    "pdw-c1-v04",
                    "pdw-c1-v04-s0",
                    "pdw-c1-v04-s1",
                    "pdw-c2-v00",
                    "pdw-c2-v01",
                    "pdw-c2-v02",
                    "pdw-c2-v03",
                    "pdw-c2-v04",
                    "pdw-c2-v04-s0",
                    "pdw-c2-v04-s1",
                    "pdw-c3-v00",
                    "pdw-c3-v01",
                    "pdw-c3-v02",
                    "pdw-c3-v03",
                    "pdw-c3-v04",
                    "pdw-c3-v04-s0",
                    "pdw-c3-v04-s1",
                ],
                "raw_exec_vnode": "(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=2+pdw-c1-v02:ncpus=1+pdw-c1-v03:ncpus=1+pdw-c1-v04:ncpus=1"
                "+pdw-c1-v04-s0:ncpus=1+pdw-c1-v04-s1:ncpus=1)+("
                "pdw-c2-v00:ncpus=4+pdw-c2-v01:ncpus=2+pdw-c2-v02:ncpus=1+pdw-c2-v03:ncpus=1+pdw-c2-v04:ncpus=1"
                "+pdw-c2-v04-s0:ncpus=1+pdw-c2-v04-s1:ncpus=1)+("
                "pdw-c3-v00:ncpus=4+pdw-c3-v01:ncpus=2+pdw-c3-v02:ncpus=1+pdw-c3-v03:ncpus=1+pdw-c3-v04:ncpus=1"
                "+pdw-c3-v04-s0:ncpus=1+pdw-c3-v04-s1:ncpus=1)",
                "ctime": datetime(2021, 4, 16, 18, 59, 22),
                "qtime": datetime(2021, 4, 16, 18, 59, 22),
                "etime": datetime(2021, 4, 16, 18, 59, 22),
                "start": datetime(2021, 4, 16, 18, 59, 22),
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 3,
                        "ncpus": "11",
                        "obj_type": "gpu",
                    },
                ],
            },
            True,
        ),
        (
            "Job end: 3 exec_hosts, 21 exec_vnodes (qsub -l select=3:ncpus=11:obj_type=gpu -- /bin/sleep 10)",
            "04/16/2021 18:59:34;E;117.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=STDIN queue=somequeue32 "
            "ctime=1618599562 qtime=1618599562 etime=1618599562 start=1618599562 exec_host=pdw-c1/0*0+pdw-c2/0*0+pdw-c3/0*0 "
            "exec_vnode=(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=2+pdw-c1-v02:ncpus=1+pdw-c1-v03:ncpus=1+pdw-c1-v04:ncpus=1+pdw-c1"
            "-v04-s0:ncpus=1+pdw-c1-v04-s1:ncpus=1)+("
            "pdw-c2-v00:ncpus=4+pdw-c2-v01:ncpus=2+pdw-c2-v02:ncpus=1+pdw-c2-v03:ncpus=1+pdw-c2-v04:ncpus=1+pdw-c2-v04-s0:ncpus=1"
            "+pdw-c2-v04-s1:ncpus=1)+(pdw-c3-v00:ncpus=4+pdw-c3-v01:ncpus=2+pdw-c3-v02:ncpus=1+pdw-c3-v03:ncpus=1+pdw-c3-v04"
            ":ncpus=1+pdw-c3-v04-s0:ncpus=1+pdw-c3-v04-s1:ncpus=1) Resource_List.ncpus=33 Resource_List.nodect=3 "
            "Resource_List.place=free Resource_List.select=3:ncpus=11:obj_type=gpu session=2063 end=1618599574 Exit_status=0 "
            "resources_used.cpupercent=0 resources_used.cput=00:00:00 resources_used.mem=864kb resources_used.ncpus=33 "
            "resources_used.vmem=12916kb resources_used.walltime=00:00:10 run_count=1",
            {
                "action": "end",
                "record": "E",
                "event_type": "JOB",
                "identifier": "117.pdw-s1",
                "scheduler_timestamp": datetime(2021, 4, 16, 18, 59, 34),
                "sub_identifier": "0",
                "user": "someuser73",
                "raw_user_name": "someuser73",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "queue": "somequeue32",
                "jobname": "STDIN",
                "group": "someuser73",
                "Resource_List.nodect": 3,
                "Resource_List.ncpus": 33,
                "exec_host": ["pdw-c1", "pdw-c2", "pdw-c3"],
                "exec_vnode": [
                    "pdw-c1-v00",
                    "pdw-c1-v01",
                    "pdw-c1-v02",
                    "pdw-c1-v03",
                    "pdw-c1-v04",
                    "pdw-c1-v04-s0",
                    "pdw-c1-v04-s1",
                    "pdw-c2-v00",
                    "pdw-c2-v01",
                    "pdw-c2-v02",
                    "pdw-c2-v03",
                    "pdw-c2-v04",
                    "pdw-c2-v04-s0",
                    "pdw-c2-v04-s1",
                    "pdw-c3-v00",
                    "pdw-c3-v01",
                    "pdw-c3-v02",
                    "pdw-c3-v03",
                    "pdw-c3-v04",
                    "pdw-c3-v04-s0",
                    "pdw-c3-v04-s1",
                ],
                "raw_exec_vnode": "(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=2+pdw-c1-v02:ncpus=1+pdw-c1-v03:ncpus=1+pdw-c1-v04:ncpus=1"
                "+pdw-c1-v04-s0:ncpus=1+pdw-c1-v04-s1:ncpus=1)+("
                "pdw-c2-v00:ncpus=4+pdw-c2-v01:ncpus=2+pdw-c2-v02:ncpus=1+pdw-c2-v03:ncpus=1+pdw-c2-v04:ncpus=1"
                "+pdw-c2-v04-s0:ncpus=1+pdw-c2-v04-s1:ncpus=1)+("
                "pdw-c3-v00:ncpus=4+pdw-c3-v01:ncpus=2+pdw-c3-v02:ncpus=1+pdw-c3-v03:ncpus=1+pdw-c3-v04:ncpus=1"
                "+pdw-c3-v04-s0:ncpus=1+pdw-c3-v04-s1:ncpus=1)",
                "ctime": datetime(2021, 4, 16, 18, 59, 22),
                "qtime": datetime(2021, 4, 16, 18, 59, 22),
                "etime": datetime(2021, 4, 16, 18, 59, 22),
                "start": datetime(2021, 4, 16, 18, 59, 22),
                "end": datetime(2021, 4, 16, 18, 59, 34),
                "Exit_status": "0",
                "resources_used.walltime": 10,
                "run_count": 1,
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 3,
                        "ncpus": "11",
                        "obj_type": "gpu",
                    },
                ],
                "session": "2063",
            },
            True,
        ),
        (
            "Job start: 3 exec_hosts, 9 exec_vnodes ( qsub -l select=3:ncpus=5:obj_type=gpu -- /bin/sleep 10)",
            "04/18/2021 19:24:23;S;119.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=STDIN queue=somequeue32 "
            "ctime=1618773863 qtime=1618773863 etime=1618773863 start=1618773863 exec_host=pdw-c1/0*0+pdw-c1/1*0+pdw-c2/0*0 "
            "exec_vnode=(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=1)+("
            "pdw-c1-v01:ncpus=1+pdw-c1-v02:ncpus=1+pdw-c1-v03:ncpus=1+pdw-c1-v04:ncpus=1+pdw-c1-v04-s0:ncpus=1)+("
            "pdw-c2-v00:ncpus=4+pdw-c2-v01:ncpus=1) Resource_List.ncpus=15 Resource_List.nodect=3 Resource_List.place=free "
            "Resource_List.select=3:ncpus=5:obj_type=gpu resource_assigned.ncpus=16",
            {
                "action": "start",
                "record": "S",
                "event_type": "JOB",
                "identifier": "119.pdw-s1",
                "scheduler_timestamp": datetime(2021, 4, 18, 19, 24, 23),
                "sub_identifier": "0",
                "user": "someuser73",
                "raw_user_name": "someuser73",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "queue": "somequeue32",
                "jobname": "STDIN",
                "group": "someuser73",
                "Resource_List.nodect": 3,
                "Resource_List.ncpus": 15,
                "exec_host": ["pdw-c1", "pdw-c1", "pdw-c2"],
                "exec_vnode": [
                    "pdw-c1-v00",
                    "pdw-c1-v01",
                    "pdw-c1-v01",
                    "pdw-c1-v02",
                    "pdw-c1-v03",
                    "pdw-c1-v04",
                    "pdw-c1-v04-s0",
                    "pdw-c2-v00",
                    "pdw-c2-v01",
                ],
                "raw_exec_vnode": "(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=1)+("
                "pdw-c1-v01:ncpus=1+pdw-c1-v02:ncpus=1+pdw-c1-v03:ncpus=1+pdw-c1-v04:ncpus=1+pdw-c1-v04-s0"
                ":ncpus=1)+(pdw-c2-v00:ncpus=4+pdw-c2-v01:ncpus=1)",
                "ctime": datetime(2021, 4, 18, 19, 24, 23),
                "qtime": datetime(2021, 4, 18, 19, 24, 23),
                "etime": datetime(2021, 4, 18, 19, 24, 23),
                "start": datetime(2021, 4, 18, 19, 24, 23),
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 3,
                        "ncpus": "5",
                        "obj_type": "gpu",
                    },
                ],
            },
            True,
        ),
        (
            "Job end: 3 exec_hosts, 9 exec_vnodes ( qsub -l select=3:ncpus=5:obj_type=gpu -- /bin/sleep 10)",
            "04/18/2021 19:24:34;E;119.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=STDIN queue=somequeue32 "
            "ctime=1618773863 qtime=1618773863 etime=1618773863 start=1618773863 exec_host=pdw-c1/0*0+pdw-c1/1*0+pdw-c2/0*0 "
            "exec_vnode=(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=1)+("
            "pdw-c1-v01:ncpus=1+pdw-c1-v02:ncpus=1+pdw-c1-v03:ncpus=1+pdw-c1-v04:ncpus=1+pdw-c1-v04-s0:ncpus=1)+("
            "pdw-c2-v00:ncpus=4+pdw-c2-v01:ncpus=1) Resource_List.ncpus=15 Resource_List.nodect=3 Resource_List.place=free "
            "Resource_List.select=3:ncpus=5:obj_type=gpu session=2072 end=1618773874 Exit_status=0 resources_used.cpupercent=0 "
            "resources_used.cput=00:00:00 resources_used.mem=868kb resources_used.ncpus=15 resources_used.vmem=12916kb "
            "resources_used.walltime=00:00:10 run_count=1",
            {
                "action": "end",
                "record": "E",
                "event_type": "JOB",
                "identifier": "119.pdw-s1",
                "scheduler_timestamp": datetime(2021, 4, 18, 19, 24, 34),
                "sub_identifier": "0",
                "user": "someuser73",
                "raw_user_name": "someuser73",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "queue": "somequeue32",
                "jobname": "STDIN",
                "group": "someuser73",
                "Resource_List.nodect": 3,
                "Resource_List.ncpus": 15,
                "exec_host": ["pdw-c1", "pdw-c1", "pdw-c2"],
                "exec_vnode": [
                    "pdw-c1-v00",
                    "pdw-c1-v01",
                    "pdw-c1-v01",
                    "pdw-c1-v02",
                    "pdw-c1-v03",
                    "pdw-c1-v04",
                    "pdw-c1-v04-s0",
                    "pdw-c2-v00",
                    "pdw-c2-v01",
                ],
                "raw_exec_vnode": "(pdw-c1-v00:ncpus=4+pdw-c1-v01:ncpus=1)+("
                "pdw-c1-v01:ncpus=1+pdw-c1-v02:ncpus=1+pdw-c1-v03:ncpus=1+pdw-c1-v04:ncpus=1+pdw-c1-v04-s0"
                ":ncpus=1)+(pdw-c2-v00:ncpus=4+pdw-c2-v01:ncpus=1)",
                "ctime": datetime(2021, 4, 18, 19, 24, 23),
                "qtime": datetime(2021, 4, 18, 19, 24, 23),
                "etime": datetime(2021, 4, 18, 19, 24, 23),
                "start": datetime(2021, 4, 18, 19, 24, 23),
                "end": datetime(2021, 4, 18, 19, 24, 34),
                "Exit_status": "0",
                "resources_used.walltime": 10,
                "run_count": 1,
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 3,
                        "ncpus": "5",
                        "obj_type": "gpu",
                    },
                ],
                "session": "2072",
            },
            True,
        ),
        (
            "13 Job start: 3 exec_hosts, 3 exec_vnodes (qsub -l select=3:ncpus=5 -- /bin/sleep 10)",
            "04/21/2021 14:16:41;S;50.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=STDIN queue=somequeue32 "
            "ctime=1619014601 qtime=1619014601 etime=1619014601 start=1619014601 exec_host=pdw-c1/0*5+pdw-c2/0*5+pdw-c3/0*5 "
            "exec_vnode=(pdw-c1:ncpus=5)+(pdw-c2:ncpus=5)+(pdw-c3:ncpus=5) Resource_List.ncpus=15 Resource_List.nodect=3 "
            "Resource_List.place=free Resource_List.select=3:ncpus=5 resource_assigned.mem=6106404kb resource_assigned.ncpus=24",
            {
                "action": "start",
                "record": "S",
                "event_type": "JOB",
                "identifier": "50.pdw-s1",
                "scheduler_timestamp": datetime(2021, 4, 21, 14, 16, 41),
                "sub_identifier": "0",
                "user": "someuser73",
                "raw_user_name": "someuser73",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "queue": "somequeue32",
                "jobname": "STDIN",
                "group": "someuser73",
                "Resource_List.nodect": 3,
                "Resource_List.ncpus": 15,
                "exec_vnode": ["pdw-c1", "pdw-c2", "pdw-c3"],
                "raw_exec_vnode": "(pdw-c1:ncpus=5)+(pdw-c2:ncpus=5)+(pdw-c3:ncpus=5)",
                # exec_host=pdw-c1/0*5+pdw-c2/0*5+pdw-c3/0*5
                "exec_host": ["pdw-c1", "pdw-c2", "pdw-c3"],
                "ctime": datetime(2021, 4, 21, 14, 16, 41),
                "qtime": datetime(2021, 4, 21, 14, 16, 41),
                "etime": datetime(2021, 4, 21, 14, 16, 41),
                "start": datetime(2021, 4, 21, 14, 16, 41),
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 3,
                        "ncpus": "5",
                    },
                ],
            },
            True,
        ),
        (
            "Job end: 3 exec_hosts, 3 exec_vnodes (qsub -l select=3:ncpus=5 -- /bin/sleep 10)",
            "04/21/2021 14:16:52;E;50.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=STDIN queue=somequeue32 "
            "ctime=1619014601 qtime=1619014601 etime=1619014601 start=1619014601 exec_host=pdw-c1/0*5+pdw-c2/0*5+pdw-c3/0*5 "
            "exec_vnode=(pdw-c1:ncpus=5)+(pdw-c2:ncpus=5)+(pdw-c3:ncpus=5) Resource_List.ncpus=15 Resource_List.nodect=3 "
            "Resource_List.place=free Resource_List.select=3:ncpus=5 session=1144 end=1619014612 Exit_status=0 "
            "resources_used.cpupercent=0 resources_used.cput=00:00:00 resources_used.mem=848kb resources_used.ncpus=15 "
            "resources_used.vmem=12916kb resources_used.walltime=00:00:10 run_count=1",
            {
                "action": "end",
                "record": "E",
                "event_type": "JOB",
                "identifier": "50.pdw-s1",
                "scheduler_timestamp": datetime(2021, 4, 21, 14, 16, 52),
                "sub_identifier": "0",
                "user": "someuser73",
                "raw_user_name": "someuser73",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "queue": "somequeue32",
                "jobname": "STDIN",
                "group": "someuser73",
                "Resource_List.nodect": 3,
                "Resource_List.ncpus": 15,
                "exec_vnode": ["pdw-c1", "pdw-c2", "pdw-c3"],
                "raw_exec_vnode": "(pdw-c1:ncpus=5)+(pdw-c2:ncpus=5)+(pdw-c3:ncpus=5)",
                # exec_host=pdw-c1/0*5+pdw-c2/0*5+pdw-c3/0*5
                "exec_host": ["pdw-c1", "pdw-c2", "pdw-c3"],
                "ctime": datetime(2021, 4, 21, 14, 16, 41),
                "qtime": datetime(2021, 4, 21, 14, 16, 41),
                "etime": datetime(2021, 4, 21, 14, 16, 41),
                "start": datetime(2021, 4, 21, 14, 16, 41),
                "end": datetime(2021, 4, 21, 14, 16, 52),
                "Exit_status": "0",
                "resources_used.walltime": 10,
                "run_count": 1,
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 3,
                        "ncpus": "5",
                    },
                ],
                "session": "1144",
            },
            True,
        ),
        (
            "Job end: 2 exec_hosts, 2 exec_vnodes (jobname=pbs_job_script.sh)",
            '04/21/2021 18:46:27;E;0.pdw-s1;user=someuser73 group=someuser73 account="account1" project=project1 '
            "jobname=pbs_job_script.sh queue=somequeue32 ctime=1619030783 qtime=1619030783 etime=1619030783 start=1619030783 "
            "exec_host=pdw-c1/0*2+pdw-c1/1*2 exec_vnode=(pdw-c1:ncpus=2)+(pdw-c1:ncpus=2) Resource_List.ncpus=4 "
            "Resource_List.nodect=2 Resource_List.place=free Resource_List.select=2:ncpus=2 Resource_List.walltime=00:05:00 "
            "session=561 end=1619030787 Exit_status=0 resources_used.cpupercent=0 resources_used.cput=00:00:00 "
            "resources_used.mem=0kb resources_used.ncpus=4 resources_used.vmem=0kb resources_used.walltime=00:00:03 run_count=1",
            {
                "Exit_status": "0",
                "Resource_List.ncpus": 4,
                "Resource_List.nodect": 2,
                "Resource_List.walltime": 300,
                "action": "end",
                "ctime": datetime(2021, 4, 21, 18, 46, 23),
                "end": datetime(2021, 4, 21, 18, 46, 27),
                "etime": datetime(2021, 4, 21, 18, 46, 23),
                "event_type": "JOB",
                # exec_host=pdw-c1/0*2+pdw-c1/1*2
                "exec_host": ["pdw-c1", "pdw-c1"],
                "raw_exec_vnode": "(pdw-c1:ncpus=2)+(pdw-c1:ncpus=2)",
                "exec_vnode": ["pdw-c1", "pdw-c1"],
                "group": "someuser73",
                "identifier": "0.pdw-s1",
                "jobname": "pbs_job_script.sh",
                "qtime": datetime(2021, 4, 21, 18, 46, 23),
                "queue": "somequeue32",
                "project": "project1",
                "raw_project_name": "project1",
                "account": "account1",
                "raw_account_name": "account1",
                "raw_user_name": "someuser73",
                "record": "E",
                "resources_used.walltime": 3,
                "scheduler_timestamp": datetime(2021, 4, 21, 18, 46, 27),
                "start": datetime(2021, 4, 21, 18, 46, 23),
                "sub_identifier": "0",
                "user": "someuser73",
                "run_count": 1,
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 2,
                        "ncpus": "2",
                    },
                ],
                "session": "561",
            },
            True,
        ),
        (
            "Job start: 3 exec_hosts, 3 exec_vnodes (jobname=pbs_job_script.sh)",
            "04/21/2021 18:48:25;S;7.pdw-s1;user=someuser73 group=someuser73 project=someuser72 jobname=pbs_job_script.sh "
            "queue=somequeue32 ctime=1619030785 qtime=1619030785 etime=1619030785 start=1619030905 "
            "exec_host=pdw-c3/0*3+pdw-c3/1*3+pdw-c4/0*3 exec_vnode=(pdw-c3:ncpus=3:mem=1kb)+(pdw-c3:ncpus=3:mem=1kb)+("
            "pdw-c4:ncpus=3:mem=1kb) Resource_List.mem=9b Resource_List.ncpus=9 Resource_List.nodect=3 Resource_List.place=free "
            "Resource_List.select=3:ncpus=3:mem=3 Resource_List.walltime=00:05:00 resource_assigned.mem=4070936kb "
            "resource_assigned.ncpus=16",
            {
                "Resource_List.ncpus": 9,
                "Resource_List.nodect": 3,
                "Resource_List.walltime": 300,
                "action": "start",
                "ctime": datetime(2021, 4, 21, 18, 46, 25),
                "etime": datetime(2021, 4, 21, 18, 46, 25),
                "event_type": "JOB",
                # exec_host=pdw-c3/0*3+pdw-c3/1*3+pdw-c4/0*3
                "exec_host": ["pdw-c3", "pdw-c3", "pdw-c4"],
                "exec_vnode": ["pdw-c3", "pdw-c3", "pdw-c4"],
                "raw_exec_vnode": "(pdw-c3:ncpus=3:mem=1kb)+(pdw-c3:ncpus=3:mem=1kb)+(pdw-c4:ncpus=3:mem=1kb)",
                "group": "someuser73",
                "identifier": "7.pdw-s1",
                "jobname": "pbs_job_script.sh",
                "qtime": datetime(2021, 4, 21, 18, 46, 25),
                "queue": "somequeue32",
                "project": "someuser72",
                "raw_project_name": "someuser72",
                "raw_user_name": "someuser73",
                "record": "S",
                "scheduler_timestamp": datetime(2021, 4, 21, 18, 48, 25),
                "start": datetime(2021, 4, 21, 18, 48, 25),
                "sub_identifier": "0",
                "user": "someuser73",
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 3,
                        "mem": "3",
                        "ncpus": "3",
                    },
                ],
            },
            True,
        ),
        (
            "Job end: 2 exec_hosts, 2 exec_vnodes (jobname=pbs_job_script_16.sh)",
            '04/22/2021 16:16:04;E;1.pdw-s1;user=someuser73 group=someuser73 account="account2" project=project2 '
            "jobname=pbs_job_script_16.sh queue=somequeue32 ctime=1619108159 qtime=1619108159 etime=1619108159 start=1619108159 "
            "exec_host=pdw-c2/0*8+pdw-c3/0*8 exec_vnode=(pdw-c2:ncpus=8)+(pdw-c3:ncpus=8) Resource_List.ncpus=16 "
            "Resource_List.nodect=2 Resource_List.place=free Resource_List.select=2:ncpus=8 Resource_List.walltime=00:05:00 "
            "session=562 end=1619108164 Exit_status=0 resources_used.cpupercent=0 resources_used.cput=00:00:00 "
            "resources_used.mem=820kb resources_used.ncpus=16 resources_used.vmem=12916kb resources_used.walltime=00:00:03 "
            "run_count=1",
            {
                "Exit_status": "0",
                "Resource_List.ncpus": 16,
                "Resource_List.nodect": 2,
                "Resource_List.walltime": 300,
                "action": "end",
                "ctime": datetime(2021, 4, 22, 16, 15, 59),
                "end": datetime(2021, 4, 22, 16, 16, 4),
                "etime": datetime(2021, 4, 22, 16, 15, 59),
                "event_type": "JOB",
                # exec_host=pdw-c2/0*8+pdw-c3/0*8
                "exec_host": ["pdw-c2", "pdw-c3"],
                "raw_exec_vnode": "(pdw-c2:ncpus=8)+(pdw-c3:ncpus=8)",
                "exec_vnode": ["pdw-c2", "pdw-c3"],
                "group": "someuser73",
                "identifier": "1.pdw-s1",
                "jobname": "pbs_job_script_16.sh",
                "qtime": datetime(2021, 4, 22, 16, 15, 59),
                "queue": "somequeue32",
                "account": "account2",
                "raw_account_name": "account2",
                "project": "project2",
                "raw_project_name": "project2",
                "raw_user_name": "someuser73",
                "record": "E",
                "resources_used.walltime": 3,
                "scheduler_timestamp": datetime(2021, 4, 22, 16, 16, 4),
                "start": datetime(2021, 4, 22, 16, 15, 59),
                "sub_identifier": "0",
                "user": "someuser73",
                "run_count": 1,
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 2,
                        "ncpus": "8",
                    },
                ],
                "session": "562",
            },
            True,
        ),
        (
            "18 Job start: (jobname=N/A mode=script)",
            "02/14/2020 23:59:33;S;409766;Resource_List.ncpus=4 Resource_List.nodect=4 Resource_List.walltime=0:30:00 "
            "account=funny args= ctime=1581722605.06 cwd=/fs/somedirectoryobj826 etime=1581722605.06 "
            'exe=/fs/somedirectoryobj825 exec_host="3824-3826,3829" group=unknown jobname=N/A '
            "mode=script qtime=1581722605.06 queue=somequeue79 resource=somemachine3 session=unknown start=1581724773.64 "
            "user=someuser234",
            {
                "Resource_List.ncpus": 4,
                "Resource_List.nodect": 4,
                "Resource_List.walltime": 1800,
                "action": "start",
                "ctime": datetime(2020, 2, 14, 23, 23, 25),
                "cwd": "/fs/somedirectoryobj826",
                "etime": datetime(2020, 2, 14, 23, 23, 25),
                "event_type": "JOB",
                "exe": "/fs/somedirectoryobj825",
                # exec_host="3824-3826,3829"
                "exec_host": ["3824-3826,3829"],
                "group": "unknown",
                "identifier": "409766",
                "jobname": "N/A",
                "mode": "script",
                "qtime": datetime(2020, 2, 14, 23, 23, 25),
                "queue": "somequeue79",
                "account": "funny",
                "raw_account_name": "funny",
                "user": "someuser234",
                "raw_user_name": "someuser234",
                "record": "S",
                "resource": "somemachine3",
                "scheduler_timestamp": datetime(2020, 2, 14, 23, 59, 33),
                "start": datetime(2020, 2, 14, 23, 59, 33),
                "sub_identifier": "0",
                "args": "",
                "session": "unknown",
            },
            True,
        ),
        (
            "19 suspend",
            "11/29/2022 19:56:18;z;54095.somepbshost;resources_used.cpupercent=0 resources_used.cput=00:00:00 "
            "resources_used.mem=0b resources_used.ncpus=1 resources_used.vmem=0kb resources_used.walltime=00:00:00",
            {
                "action": "suspend",
                "event_type": "JOB",
                "identifier": "54095.somepbshost",
                "record": "z",
                "resources_used.cpupercent": "0",
                "resources_used.cput": "00:00:00",
                "resources_used.mem": "0b",
                "resources_used.ncpus": "1",
                "resources_used.vmem": "0kb",
                "resources_used.walltime": 0,
                "scheduler_timestamp": datetime(2022, 11, 29, 19, 56, 18),
                "sub_identifier": "0",
            },
            True,
        ),
        (
            "20 resume",
            "11/29/2022 19:56:34;r;54095.somepbshost;",
            {
                "action": "resume",
                "event_type": "JOB",
                "identifier": "54095.somepbshost",
                "record": "r",
                "scheduler_timestamp": datetime(2022, 11, 29, 19, 56, 34),
                "sub_identifier": "0",
            },
            True,
        ),
        (
            "21 log lines with different nodes used than requested",
            "01/17/2013 21:01:10;Q;50888;queue=somequeue78",
            {
                "event_type": "JOB",
                "sub_identifier": "0",
                "action": "queue",
                "identifier": "50888",
                "queue": "somequeue78",
                "record": "Q",
                "scheduler_timestamp": datetime(2013, 1, 17, 21, 1, 10),
            },
            True,
        ),
        (
            "22 Fix for PBS_Utils.pbs_accounting.BadLineError A",
            "10/26/2021 16:24:55;D;52.pdw-s1;requestor=someuser73@pdw-s1.pdw.local",
            {
                "action": "delete",
                "record": "D",
                "event_type": "JOB",
                "identifier": "52.pdw-s1",
                "scheduler_timestamp": datetime(2021, 10, 26, 16, 24, 55),
                "sub_identifier": "0",
                "requester": "someuser73@pdw-s1.pdw.local",
            },
            True,
        ),
        (
            "23 Fix for PBS_Utils.pbs_accounting.BadLineError B",
            "10/26/2021 16:24:55;A;53.pdw-s1;Job deleted as result of dependency on job 52.pdw-s1",
            {
                "action": "abort",
                "record": "A",
                "event_type": "JOB",
                "identifier": "53.pdw-s1",
                "scheduler_timestamp": datetime(2021, 10, 26, 16, 24, 55),
                "sub_identifier": "0",
            },
            True,
        ),
        (
            "24 job array end",
            '05/10/2023 17:33:35;E;5[].pdw-s1;user=someuser73 group=someuser73 account="account3" project=Project3 jobname=name3 '
            "queue=somequeue32 ctime=1683740001 qtime=1683740001 etime=1683740001 start=1683740001 Resource_List.ncpus=6 "
            "Resource_List.nodect=2 Resource_List.place=free Resource_List.select=2:ncpus=3 Resource_List.walltime=00:05:00 "
            "session=0 end=1683740015 Exit_status=0 run_count=0",
            {
                "Exit_status": "0",
                "Resource_List.ncpus": 6,
                "Resource_List.nodect": 2,
                "Resource_List.walltime": 300,
                "action": "end",
                "ctime": datetime(2023, 5, 10, 17, 33, 21),
                "end": datetime(2023, 5, 10, 17, 33, 35),
                "etime": datetime(2023, 5, 10, 17, 33, 21),
                "event_type": "JOB",
                "group": "someuser73",
                "identifier": "5[].pdw-s1",
                "jobname": "name3",
                "qtime": datetime(2023, 5, 10, 17, 33, 21),
                "queue": "somequeue32",
                "account": "account3",
                "raw_account_name": "account3",
                "project": "Project3",
                "raw_project_name": "Project3",
                "raw_user_name": "someuser73",
                "record": "E",
                "run_count": 0,
                "scheduler_timestamp": datetime(2023, 5, 10, 17, 33, 35),
                "start": datetime(2023, 5, 10, 17, 33, 21),
                "sub_identifier": "0",
                "user": "someuser73",
                "Resource_List.place": {
                    "arrangement": "free",
                    "groups": [],
                    "sharing": None,
                },
                "Resource_List.select": [
                    {
                        "N": 2,
                        "ncpus": "3",
                    },
                ],
                "session": "0",
            },
            True,
        ),
        (
            "25 job with one quote",
            """10/24/2023 17:31:14;S;674670;Resource_List.ncpus=2 Resource_List.nodect=2 Resource_List.walltime=1:00:00 
          account=account831 args= ctime=1698168671.61 cwd=/fs/user827/MA etime=1698168671.61 
          exe=/fs/somedirectoryobj827 exec_host="3828,3837" group=unknown jobname=somemachine3_debug_1_core.txt' mode=script 
          qtime=1698168671.61 queue=somequeue1 resource=somemachine3 session=unknown start=1698168674.61 user=user827""",
            {
                "Resource_List.ncpus": 2,
                "Resource_List.nodect": 2,
                "Resource_List.walltime": 3600,
                "account": "account831",
                "action": "start",
                "ctime": datetime(2023, 10, 24, 17, 31, 11),
                "cwd": "/fs/user827/MA",
                "etime": datetime(2023, 10, 24, 17, 31, 11),
                "event_type": "JOB",
                "exe": "/fs/somedirectoryobj827",
                "exec_host": ["3828,3837"],
                "group": "unknown",
                "identifier": "674670",
                "jobname": "somemachine3_debug_1_core.txt'",
                "mode": "script",
                "qtime": datetime(2023, 10, 24, 17, 31, 11),
                "queue": "somequeue1",
                "raw_account_name": "account831",
                "raw_user_name": "user827",
                "record": "S",
                "resource": "somemachine3",
                "scheduler_timestamp": datetime(2023, 10, 24, 17, 31, 14),
                "start": datetime(2023, 10, 24, 17, 31, 14),
                "sub_identifier": "0",
                "user": "user827",
                "args": "",
                "session": "unknown",
            },
            True,
        ),
        (
            "26 job with one quote",
            """10/24/2023 18:10:32;E;674670;Exit_status=0 Resource_List.ncpus=2 Resource_List.nodect=2 
            Resource_List.walltime=1:00:00 account=account831 args= ctime=1698168671.61 cwd=/fs/user827/MA 
            end=1698171032.04 etime=1698168671.61 exe=/fs/somedirectoryobj827 exec_host="3828,3837" group=unknown 
            jobname=somemachine3_debug_1_core.txt' mode=script qtime=1698168671.61 queue=somequeue1 resource=somemachine3 
            resources_used.location="3828,3837" resources_used.nodect=2 resources_used.walltime=0:39:17 session=unknown 
            start=1698168674.61 user=user827""",
            {
                "Exit_status": "0",
                "Resource_List.ncpus": 2,
                "Resource_List.nodect": 2,
                "Resource_List.walltime": 3600,
                "account": "account831",
                "action": "end",
                "ctime": datetime(2023, 10, 24, 17, 31, 11),
                "cwd": "/fs/user827/MA",
                "end": datetime(2023, 10, 24, 18, 10, 32),
                "etime": datetime(2023, 10, 24, 17, 31, 11),
                "event_type": "JOB",
                "exe": "/fs/somedirectoryobj827",
                "exec_host": ["3828,3837"],
                "group": "unknown",
                "identifier": "674670",
                "jobname": "somemachine3_debug_1_core.txt'",
                "mode": "script",
                "qtime": datetime(2023, 10, 24, 17, 31, 11),
                "queue": "somequeue1",
                "raw_account_name": "account831",
                "raw_user_name": "user827",
                "record": "E",
                "resource": "somemachine3",
                "resources_used.nodect": 2,
                "resources_used.walltime": 2357,
                "scheduler_timestamp": datetime(2023, 10, 24, 18, 10, 32),
                "start": datetime(2023, 10, 24, 17, 31, 14),
                "sub_identifier": "0",
                "user": "user827",
                "args": "",
                "session": "unknown",
            },
            True,
        ),
        (
            "27 walltime test",
            """10/26/2020 22:10:50;E;475302;Exit_status=2 Resource_List.ncpus=128 Resource_List.nodect=128 
         Resource_List.walltime=00:10:00""",
            {
                "Exit_status": "2",
                "Resource_List.ncpus": 128,
                "Resource_List.nodect": 128,
                "Resource_List.walltime": 600,
                "action": "end",
                "event_type": "JOB",
                "identifier": "475302",
                "record": "E",
                "scheduler_timestamp": datetime(2020, 10, 26, 22, 10, 50),
                "sub_identifier": "0",
            },
            True,
        ),
        (
            "28 this is an invalid key, which is part of walltime."
            "Resource_List.walltime=00:10:00 Resource_List.select:2:ncpus=128:mpiprocs=128",
            """10/26/2023 08:41:14;Q;1702.imgt1;user=user968 group=somegroup21 account="someproject23" project=someproject23 
         jobname=hello_world queue=somequeue65 ctime=1698327674 qtime=1698327674 etime=0 
         Resource_List.allow_account_check_failure=True Resource_List.allow_negative_allocation=True 
         Resource_List.award_category=someword37 Resource_List.award_type=someword37 Resource_List.burn_ratio=0 
         Resource_List.current_allocation=460795510784 Resource_List.ncpus=128 Resource_List.ni_resource=improv 
         Resource_List.nodect=1 Resource_List.overburn=False Resource_List.place=pack Resource_List.project_priority=2 
         Resource_List.route_backfill=False Resource_List.select=1:ncpus=128 Resource_List.total_allocation=460800000000 
         Resource_List.walltime=00:10:00 Resource_List.select:2:ncpus=128:mpiprocs=128""",
            Exception,
            False,
        ),
        (
            "malformed command line leakage: Resources_List.-I=",
            """02/23/2024 22:33:55;Q;621652.somehost;user=someuser92 group=somegroup20
            account="someproject" project=someproject jobname=hello_world.sh queue=somequeue ctime=1708727635 qtime=1708727635
            etime=0 Resource_List.allow_account_check_failure=True Resource_List.allow_negative_allocation=True
            Resource_List.award_category=someword34 Resource_List.award_type=someword234 Resource_List.burn_ratio=0
            Resource_List.current_allocation=-280471872 Resource_List.hbm_mode=flat
            Resource_List.hpcm_image=compute_test_20240120T003017_0051f5d Resource_List.ncpus=208 Resource_List.ni_resource=aurora
            Resource_List.nodect=1 Resource_List.overburn=False Resource_List.place=scatter Resource_List.project_priority=2
            Resource_List.route_backfill=True Resource_List.select=1 Resource_List.snc_mode=quad Resource_List.total_allocation=0
            Resource_List.walltime=00:30:00 Resource_List.-I=""",
            {
                "Resource_List.ncpus": 208,
                "Resource_List.nodect": 1,
                "Resource_List.place": {"arrangement": "scatter", "groups": [], "sharing": None},
                "Resource_List.select": [{"N": 1}],
                "Resource_List.walltime": 1800,
                "account": "someproject",
                "action": "queue",
                "ctime": datetime(2024, 2, 23, 22, 33, 55),
                "event_type": "JOB",
                "group": "somegroup20",
                "identifier": "621652.somehost",
                "jobname": "hello_world.sh",
                "project": "someproject",
                "qtime": datetime(2024, 2, 23, 22, 33, 55),
                "queue": "somequeue",
                "raw_account_name": "someproject",
                "raw_project_name": "someproject",
                "raw_user_name": "someuser92",
                "record": "Q",
                "scheduler_timestamp": datetime(2024, 2, 23, 22, 33, 55),
                "sub_identifier": "0",
                "user": "someuser92",
            },
            True,
        ),
        (
            "'-' in Resource name: Resources_List.foo-bar",
            """02/23/2024 22:33:55;Q;621652.somehost;user=someuser92 group=somegroup21
            account="someproject" project=someproject jobname=hello_world.sh queue=somequeue ctime=1708727635 qtime=1708727635
            etime=0 Resource_List.allow_account_check_failure=True Resource_List.allow_negative_allocation=True
            Resource_List.award_category=someword34 Resource_List.award_type=someword234 Resource_List.burn_ratio=0
            Resource_List.current_allocation=-280471872 Resource_List.hbm_mode=flat
            Resource_List.hpcm_image=compute_test_20240120T003017_0051f5d Resource_List.ncpus=208 Resource_List.ni_resource=aurora
            Resource_List.nodect=1 Resource_List.overburn=False Resource_List.place=scatter Resource_List.project_priority=2
            Resource_List.route_backfill=True Resource_List.select=1 Resource_List.snc_mode=quad Resource_List.total_allocation=0
            Resource_List.walltime=00:30:00 Resource_List.foo-bar=baz""",
            {
                "Resource_List.ncpus": 208,
                "Resource_List.nodect": 1,
                "Resource_List.place": {"arrangement": "scatter", "groups": [], "sharing": None},
                "Resource_List.select": [{"N": 1}],
                "Resource_List.walltime": 1800,
                "account": "someproject",
                "action": "queue",
                "ctime": datetime(2024, 2, 23, 22, 33, 55),
                "event_type": "JOB",
                "group": "somegroup21",
                "identifier": "621652.somehost",
                "jobname": "hello_world.sh",
                "project": "someproject",
                "qtime": datetime(2024, 2, 23, 22, 33, 55),
                "queue": "somequeue",
                "raw_account_name": "someproject",
                "raw_project_name": "someproject",
                "raw_user_name": "someuser92",
                "record": "Q",
                "scheduler_timestamp": datetime(2024, 2, 23, 22, 33, 55),
                "sub_identifier": "0",
                "user": "someuser92",
            },
            True,
        ),
        (
            "malformed command line leakage: Resource_List.selec816:ncpus=208",
            """02/22/2024 04:30:47;Q;620255.somehost;user=someuser45 group=somegroup21
            account="someproject" project=someproject jobname=STDIN queue=somequeue ctime=1708576247 qtime=1708576247 etime=0
            Resource_List.allow_account_check_failure=True Resource_List.allow_negative_allocation=True
            Resource_List.award_category=someword34 Resource_List.award_type=someword34 Resource_List.burn_ratio=0
            Resource_List.current_allocation=-288779960320 Resource_List.hbm_mode=flat
            Resource_List.hpcm_image=compute_test_20240120T003017_0051f5d Resource_List.ncpus=1 Resource_List.ni_resource=aurora
            Resource_List.nodect=1 Resource_List.overburn=False Resource_List.place=pack Resource_List.project_priority=2
            Resource_List.route_backfill=False Resource_List.select=1:ncpus=1 Resource_List.snc_mode=quad 
            Resource_List.total_allocation=0
            Resource_List.walltime=04:00:00 Resource_List.selec816:ncpus=208""",
            Exception,
            False,
        ),
        (
            "line with resvname",
            '01/21/2023 22:10:18;E;396595.somehost;user=someuser322 group=somegroup21 account="RL-fold" project=projecta '
            "jobname=submit.sh queue=someresv23 resvname=ResvName resvID=someresv23.somehost ctime=1674338973 "
            "qtime=1674338974 etime=1674338974 start=1674338976 exec_host=x3006c0s13b1n0/0*64+x3006c0s19b0n0/0*64 exec_vnode=("
            "x3006c0s13b1n0:ncpus=64)+(x3006c0s19b0n0:ncpus=64) Resource_List.allow_account_check_failure=True "
            "Resource_List.allow_negative_allocation=True Resource_List.award_category=someword34 "
            "Resource_List.award_type=someword34 Resource_List.backfill_factor=84600 Resource_List.backfill_max=50 "
            "Resource_List.base_score=0 Resource_List.burn_ratio=0.2028 Resource_List.current_allocation=4018003968 "
            "Resource_List.eagle_fs=True Resource_List.enable_backfill=0 Resource_List.enable_fifo=0 Resource_List.enable_wfp=0 "
            "Resource_List.fifo_factor=1800 Resource_List.filesystems=home:eagle Resource_List.home_fs=True "
            "Resource_List.ncpus=128 Resource_List.ni_resource=somemachine4 Resource_List.nodect=2 Resource_List.overburn=False "
            "Resource_List.place=scatter Resource_List.preempt_targets=NONE Resource_List.project_priority=2 "
            "Resource_List.route_backfill=False Resource_List.score_boost=0 Resource_List.select=2:system=somemachine4 "
            "Resource_List.total_allocation=5040000000 Resource_List.total_cpus=560 Resource_List.walltime=00:30:00 "
            "Resource_List.wfp_factor=100000 session=0 end=1674339018 Exit_status=1 resources_used.cpupercent=0 "
            "resources_used.cput=00:00:00 resources_used.mem=3880kb resources_used.ncpus=128 resources_used.vmem=16212kb "
            "resources_used.walltime=00:00:01 eligible_time=00:00:08 run_count=1",
            {
                "Exit_status": "1",
                "Resource_List.ncpus": 128,
                "Resource_List.nodect": 2,
                "Resource_List.place": {"arrangement": "scatter", "groups": [], "sharing": None},
                "Resource_List.select": [{"N": 2, "system": "somemachine4"}],
                "Resource_List.walltime": 1800,
                "account": "RL-fold",
                "action": "end",
                "ctime": datetime(2023, 1, 21, 22, 9, 33),
                "end": datetime(2023, 1, 21, 22, 10, 18),
                "etime": datetime(2023, 1, 21, 22, 9, 34),
                "event_type": "JOB",
                "exec_host": ["x3006c0s13b1n0", "x3006c0s19b0n0"],
                "exec_vnode": ["x3006c0s13b1n0", "x3006c0s19b0n0"],
                "group": "somegroup21",
                "identifier": "396595.somehost",
                "jobname": "submit.sh",
                "project": "projecta",
                "qtime": datetime(2023, 1, 21, 22, 9, 34),
                "queue": "someresv23",
                "raw_account_name": "RL-fold",
                "raw_exec_vnode": "(x3006c0s13b1n0:ncpus=64)+(x3006c0s19b0n0:ncpus=64)",
                "raw_project_name": "projecta",
                "raw_user_name": "someuser322",
                "record": "E",
                "resources_used.walltime": 1,
                "resvID": "someresv23.somehost",
                "resvname": "ResvName",
                "run_count": 1,
                "scheduler_timestamp": datetime(2023, 1, 21, 22, 10, 18),
                "session": "0",
                "start": datetime(2023, 1, 21, 22, 9, 36),
                "sub_identifier": "0",
                "user": "someuser322",
            },
            True,
        ),
        (
            "line with resvname",
            "07/25/2023 23:54:37;E;562207.somemachine4-pbs-01.hsn.cm.somemachine4;user=someuser93 group=somegroup23 "
            'account="someproject2" project=someproject2 jobname=HACC_0512_C04_MPI04_GPU04_job queue=somequeue23 ctime=1690299801 '
            "qtime=1690299801 etime=1690299801 start=1690328787 exec_host=x3013c0s31b0n0/0*64 exec_vnode=("
            "x3013c0s31b0n0:ncpus=64) Resource_List.allow_account_check_failure=True Resource_List.allow_negative_allocation=True "
            "Resource_List.award_category=someword34 Resource_List.award_type=someword34 "
            "Resource_List.backfill_factor=84600 Resource_List.backfill_max=50 Resource_List.base_score=51 "
            "Resource_List.burn_ratio=0.0809 Resource_List.current_allocation=9264516096 Resource_List.enable_backfill=0 "
            "Resource_List.enable_fifo=1 Resource_List.enable_wfp=0 Resource_List.fifo_factor=1800 "
            "Resource_List.filesystems=grand Resource_List.grand_fs=True Resource_List.ncpus=64 Resource_List.ni_resource=somemachine4 "
            "Resource_List.nodect=1 Resource_List.overburn=False Resource_List.place=scatter Resource_List.preempt_targets=NONE "
            "Resource_List.project_priority=20 Resource_List.route_backfill=False Resource_List.score_boost=0 "
            "Resource_List.select=1 Resource_List.total_allocation=10080000000 Resource_List.total_cpus=560 "
            "Resource_List.walltime=00:30:00 Resource_List.wfp_factor=100000 session=49469 end=1690329277 Exit_status=0 "
            "resources_used.cpupercent=357 resources_used.cput=00:29:05 resources_used.mem=22536524kb resources_used.ncpus=64 "
            "resources_used.vmem=98586472kb resources_used.walltime=00:08:03 eligible_time=08:03:08 run_count=1",
            {
                "Exit_status": "0",
                "Resource_List.ncpus": 64,
                "Resource_List.nodect": 1,
                "Resource_List.place": {"arrangement": "scatter", "groups": [], "sharing": None},
                "Resource_List.select": [{"N": 1}],
                "Resource_List.walltime": 1800,
                "account": "someproject2",
                "action": "end",
                "ctime": datetime(2023, 7, 25, 15, 43, 21),
                "end": datetime(2023, 7, 25, 23, 54, 37),
                "etime": datetime(2023, 7, 25, 15, 43, 21),
                "event_type": "JOB",
                "exec_host": ["x3013c0s31b0n0"],
                "exec_vnode": ["x3013c0s31b0n0"],
                "group": "somegroup23",
                "identifier": "562207.somemachine4-pbs-01.hsn.cm.somemachine4",
                "jobname": "HACC_0512_C04_MPI04_GPU04_job",
                "project": "someproject2",
                "qtime": datetime(2023, 7, 25, 15, 43, 21),
                "queue": "somequeue23",
                "raw_account_name": "someproject2",
                "raw_exec_vnode": "(x3013c0s31b0n0:ncpus=64)",
                "raw_project_name": "someproject2",
                "raw_user_name": "someuser93",
                "record": "E",
                "resources_used.walltime": 483,
                "run_count": 1,
                "scheduler_timestamp": datetime(2023, 7, 25, 23, 54, 37),
                "session": "49469",
                "start": datetime(2023, 7, 25, 23, 46, 27),
                "sub_identifier": "0",
                "user": "someuser93",
            },
            True,
        ),
    ]
    if ids:
        deck = [f"{i:0>2}" for i, _ in enumerate(deck)]
    return deck


@pytest.mark.parametrize("line,valid_lt_line", input_pbs_line_extract_time(), ids=input_pbs_line_extract_time(ids=True))
def test_parse_pbs_log_line_extract_time(line, valid_lt_line):
    correct_logtime, correct_data = valid_lt_line
    logtime, data = pbs_line_extract_time(line)
    assert logtime == correct_logtime, f"RESULT:{logtime} != CORRECTRESULT:{correct_logtime}"
    assert data == correct_data, f"RESULT:{logtime} != CORRECTRESULT:{correct_data}"
    dct = parse_pbs_log_line(logtime, data)
    assert dct["record"] in ["S", "D", "Q", "E", "A"]
    assert dct["identifier"] is not None
    result_line = create_pbs_log_line(logtime, dct["record"], dct["identifier"], dct)
    new_logtime, new_data = pbs_line_extract_time(result_line)
    new_dct = parse_pbs_log_line(new_logtime, new_data)
    assert dct == new_dct


@pytest.mark.parametrize(
    "comment,line,valid_record,valid_data", input_parse_pbs_log_line(), ids=input_parse_pbs_log_line(ids=True)
)
def test_parse_pbs_log_line(comment, line, valid_record, valid_data):
    default_resource = "someresource"
    logtime, data = pbs_line_extract_time(line)
    if valid_record == BadLineError:
        with pytest.raises(BadLineError):
            parse_pbs_log_line(logtime, data, resource=default_resource)
    elif valid_record == Exception:
        with pytest.raises(Exception):
            parse_pbs_log_line(logtime, data, resource=default_resource)
    else:
        result = parse_pbs_log_line(logtime, data, resource=default_resource)
        pprint(result, width=132)
        # if 'resource' in result and result['resource'] != default_resource:
        #     # this is the case where the resource doesn't equal the default given resource.
        #     pass
        if " resource=" not in data:
            # add it in the records that should have it now.
            valid_record["resource"] = default_resource
        for key in result.keys():
            if key not in FieldHandlers.builtin_keys and key not in FieldHandlers.encode_passthrough_lst:
                assert key in data
        assert result == valid_record, pformat(result, width=132)


@pytest.mark.parametrize(
    "comment,line,valid_record,valid_data", input_parse_pbs_log_line(), ids=input_parse_pbs_log_line(ids=True)
)
def test_parse_pbs_log_line_with_resource(comment, line, valid_record, valid_data):
    logtime, data = pbs_line_extract_time(line)
    if valid_record == BadLineError:
        with pytest.raises(BadLineError):
            parse_pbs_log_line(logtime, data)
    elif valid_record == Exception:
        with pytest.raises(Exception):
            parse_pbs_log_line(logtime, data)
    else:
        result = parse_pbs_log_line(logtime, data)
        pprint(result, width=132)
        assert result == valid_record, pformat(result, width=132)


def check_kv_pair_records(src, dst):
    # these need to be split but account for quotes.
    assert parse_key_value_pairs(src) == parse_key_value_pairs(dst)


# fixme: add a bad data marker.
@pytest.mark.parametrize("comment,line,valid_result,valid_data", input_parse_kv_pairs(), ids=input_parse_kv_pairs(ids=True))
def test_pbs_parse_key_value_pairs(comment, line, valid_result, valid_data):
    logtime, data = pbs_line_extract_time(line)
    record_type, identifier, record_keyvals = split_record(data)
    if valid_result == BadLineError:
        with pytest.raises(BadLineError):
            parse_key_value_pairs(record_keyvals)
    else:
        result = parse_key_value_pairs(record_keyvals)
        for key, value in valid_result.items():
            assert value == result[key], f"{key}: {value} != {result[key]}"
        assert result == valid_result, f"{line=}\n{pformat(result)=}\n != \n {pformat(valid_result)=}"
        if valid_data:  # we lost and added data, we cannot regenerate.
            generated_record_keyvals = format_key_value_pairs(result)
            check_kv_pair_records(record_keyvals, generated_record_keyvals)


def input_get_time():
    deck = [
        # format,                   line ts,              valid
        (LOG_REGEX.LOG_FORMAT_PBS, "03/24/2011 13:43:28", 1300974208.0),
        (LOG_REGEX.LOG_FORMAT_PBS_US, "03/24/2011 13:43:28.012345", 1300974208.012345),
    ]
    return deck


@pytest.mark.parametrize("fmt,line_ts,valid", input_get_time())
def test_get_time(fmt, line_ts, valid):
    result = get_time(line_ts, fmt)
    assert result == valid


def input_hms_to_seconds():
    deck = [
        ("0:29:00", 1740),
        ("-1:52:00", 0),  # fixme: why???
        ("03:58:00", 14280),
        ("154:00:00", 554400),
        ("154:00", 9240),
        ("0:29", 29),
        ("1:16:00", 4560),
        ("0:01:45", 105),
    ]
    return deck


@pytest.mark.parametrize("walltime,correct_result", input_hms_to_seconds())
def test_hms_to_seconds(walltime, correct_result):
    result = hms_to_seconds(walltime)
    assert result == correct_result, "RESULT:%s CORRECT RESULT:%s" % (result, correct_result)
    result_seconds = hms_to_seconds(seconds_to_hms(result))
    assert result_seconds == result


def test_fix_newlines_extract_time_00():
    lines = [
        "05/02/2014 20:27:02;E;257928;Exit_status=0 Resource_List.ncpus=65536 Resource_List.nodect=4096 "
        "Resource_List.walltime=0:29:00 account=SSSPPg args=-g,-t,256k,-i,20,-p,-c,-w,-b,256k,-a,MPIIO ctime=1399061907.45 "
        "cwd=/fs/somedirectoryobj3 end=1399062422.19 etime=1399061907.45 "
        "exe=/fs/somedirectoryobj2"
        "-bridgeringagg-automated-nodebug exec_host=MIR-04C00-77FF1-4096 group=unknown jobname=N/A mode=c16 qtime=1399061907.45 "
        "queue=somequeue09 resources_used.location=MIR-04C00-77FF1-4096 resources_used.nodect=4096 resources_used.walltime=0:04:30 "
        "session=unknown start=1399062151.69 user=userg",
        "05/02/2014 20:27:35;S;256107;Resource_List.ncpus=1024 Resource_List.nodect=1024 Resource_List.walltime=6:00:00 "
        "account=someaccount2 args= ctime=1398890600.49 "
        "cwd=/fs/somedirectoryobj "
        "etime=1398890600.49 exe=/fs/somedirectoryobj4"
        ".0_ep_30.0/run_ensemble_job.292 exec_host=MIR-48800-7BB71-1-1024 group=unknown jobname=npt2-job292 mode=script "
        "qtime=1398890600.49 queue=somequeue09 session=unknown start=1399062455.04 user=someuser93",
        "05/02/2014 20:28:11;S;257828;Resource_List.ncpus=32768 Resource_List.nodect=2048 Resource_List.walltime=4:00:00 "
        "account=someproject2 args=outFoamx_SecXY.input",
        "ctime=1399056664.55 cwd=/fs/somedirectoryobjE etime=1399056664.55 "
        "exe=/fs/somedirectoryobj6 exec_host=MIR-08000-3B3F1-2048 "
        "group=unknown jobname=N/A mode=c16 qtime=1399056664.55 queue=somequeue09 session=unknown start=1399062491.92 user=useri",
        "05/02/2014 20:28:43;S;257924;Resource_List.ncpus=1024 Resource_List.nodect=1024 Resource_List.walltime=0:40:00 "
        "account=LiAirBattery args= ctime=1399061768.84 cwd=/fs/somedirectoryobjG "
        "etime=1399061768.84 exe=/fs/somedirectoryobj5 "
        "exec_host=MIR-48000-7B371-1-1024 group=unknown jobname=N/A mode=script qtime=1399061768.84 queue=somequeue09 "
        "session=unknown start=1399062523.88 user=userj",
        "05/02/2014 20:29:19;S;257931;Resource_List.ncpus=65536 Resource_List.nodect=2048 Resource_List.walltime=2:00:00 "
        "account=Ray_Benard args= ctime=1399062080.19 cwd=/fs/somedirectoryobjF etime=1399062080.19 "
        "exe=/fs/somedirectoryobj7 exec_host=MIR-48400-7B7F1-2048 group=unknown jobname=N/A mode=c32 "
        "qtime=1399062080.19 queue=somequeue09 session=unknown start=1399062559.5 user=userk",
        "05/02/2014 20:29:39;Q;257933;queue=somequeue64",
    ]
    parsed = []
    times_and_data = fix_newlines_extract_time(lines, version=2)
    assert len(times_and_data) == len(lines) - 1
    for logtime, data in times_and_data:
        parsed_line = parse_pbs_log_line(logtime, data)
        parsed.append(parsed_line)
    assert len(parsed) == len(lines) - 1


def test_fix_newlines_extract_time_01():
    lines = [
        "05/02/2014 20:27:02;E;257928;Exit_status=0 Resource_List.ncpus=65536 Resource_List.nodect=4096 "
        "Resource_List.walltime=0:29:00 account=SSSPPg args=-g,-t,256k,-i,20,-p,-c,-w,-b,256k,-a,MPIIO ctime=1399061907.45 "
        "cwd=/fs/somedirectoryobj3 end=1399062422.19 etime=1399061907.45 "
        "exe=/fs/somedirectoryobj2"
        "-bridgeringagg-automated-nodebug exec_host=MIR-04C00-77FF1-4096 group=unknown jobname=N/A mode=c16 qtime=1399061907.45 "
        "queue=somequeue09 resources_used.location=MIR-04C00-77FF1-4096 resources_used.nodect=4096 resources_used.walltime=0:04:30 "
        "session=unknown start=1399062151.69 user=userg",
        "05/02/2014 20:27:35;S;256107;Resource_List.ncpus=1024 Resource_List.nodect=1024 Resource_List.walltime=6:00:00 "
        "account=someaccount2 args= ctime=1398890600.49 "
        "cwd=/fs/somedirectoryobj "
        "etime=1398890600.49 exe=/fs/somedirectoryobj4"
        ".0_ep_30.0/run_ensemble_job.292 exec_host=MIR-48800-7BB71-1-1024 group=unknown jobname=npt2-job292 mode=script "
        "qtime=1398890600.49 queue=somequeue09 session=unknown start=1399062455.04 user=someuser93",
        "05/02/2014 20:28:11;S;257828;Resource_List.ncpus=32768 Resource_List.nodect=2048 Resource_List.walltime=4:00:00 "
        "account=someproject2 args=outFoamx_SecXY.input",
        "ctime=1399056664.55 cwd=/fs/somedirectoryobjE etime=1399056664.55 "
        "exe=/fs/somedirectoryobj6 exec_host=MIR-08000-3B3F1-2048 "
        "group=unknown jobname=N/A mode=c16 qtime=1399056664.55 queue=somequeue09 session=unknown start=1399062491.92 user=useri",
        "05/02/2014 20:28:43;S;257924;Resource_List.ncpus=1024 Resource_List.nodect=1024 Resource_List.walltime=0:40:00 "
        "account=LiAirBattery args= ctime=1399061768.84 cwd=/fs/somedirectoryobjG "
        "etime=1399061768.84 exe=/fs/somedirectoryobj5 "
        "exec_host=MIR-48000-7B371-1-1024 group=unknown jobname=N/A mode=script qtime=1399061768.84 queue=somequeue09 "
        "session=unknown start=1399062523.88 user=userj",
        "05/02/2014 20:29:19;S;257931;Resource_List.ncpus=65536 Resource_List.nodect=2048 Resource_List.walltime=2:00:00 "
        "account=Ray_Benard args= ctime=1399062080.19 cwd=/fs/somedirectoryobjF etime=1399062080.19 "
        "exe=/fs/somedirectoryobj7 exec_host=MIR-48400-7B7F1-2048 group=unknown jobname=N/A mode=c32 "
        "qtime=1399062080.19 queue=somequeue09 session=unknown start=1399062559.5 user=userk",
        "05/02/2014 20:29:39;Q;257933;queue=somequeue64",
    ]
    correct_lineno = [1, 2, 3, 5, 6, 7]
    parsed = []
    times_and_data = fix_newlines_extract_time(lines, include_lineno=True, version=2)
    assert len(times_and_data) == len(lines) - 1
    for idx, (logtime, lineno, data) in enumerate(times_and_data):
        parsed_line = parse_pbs_log_line(logtime, data)
        parsed.append(parsed_line)
        assert lineno == correct_lineno[idx]
    assert len(parsed) == len(lines) - 1


def test_fix_newlines_extract_time_02():
    """Fix for PBS_Utils.pbs_accounting.BadLineError"""
    lines = [
        "10/26/2021 16:24:55;D;52.pdw-s1;requestor=someuser73@pdw-s1.pdw.local",
        "10/26/2021 16:24:55;A;53.pdw-s1;Job deleted as result of dependency on job 52.pdw-s1",
    ]
    parsed = []
    times_and_data = fix_newlines_extract_time(lines)
    for logtime, data in times_and_data:
        parsed_line = parse_pbs_log_line(logtime, data)
        parsed.append(parsed_line)
    assert len(parsed) == len(lines)


def test_fix_newlines_extract_time_03():
    """an interesting a record."""
    lines = [
        r"07/17/2024 18:07:24;a;1624.somemachine;Variable_List=PBS_O_HOME=/fs/someuser,"
        "PBS_O_LANG=en_US.UTF-8,PBS_O_LOGNAME=someuser,"
        "PBS_O_PATH=/fs/somedirectoryobj12:/fs/somedirectoryobj13:/usr/local/bin:/usr/bin:/usr/local/sbin:/usr,"
        "PBS_O_SHELL=/bin/bash,PBS_O_HOST=login-01,PBS_O_WORKDIR=/fs/somedirectoryobj11,"
        "PBS_O_SYSTEM=Linux,SHELL=/bin/bash,HISTCONTROL=ignoredups,"
        "LS_COLORS=rs=0:di=01;34:ln=01;36:mh=00:pi=40;33:so=01;35:do=01;35:bd=40;33;01:cd=40;33;01:or=40;31;01:mi=01;37;41:su=37"
        ";41:sg=30;43:ca=30;41:tw=30;42:ow=34;42:st=37;44:ex=01;32:*.tar=01;31:*.tgz=01;31:*.arc=01;31:*.arj=01;31:*.taz=01;31"
        ":*.lha=01;31:*.lz4=01;31:*.lzh=01;31:*.lzma=01;31:*.tlz=01;31:*.txz=01;31:*.tzo=01;31:*.t7z=01;31:*.zip=01;31:*.z=01;31"
        ":*.dz=01;31:*.gz=01;31:*.lrz=01;31:*.lz=01;31:*.lzo=01;31:*.xz=01;31:*.zst=01;31:*.tzst=01;31:*.bz2=01;31:*.bz=01;31"
        ":*.tbz=01;31:*.tbz2=01;31:*.tz=01;31:*.deb=01;31:*.rpm=01;31:*.jar=01;31:*.war=01;31:*.ear=01;31:*.sar=01;31:*.rar=01;31"
        ":*.alz=01;31:*.ace=01;31:*.zoo=01;31:*.cpio=01;31:*.7z=01;31:*.rz=01;31:*.cab=01;31:*.wim=01;31:*.swm=01;31:*.dwm=01;31"
        ":*.esd=01;31:*.jpg=01;35:*.jpeg=01;35:*.mjpg=01;35:*.mjpeg=01;35:*.gif=01;35:*.bmp=01;35:*.pbm=01;35:*.pgm=01;35:*.ppm=01"
        ";35:*.tga=01;35:*.xbm=01;35:*.xpm=01;35:*.tif=01;35:*.tiff=01;35:*.png=01;35:*.svg=01;35:*.svgz=01;35:*.mng=01;35:*.pcx"
        "=01;35:*.mov=01;35:*.mpg=01;35:*.mpeg=01;35:*.m2v=01;35:*.mkv=01;35:*.webm=01;35:*.webp=01;35:*.ogm=01;35:*.mp4=01;35"
        ":*.m4v=01;35:*.mp4v=01;35:*.vob=01;35:*.qt=01;35:*.nuv=01;35:*.wmv=01;35:*.asf=01;35:*.rm=01;35:*.rmvb=01;35:*.flc=01;35"
        ":*.avi=01;35:*.fli=01;35:*.flv=01;35:*.gl=01;35:*.dl=01;35:*.xcf=01;35:*.xwd=01;35:*.yuv=01;35:*.cgm=01;35:*.emf=01;35"
        ":*.ogv=01;35:*.ogx=01;35:*.aac=01;36:*.au=01;36:*.flac=01;36:*.m4a=01;36:*.mid=01;36:*.midi=01;36:*.mka=01;36:*.mp3=01;36"
        ":*.mpc=01;36:*.ogg=01;36:*.ra=01;36:*.wav=01;36:*.oga=01;36:*.opus=01;36:*.spx=01;36:*.xspf=01;36:,"
        "LMOD_SETTARG_FULL_SUPPORT=no,CONDA_PROMPT_MODIFIER=(base),,LMOD_VERSION=8.7.32,"
        'SSH_TTY=/dev/pts/15,BASH_FUNC_ml%%=() {  eval "$($LMOD_DIR/ml_cmd "$@")",',
        r"},BASH_FUNC_which%%=() {  ( alias; eval ${which_declare} ) | /usr/bin/which --tty-only --read-alias --read-functions "
        r"--show-tilde --show-dot $@",
        r"},BASH_FUNC_module%%=() {  if [ -z \"${LMOD_SH_DBG_ON+x}\" ]; then",
        r" case \"$-\" in ",
        r" *v*x*)",
        r" __lmod_sh_dbg=\'vx\'",
        r" ;; *v*)" "",
        r" __lmod_sh_dbg=\'v\'",
        r" ;; *x*)",
        r" __lmod_sh_dbg=\'x\'",
        r" ;; esac; fi; if [ -n \"${__lmod_sh_dbg:-}\" ]; then",
        r"set +$__lmod_sh_dbg; echo \"Shell debugging temporarily silenced: export LMOD_SH_DBG_ON=1 for Lmod\'s output\" 1>&2; "
        r"fi; eval \"$($LMOD_CMD shel",
    ]
    parsed = []
    times_and_data = fix_newlines_extract_time(lines, include_lineno=True)
    print(times_and_data)
    assert len(times_and_data) == 1
    for idx, (logtime, lineno, data) in enumerate(times_and_data):
        print(f"{idx=} {logtime=} {data=}")
        parsed_line = parse_pbs_log_line(logtime, data)
        print(f" *** {parsed_line=} ***")
        parsed.append(parsed_line)


def input_pbs_accounting_place(ids=False) -> List[Tuple[str, str, list, bool]]:
    # comment, input, output, valid
    deck = [
        ("", "excl", {"arrangement": None, "sharing": "excl", "groups": []}, True),
        ("", "exclhost", {"arrangement": None, "sharing": "exclhost", "groups": []}, True),
        ("", "free", {"arrangement": "free", "sharing": None, "groups": []}, True),
        ("", "group=somegroup1", {"arrangement": None, "sharing": None, "groups": ["somegroup1"]}, True),
        ("", "pack", {"arrangement": "pack", "sharing": None, "groups": []}, True),
        ("", "scatter", {"arrangement": "scatter", "sharing": None, "groups": []}, True),
        ("", "scatter:excl", {"arrangement": "scatter", "sharing": "excl", "groups": []}, True),
        ("", "scatter:exclhost", {"arrangement": "scatter", "sharing": "exclhost", "groups": []}, True),
        ("", "scatter:group=somegroup0", {"arrangement": "scatter", "sharing": None, "groups": ["somegroup0"]}, True),
        ("", "vscatter", {"arrangement": "vscatter", "sharing": None, "groups": []}, True),
        (
            "",
            "scatter:group=somegroup4:group=somegroup8",
            {"arrangement": "scatter", "sharing": None, "groups": ["somegroup4", "somegroup8"]},
            True,
        ),
        ("", "scatter:shared:group=somegroup8", {"arrangement": "scatter", "sharing": "shared", "groups": ["somegroup8"]}, True),
        ("", "", {}, True),
    ]
    if ids:
        deck = [f"{i:0>2}" for i, _ in enumerate(deck)]
    return deck


@pytest.mark.parametrize("comment,value,correct,valid", input_pbs_accounting_place(), ids=input_pbs_accounting_place(ids=True))
def test_parse_place(comment, value, correct, valid):
    result = cast_to_place_dict(value)
    # print(f'("{comment}", "{value}", {result}, {valid}),')
    if valid:
        assert result == correct, value
    else:
        assert result != correct, value
    new_value = join_place_dict(result)
    # we cannot compare directly because dictionaries are not ordered, convert again
    new_result = cast_to_place_dict(new_value)
    if valid:
        assert result == new_result, value
    else:
        assert result != new_result, value


def input_pbs_accounting_select(ids=False) -> List[Tuple[str, str, list, bool]]:
    # comment, input, output, valid
    deck = [
        ("", "128:ngpus=4", [{"N": 128, "ngpus": "4"}], True),
        (
            "",
            "128:ncpus=64:ngpus=4:gputype=A100+32:ncpus=64:ngpus=2:gputype=A40",
            [{"N": 128, "ncpus": "64", "ngpus": "4", "gputype": "A100"}, {"N": 32, "ncpus": "64", "ngpus": "2", "gputype": "A40"}],
            True,
        ),
        ("", "128:ngpus=4:system=somemachine4", [{"N": 128, "ngpus": "4", "system": "somemachine4"}], True),
        ("", "128:system=somemachine4", [{"N": 128, "system": "somemachine4"}], True),
        (
            "",
            "128:system=somemachine4:ncpus=32:ngpus=4",
            [{"N": 128, "system": "somemachine4", "ncpus": "32", "ngpus": "4"}],
            True,
        ),
        (
            "",
            "128:system=somemachine4:vnode=^x3008c0s13b0n0",
            [{"N": 128, "system": "somemachine4", "vnode": "^x3008c0s13b0n0"}],
            True,
        ),
        (
            "",
            "128:system=somemachine4:vnode=^x3016c0s19b0n0",
            [{"N": 128, "system": "somemachine4", "vnode": "^x3016c0s19b0n0"}],
            True,
        ),
        (
            "",
            "128:system=somemachine4:vnode=^x3108c0s19b1n0",
            [{"N": 128, "system": "somemachine4", "vnode": "^x3108c0s19b1n0"}],
            True,
        ),
        (
            "",
            "128:vnode=^x3206c0s37b0n0+1:vnode=^x3202c0s31b1n0",
            [{"N": 128, "vnode": "^x3206c0s37b0n0"}, {"N": 1, "vnode": "^x3202c0s31b1n0"}],
            True,
        ),
        ("", "129:ncpus=64:ngpus=4", [{"N": 129, "ncpus": "64", "ngpus": "4"}], True),
        ("", "12:ncpus=12", [{"N": 12, "ncpus": "12"}], True),
        ("", "12:ncpus=32:ngpus=4", [{"N": 12, "ncpus": "32", "ngpus": "4"}], True),
        ("", "12:ncpus=64:ngpus=4", [{"N": 12, "ncpus": "64", "ngpus": "4"}], True),
        ("", "12:system=somemachine4", [{"N": 12, "system": "somemachine4"}], True),
        ("", "12:system=somemachine4:ncpus=32:ngpus=4", [{"N": 12, "system": "somemachine4", "ncpus": "32", "ngpus": "4"}], True),
        ("", "130:ncpus=64:ngpus=4", [{"N": 130, "ncpus": "64", "ngpus": "4"}], True),
        (
            "",
            "16:ncpus=64:build=False:debug=False:demand=False",
            [{"N": 16, "ncpus": "64", "build": "False", "debug": "False", "demand": "False"}],
            True,
        ),
        ("", "1:ncpus=1:system=generic_pbs", [{"N": 1, "ncpus": "1", "system": "generic_pbs"}], True),
        (
            "",
            "1:ncpus=1:obj_type=node:system=generic_pbs",
            [{"N": 1, "ncpus": "1", "obj_type": "node", "system": "generic_pbs"}],
            True,
        ),
        ("", "host=x3212c0s7b1n0:ncpus=64", [{"host": "x3212c0s7b1n0", "ncpus": "64"}], True),
        ("", "ngpus=1", [{"ngpus": "1"}], True),
        ("", "vnode=16", [{"vnode": "16"}], True),
        ("", "", [], True),
        ("", "vnode=1:system=somemachine4", [{"vnode": "1", "system": "somemachine4"}], True),
        (
            "",
            '1:ncpus=1:obj_type="bl:ck":system=generic_pbs',
            [{"N": 1, "ncpus": "1", "obj_type": "bl:ck", "system": "generic_pbs"}],
            True,
        ),
        (
            "",
            '1:ncpus=1:obj_type="bl:ck":system="gener+ic_pbs"',
            [{"N": 1, "ncpus": "1", "obj_type": "bl:ck", "system": "gener+ic_pbs"}],
            True,
        ),
        ("", "1:vnode=^p0015-01+2:vnode=^p0001-01", [{"N": 1, "vnode": "^p0015-01"}, {"N": 2, "vnode": "^p0001-01"}], True),
        (
            "",
            '1:ncpus=1:obj_type="bl\\:o\\+ck2":system=dog',
            [{"N": 1, "ncpus": "1", "obj_type": "bl\\:o\\+ck2", "system": "dog"}],
            True,
        ),
        ("", "2:ncpus=2:mpiprocs=2+1:nfpus=4", [{"N": 2, "ncpus": "2", "mpiprocs": "2"}, {"N": 1, "nfpus": "4"}], True),
        (
            "",
            '1:ncpus=1:obj_type="node+4:obj_type=block":system="gener:ic_pbs":tier_16="fun+fun"',
            [{"N": 1, "ncpus": "1", "obj_type": "node+4:obj_type=block", "system": "gener:ic_pbs", "tier_16": "fun+fun"}],
            True,
        ),
        (
            "",
            '1:ncpus=1:obj_type="bl:ck+arch=linux+host=gc-svr":system="gener+ic_pbs"',
            [{"N": 1, "ncpus": "1", "obj_type": "bl:ck+arch=linux+host=gc-svr", "system": "gener+ic_pbs"}],
            True,
        ),
    ]
    if ids:
        deck = [f"{i:0>2}" for i, _ in enumerate(deck)]
    return deck


@pytest.mark.parametrize("comment,value,correct,valid", input_pbs_accounting_select(), ids=input_pbs_accounting_select(ids=True))
def test_parse_select(comment, value, correct, valid):
    result = cast_to_select_list(value)
    # print(f"('{comment}', '{value}', {result}, {valid}),")
    if valid:
        assert result == correct
    else:
        assert result != correct
    new_value = join_select_list(result)
    # we cannot compare directly because dictionaries are not ordered, convert again
    new_result = cast_to_select_list(new_value)
    if valid:
        assert result == new_result, value
    else:
        assert result != new_result, value


def input_create_pbs_log_line_from_record(ids=False):
    deck = [
        ("", (datetime.now(), RecType.X, "1234", {}), "", True),
        ("", (datetime.now(), RecType.X, "1234", {}), "", True),
        ("", (datetime.now(), RecType.X, "1234", {"a": "1", "b": "2", "c": "3"}), "", True),
        ("", (datetime.now(), RecType.X, "1234", {"a": 1, "b": 2, "c": 3}), "", False),
    ]
    if ids:
        deck = [f"{i:0>2}" for i, _ in enumerate(deck)]
    return deck


@pytest.mark.parametrize(
    "comment,value,correct,valid", input_create_pbs_log_line_from_record(), ids=input_create_pbs_log_line_from_record(ids=True)
)
def test_create_pbs_log_line_from_record_00(comment, value, correct, valid):
    timestamp, record_type, identifier, dct = value
    edct = enhance_pbs_record(timestamp, record_type, identifier, dct=dct)
    record_obj = create_pbs_record(timestamp, record_type, identifier, dct)
    line = create_pbs_log_line_from_record(record_obj, attempt_str=True, iso8601=True)
    logtime, data = pbs_line_extract_time(line, iso8601=True)
    pbs_dct = parse_pbs_log_line(logtime, data)
    if valid:
        assert edct == pbs_dct
    else:
        assert edct != pbs_dct


@pytest.mark.parametrize(
    "comment,line,valid_record,valid_data", input_parse_pbs_log_line(), ids=input_parse_pbs_log_line(ids=True)
)
def test_create_pbs_log_line_from_record_01(comment, line, valid_record, valid_data):
    logtime, data = pbs_line_extract_time(line)
    if valid_record == BadLineError:
        with pytest.raises(BadLineError):
            parse_pbs_log_line(logtime, data)
    elif valid_record == Exception:
        with pytest.raises(Exception):
            parse_pbs_log_line(logtime, data)
    else:
        result = parse_pbs_log_line(logtime, data)
        pprint(result, width=132)
        assert result == valid_record, pformat(result, width=132)
        record_type = result["record"]
        identifier = result["identifier"]
        record_obj = create_pbs_record(logtime, record_type, identifier, result)
        new_line = create_pbs_log_line_from_record(record_obj, attempt_str=True, iso8601=False)
        logtime_new, data_new = pbs_line_extract_time(new_line)
        new_result = parse_pbs_log_line(logtime_new, data_new)
        assert result == new_result


def input_reservations(ids=False):
    """
    Test of PBS reservation record.
    What are nodes=?  What other data are we missing.  Pull all somemachine4 and somemachine3 data here.
    """
    deck = [
        (
            (
                "04/04/2023 13:21:00;B;someresv43.somemachine4;owner=user1234@somemachine4-login-01 name=NULL queue=someresv43 "
                "ctime=1680614422 start=1680614460 end=1680650460 duration=36000 nodes=(x3013c0s7b0n0:ncpus=64) "
                "Authorized_Users=user1234@somemachine4-login-01 Resource_List.allow_account_check_failure=True "
                "Resource_List.allow_negative_allocation=True Resource_List.backfill_factor=84600 "
                "Resource_List.backfill_max=50 Resource_List.base_score=0 Resource_List.enable_backfill=0 "
                "Resource_List.enable_fifo=0 Resource_List.enable_wfp=0 Resource_List.fifo_factor=1800 "
                "Resource_List.ncpus=64 Resource_List.ni_resource=somemachine4 Resource_List.preempt_targets=NONE "
                "Resource_List.score_boost=0 Resource_List.total_cpus=560 Resource_List.wfp_factor=100000 "
                "Resource_List.nodect=1 Resource_List.select=host=x3013c0s7b0n0:ncpus=64 Resource_List.place=exclhost "
                "Resource_List.walltime=10:00:00"
            ),
            28,
            33,
            {
                "Resource_List.allow_account_check_failure": "True",
                "Resource_List.allow_negative_allocation": "True",
                "Resource_List.backfill_factor": "84600",
                "Resource_List.backfill_max": "50",
                "Resource_List.base_score": "0",
                "Resource_List.enable_backfill": "0",
                "Resource_List.enable_fifo": "0",
                "Resource_List.enable_wfp": "0",
                "Resource_List.fifo_factor": "1800",
                "Resource_List.ncpus": 64,
                "Resource_List.ni_resource": "somemachine4",
                "Resource_List.nodect": 1,
                "Resource_List.place": {"arrangement": None, "groups": [], "sharing": "exclhost"},
                "Resource_List.preempt_targets": "NONE",
                "Resource_List.score_boost": "0",
                "Resource_List.select": [{"host": "x3013c0s7b0n0", "ncpus": "64"}],
                "Resource_List.total_cpus": "560",
                "Resource_List.walltime": 36000,
                "Resource_List.wfp_factor": "100000",
                "action": "begin",
                "Authorized_Users": ["user1234@somemachine4-login-01"],
                "ctime": datetime(2023, 4, 4, 13, 20, 22),
                "duration": 36000,
                "end": datetime(2023, 4, 4, 23, 21),
                "event_type": "RESERVATION",
                "identifier": "someresv43.somemachine4",
                "name": "NULL",
                "queue": "someresv43",
                "record": "B",
                "scheduler_timestamp": datetime(2023, 4, 4, 13, 21),
                "start": datetime(2023, 4, 4, 13, 21),
                "nodes": [
                    "x3013c0s7b0n0",
                ],
                "owner": "user1234@somemachine4-login-01",
            },
            [],
        ),
        (
            (
                "01/21/2023 22:00:00;B;someresv23.pbsserver;owner=someadmin@somehost"
                " name=ResvName queue=someresv23 ctime=1674250622 start=1674338400 end=1674424800 "
                "duration=86400 nodes=(x3006c0s13b1n0:ncpus=64)+(x3006c0s19b0n0:ncpus=64)+(x3006c0s19b1n0:ncpus=64)+("
                "x3006c0s1b0n0:ncpus=64)+(x3006c0s1b1n0:ncpus=64)+(x3006c0s25b0n0:ncpus=64)+(x3006c0s25b1n0:ncpus=64)+("
                "x3006c0s31b0n0:ncpus=64)+(x3006c0s31b1n0:ncpus=64)+(x3006c0s37b0n0:ncpus=64)+(x3006c0s37b1n0:ncpus=64)+("
                "x3006c0s7b0n0:ncpus=64)+(x3006c0s7b1n0:ncpus=64)+(x3007c0s13b0n0:ncpus=64)+(x3007c0s13b1n0:ncpus=64)+("
                "x3007c0s19b0n0:ncpus=64)+(x3007c0s19b1n0:ncpus=64)+(x3007c0s1b0n0:ncpus=64)+(x3007c0s1b1n0:ncpus=64)+("
                "x3007c0s25b0n0:ncpus=64)+(x3007c0s25b1n0:ncpus=64)+(x3007c0s31b0n0:ncpus=64)+(x3007c0s31b1n0:ncpus=64)+("
                "x3007c0s37b0n0:ncpus=64)+(x3007c0s37b1n0:ncpus=64)+(x3007c0s7b0n0:ncpus=64)+(x3007c0s7b1n0:ncpus=64)+("
                "x3008c0s13b0n0:ncpus=64)+(x3008c0s13b1n0:ncpus=64)+(x3008c0s19b0n0:ncpus=64)+(x3008c0s19b1n0:ncpus=64)+("
                "x3008c0s1b0n0:ncpus=64)+(x3008c0s1b1n0:ncpus=64)+(x3008c0s25b0n0:ncpus=64)+(x3008c0s25b1n0:ncpus=64)+("
                "x3008c0s31b0n0:ncpus=64)+(x3008c0s31b1n0:ncpus=64)+(x3008c0s37b0n0:ncpus=64)+(x3008c0s37b1n0:ncpus=64)+("
                "x3008c0s7b0n0:ncpus=64)+(x3008c0s7b1n0:ncpus=64)+(x3009c0s13b0n0:ncpus=64)+(x3009c0s13b1n0:ncpus=64)+("
                "x3009c0s19b0n0:ncpus=64)+(x3009c0s19b1n0:ncpus=64)+(x3009c0s1b0n0:ncpus=64)+(x3009c0s1b1n0:ncpus=64)+("
                "x3009c0s25b0n0:ncpus=64)+(x3009c0s25b1n0:ncpus=64)+(x3009c0s31b0n0:ncpus=64)+(x3009c0s31b1n0:ncpus=64)+("
                "x3009c0s37b0n0:ncpus=64)+(x3009c0s37b1n0:ncpus=64)+(x3009c0s7b0n0:ncpus=64)+(x3009c0s7b1n0:ncpus=64)+("
                "x3010c0s13b0n0:ncpus=64)+(x3010c0s13b1n0:ncpus=64)+(x3010c0s19b0n0:ncpus=64)+(x3010c0s19b1n0:ncpus=64)+("
                "x3010c0s1b0n0:ncpus=64) Authorized_Users=someadmin@somehost "
                "Resource_List.allow_account_check_failure=True Resource_List.allow_negative_allocation=True "
                "Resource_List.backfill_factor=84600 Resource_List.backfill_max=50 Resource_List.base_score=0 "
                "Resource_List.enable_backfill=0 Resource_List.enable_fifo=0 Resource_List.enable_wfp=0 "
                "Resource_List.fifo_factor=1800 Resource_List.ncpus=3840 Resource_List.ni_resource=somemachine4 "
                "Resource_List.preempt_targets=NONE Resource_List.score_boost=0 Resource_List.total_cpus=560 "
                "Resource_List.wfp_factor=100000 Resource_List.nodect=60 Resource_List.select=60 Resource_List.place=free "
                "Resource_List.walltime=24:00:00"
            ),
            28,
            33,
            {
                "Authorized_Users": ["someadmin@somehost"],
                "Resource_List.allow_account_check_failure": "True",
                "Resource_List.allow_negative_allocation": "True",
                "Resource_List.backfill_factor": "84600",
                "Resource_List.backfill_max": "50",
                "Resource_List.base_score": "0",
                "Resource_List.enable_backfill": "0",
                "Resource_List.enable_fifo": "0",
                "Resource_List.enable_wfp": "0",
                "Resource_List.fifo_factor": "1800",
                "Resource_List.ncpus": 3840,
                "Resource_List.ni_resource": "somemachine4",
                "Resource_List.nodect": 60,
                "Resource_List.place": {"arrangement": "free", "groups": [], "sharing": None},
                "Resource_List.preempt_targets": "NONE",
                "Resource_List.score_boost": "0",
                "Resource_List.select": [{"N": 60}],
                "Resource_List.total_cpus": "560",
                "Resource_List.walltime": 86400,
                "Resource_List.wfp_factor": "100000",
                "action": "begin",
                "ctime": datetime(2023, 1, 20, 21, 37, 2),
                "duration": 86400,
                "end": datetime(2023, 1, 22, 22, 0),
                "event_type": "RESERVATION",
                "identifier": "someresv23.pbsserver",
                "name": "ResvName",
                "nodes": [
                    "x3006c0s13b1n0",
                    "x3006c0s19b0n0",
                    "x3006c0s19b1n0",
                    "x3006c0s1b0n0",
                    "x3006c0s1b1n0",
                    "x3006c0s25b0n0",
                    "x3006c0s25b1n0",
                    "x3006c0s31b0n0",
                    "x3006c0s31b1n0",
                    "x3006c0s37b0n0",
                    "x3006c0s37b1n0",
                    "x3006c0s7b0n0",
                    "x3006c0s7b1n0",
                    "x3007c0s13b0n0",
                    "x3007c0s13b1n0",
                    "x3007c0s19b0n0",
                    "x3007c0s19b1n0",
                    "x3007c0s1b0n0",
                    "x3007c0s1b1n0",
                    "x3007c0s25b0n0",
                    "x3007c0s25b1n0",
                    "x3007c0s31b0n0",
                    "x3007c0s31b1n0",
                    "x3007c0s37b0n0",
                    "x3007c0s37b1n0",
                    "x3007c0s7b0n0",
                    "x3007c0s7b1n0",
                    "x3008c0s13b0n0",
                    "x3008c0s13b1n0",
                    "x3008c0s19b0n0",
                    "x3008c0s19b1n0",
                    "x3008c0s1b0n0",
                    "x3008c0s1b1n0",
                    "x3008c0s25b0n0",
                    "x3008c0s25b1n0",
                    "x3008c0s31b0n0",
                    "x3008c0s31b1n0",
                    "x3008c0s37b0n0",
                    "x3008c0s37b1n0",
                    "x3008c0s7b0n0",
                    "x3008c0s7b1n0",
                    "x3009c0s13b0n0",
                    "x3009c0s13b1n0",
                    "x3009c0s19b0n0",
                    "x3009c0s19b1n0",
                    "x3009c0s1b0n0",
                    "x3009c0s1b1n0",
                    "x3009c0s25b0n0",
                    "x3009c0s25b1n0",
                    "x3009c0s31b0n0",
                    "x3009c0s31b1n0",
                    "x3009c0s37b0n0",
                    "x3009c0s37b1n0",
                    "x3009c0s7b0n0",
                    "x3009c0s7b1n0",
                    "x3010c0s13b0n0",
                    "x3010c0s13b1n0",
                    "x3010c0s19b0n0",
                    "x3010c0s19b1n0",
                    "x3010c0s1b0n0",
                ],
                "queue": "someresv23",
                "record": "B",
                "scheduler_timestamp": datetime(2023, 1, 21, 22, 0),
                "start": datetime(2023, 1, 21, 22, 0),
                "owner": "someadmin@somehost",
            },
            [],
        ),
    ]
    if ids:
        deck = [f"{i:0>2}" for i, _ in enumerate(deck)]
    return deck


@pytest.mark.parametrize(
    "line,len_key_value_dct,len_parse_dct,valid,valid_missing", input_reservations(), ids=input_reservations(ids=True)
)
def test_parse_pbs_reservations(line, len_key_value_dct, len_parse_dct, valid, valid_missing):
    logtime, data = pbs_line_extract_time(line)
    try:
        parse_dct = parse_pbs_log_line(logtime, data)
    except Parse_Error_PBS as e:
        parse_dct = {"EXCEPTION": e}
    record_type, identifier, record_keyvals = split_record(data)
    key_value_dct = parse_key_value_pairs(record_keyvals)
    missing_fields = []
    for field in key_value_dct.keys():
        try:
            parse_dct[field]
        except (KeyError, TypeError) as e:
            if field != "active_id":
                missing_fields.append(field)
    assert len_key_value_dct == len(key_value_dct)
    assert parse_dct == valid
    assert missing_fields == valid_missing
    assert parse_dct["ctime"] <= parse_dct["scheduler_timestamp"]
    if parse_dct["record"] == "K":
        assert parse_dct["stime"] <= parse_dct["etime"]
        assert parse_dct["stime"] <= parse_dct["scheduler_timestamp"]
        assert parse_dct["etime"] <= parse_dct["scheduler_timestamp"]
